/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball, Josh MacDonald, 
 * Copyright (C) 1997-1998 Jay Painter <jpaint@serv.net><jpaint@gimp.org>  
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GTK+ Team and others 1997-2000.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
 */

#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include "../include/gtkunilist.h"
#include "../include/gtkmarshalers.h"
#include <gdk/gdkkeysyms.h>

/* length of button_actions array */
#define MAX_BUTTON 5

/* the number rows memchunk expands at a time */
#define UNILIST_OPTIMUM_SIZE 64

/* the width of the column resize windows */
#define DRAG_WIDTH  6

/* minimum allowed width of a column */
#define COLUMN_MIN_WIDTH 5

/* this defigns the base grid spacing */
#define CELL_SPACING 1

/* added the horizontal space at the beginning and end of a row*/
#define COLUMN_INSET 3

/* used for auto-scrolling */
#define SCROLL_TIME  100

/* gives the top pixel of the given row in context of
 * the unilist's voffset */
#define ROW_TOP_YPIXEL(unilist, row) (((unilist)->row_height * (row)) + \
				    (((row) + 1) * CELL_SPACING) + \
				    (unilist)->voffset)

/* returns the row index from a y pixel location in the 
 * context of the unilist's voffset */
#define ROW_FROM_YPIXEL(unilist, y)  (((y) - (unilist)->voffset) / \
				    ((unilist)->row_height + CELL_SPACING))

/* gives the left pixel of the given column in context of
 * the unilist's hoffset */
#define COLUMN_LEFT_XPIXEL(unilist, colnum)  ((unilist)->column[(colnum)].area.x + \
					    (unilist)->hoffset)

/* returns the column index from a x pixel location in the 
 * context of the unilist's hoffset */
static inline gint
COLUMN_FROM_XPIXEL (GtkUniList * unilist,
		    gint x)
{
  gint i, cx;

  for (i = 0; i < unilist->columns; i++)
    if (unilist->column[i].visible)
      {
	cx = unilist->column[i].area.x + unilist->hoffset;

	if (x >= (cx - (COLUMN_INSET + CELL_SPACING)) &&
	    x <= (cx + unilist->column[i].area.width + COLUMN_INSET))
	  return i;
      }

  /* no match */
  return -1;
}

/* returns the top pixel of the given row in the context of
 * the list height */
#define ROW_TOP(unilist, row)        (((unilist)->row_height + CELL_SPACING) * (row))

/* returns the left pixel of the given column in the context of
 * the list width */
#define COLUMN_LEFT(unilist, colnum) ((unilist)->column[(colnum)].area.x)

/* returns the total height of the list */
#define LIST_HEIGHT(unilist)         (((unilist)->row_height * ((unilist)->rows)) + \
				    (CELL_SPACING * ((unilist)->rows + 1)))


/* returns the total width of the list */
static inline gint
LIST_WIDTH (GtkUniList * unilist) 
{
  gint last_column;

  for (last_column = unilist->columns - 1;
       last_column >= 0 && !unilist->column[last_column].visible; last_column--);

  if (last_column >= 0)
    return (unilist->column[last_column].area.x +
	    unilist->column[last_column].area.width +
	    COLUMN_INSET + CELL_SPACING);
  return 0;
}

/* returns the GList item for the nth row */
#define	ROW_ELEMENT(unilist, row)	(((row) == (unilist)->rows - 1) ? \
				 (unilist)->row_list_end : \
				 g_list_nth ((unilist)->row_list, (row)))


/* redraw the list if it's not frozen */
#define UNILIST_UNFROZEN(unilist)     (((GtkUniList*) (unilist))->freeze_count == 0)
#define	UNILIST_REFRESH(unilist)	G_STMT_START { \
  if (UNILIST_UNFROZEN (unilist)) \
    GTK_UNILIST_GET_CLASS (unilist)->refresh ((GtkUniList*) (unilist)); \
} G_STMT_END


/* Signals */
enum {
  SELECT_ROW,
  UNSELECT_ROW,
  ROW_MOVE,
  CLICK_COLUMN,
  RESIZE_COLUMN,
  TOGGLE_FOCUS_ROW,
  SELECT_ALL,
  UNSELECT_ALL,
  UNDO_SELECTION,
  START_SELECTION,
  END_SELECTION,
  TOGGLE_ADD_MODE,
  EXTEND_SELECTION,
  SCROLL_VERTICAL,
  SCROLL_HORIZONTAL,
  ABORT_COLUMN_RESIZE,
  SELECTION_CHANGED,
  LAST_SIGNAL
};

enum {
  SYNC_REMOVE,
  SYNC_INSERT
};

enum {
  ARG_0,
  ARG_N_COLUMNS,
  ARG_SHADOW_TYPE,
  ARG_SELECTION_MODE,
  ARG_ROW_HEIGHT,
  ARG_TITLES_ACTIVE,
  ARG_REORDERABLE,
  ARG_USE_DRAG_ICONS,
  ARG_SORT_TYPE
};

/* GtkUniList Methods */
static void     gtk_unilist_class_init  (GtkUniListClass         *klass);
static void     gtk_unilist_init        (GtkUniList              *unilist);
static GObject* gtk_unilist_constructor (GType                  type,
				       guint                  n_construct_properties,
				       GObjectConstructParam *construct_params);

/* GtkObject Methods */
static void gtk_unilist_destroy  (GtkObject *object);
static void gtk_unilist_finalize (GObject   *object);
static void gtk_unilist_set_arg  (GtkObject *object,
				GtkArg    *arg,
				guint      arg_id);
static void gtk_unilist_get_arg  (GtkObject *object,
				GtkArg    *arg,
				guint      arg_id);

/* GtkWidget Methods */
static void gtk_unilist_set_scroll_adjustments (GtkUniList      *unilist,
					      GtkAdjustment *hadjustment,
					      GtkAdjustment *vadjustment);
static void gtk_unilist_realize         (GtkWidget        *widget);
static void gtk_unilist_unrealize       (GtkWidget        *widget);
static void gtk_unilist_map             (GtkWidget        *widget);
static void gtk_unilist_unmap           (GtkWidget        *widget);
static gint gtk_unilist_expose          (GtkWidget        *widget,
			               GdkEventExpose   *event);
static gint gtk_unilist_button_press    (GtkWidget        *widget,
				       GdkEventButton   *event);
static gint gtk_unilist_button_release  (GtkWidget        *widget,
				       GdkEventButton   *event);
static gint gtk_unilist_motion          (GtkWidget        *widget, 
			               GdkEventMotion   *event);
static void gtk_unilist_size_request    (GtkWidget        *widget,
				       GtkRequisition   *requisition);
static void gtk_unilist_size_allocate   (GtkWidget        *widget,
				       GtkAllocation    *allocation);
static void gtk_unilist_draw_focus      (GtkWidget        *widget);
static gint gtk_unilist_focus_in        (GtkWidget        *widget,
				       GdkEventFocus    *event);
static gint gtk_unilist_focus_out       (GtkWidget        *widget,
				       GdkEventFocus    *event);
static gint gtk_unilist_focus           (GtkWidget        *widget,
				       GtkDirectionType  direction);
static void gtk_unilist_set_focus_child (GtkContainer     *container,
				       GtkWidget        *child);
static void gtk_unilist_style_set       (GtkWidget        *widget,
				       GtkStyle         *previous_style);
static void gtk_unilist_drag_begin      (GtkWidget        *widget,
				       GdkDragContext   *context);
static gint gtk_unilist_drag_motion     (GtkWidget        *widget,
				       GdkDragContext   *context,
				       gint              x,
				       gint              y,
				       guint             time);
static void gtk_unilist_drag_leave      (GtkWidget        *widget,
				       GdkDragContext   *context,
				       guint             time);
static void gtk_unilist_drag_end        (GtkWidget        *widget,
				       GdkDragContext   *context);
static gboolean gtk_unilist_drag_drop   (GtkWidget      *widget,
				       GdkDragContext *context,
				       gint            x,
				       gint            y,
				       guint           time);
static void gtk_unilist_drag_data_get   (GtkWidget        *widget,
				       GdkDragContext   *context,
				       GtkSelectionData *selection_data,
				       guint             info,
				       guint             time);
static void gtk_unilist_drag_data_received (GtkWidget        *widget,
					  GdkDragContext   *context,
					  gint              x,
					  gint              y,
					  GtkSelectionData *selection_data,
					  guint             info,
					  guint             time);

/* GtkContainer Methods */
static void gtk_unilist_forall          (GtkContainer  *container,
			               gboolean       include_internals,
			               GtkCallback    callback,
			               gpointer       callback_data);

/* Selection */
static void toggle_row                (GtkUniList      *unilist,
			               gint           row,
			               gint           column,
			               GdkEvent      *event);
static void real_select_row           (GtkUniList      *unilist,
			               gint           row,
			               gint           column,
			               GdkEvent      *event);
static void real_unselect_row         (GtkUniList      *unilist,
			               gint           row,
			               gint           column,
			               GdkEvent      *event);
static void update_extended_selection (GtkUniList      *unilist,
				       gint           row);
static GList *selection_find          (GtkUniList      *unilist,
			               gint           row_number,
			               GList         *row_list_element);
static void real_select_all           (GtkUniList      *unilist);
static void real_unselect_all         (GtkUniList      *unilist);
static void move_vertical             (GtkUniList      *unilist,
			               gint           row,
			               gfloat         align);
static void move_horizontal           (GtkUniList      *unilist,
			               gint           diff);
static void real_undo_selection       (GtkUniList      *unilist);
static void fake_unselect_all         (GtkUniList      *unilist,
			               gint           row);
static void fake_toggle_row           (GtkUniList      *unilist,
			               gint           row);
static void resync_selection          (GtkUniList      *unilist,
			               GdkEvent      *event);
static void sync_selection            (GtkUniList      *unilist,
	                               gint           row,
                                       gint           mode);
static void set_anchor                (GtkUniList      *unilist,
			               gboolean       add_mode,
			               gint           anchor,
			               gint           undo_anchor);
static void start_selection           (GtkUniList      *unilist);
static void end_selection             (GtkUniList      *unilist);
static gint selection_changed         (GtkUniList      *unilist);
static void toggle_add_mode           (GtkUniList      *unilist);
static void toggle_focus_row          (GtkUniList      *unilist);
static void extend_selection          (GtkUniList      *unilist,
			               GtkScrollType  scroll_type,
			               gfloat         position,
			               gboolean       auto_start_selection);
static gint get_selection_info        (GtkUniList       *unilist,
				       gint            x,
				       gint            y,
				       gint           *row,
				       gint           *column);

/* Scrolling */
static void move_focus_row     (GtkUniList      *unilist,
			        GtkScrollType  scroll_type,
			        gfloat         position);
static void scroll_horizontal  (GtkUniList      *unilist,
			        GtkScrollType  scroll_type,
			        gfloat         position);
static void scroll_vertical    (GtkUniList      *unilist,
			        GtkScrollType  scroll_type,
			        gfloat         position);
static void move_horizontal    (GtkUniList      *unilist,
				gint           diff);
static void move_vertical      (GtkUniList      *unilist,
				gint           row,
				gfloat         align);
static gint horizontal_timeout (GtkUniList      *unilist);
static gint vertical_timeout   (GtkUniList      *unilist);
static void remove_grab        (GtkUniList      *unilist);


/* Resize Columns */
static void draw_xor_line             (GtkUniList       *unilist);
static gint new_column_width          (GtkUniList       *unilist,
			               gint            column,
			               gint           *x);
static void column_auto_resize        (GtkUniList       *unilist,
				       GtkUniListRow    *unilist_row,
				       gint            column,
				       gint            old_width);
static void real_resize_column        (GtkUniList       *unilist,
				       gint            column,
				       gint            width);
static void abort_column_resize       (GtkUniList       *unilist);
static void cell_size_request         (GtkUniList       *unilist,
			               GtkUniListRow    *unilist_row,
			               gint            column,
				       GtkRequisition *requisition);

/* Buttons */
static void column_button_create      (GtkUniList       *unilist,
				       gint            column);
static void column_button_clicked     (GtkWidget      *widget,
				       gpointer        data);

/* Adjustments */
static void adjust_adjustments        (GtkUniList       *unilist,
				       gboolean        block_resize);
static void vadjustment_changed       (GtkAdjustment  *adjustment,
				       gpointer        data);
static void vadjustment_value_changed (GtkAdjustment  *adjustment,
				       gpointer        data);
static void hadjustment_changed       (GtkAdjustment  *adjustment,
				       gpointer        data);
static void hadjustment_value_changed (GtkAdjustment  *adjustment,
				       gpointer        data);

/* Drawing */
static void get_cell_style   (GtkUniList      *unilist,
			      GtkUniListRow   *unilist_row,
			      gint           state,
			      gint           column,
			      GtkStyle     **style,
			      GdkGC        **fg_gc,
			      GdkGC        **bg_gc);
static gint draw_cell_pixmap (GdkWindow     *window,
			      GdkRectangle  *clip_rectangle,
			      GdkGC         *fg_gc,
			      GdkPixmap     *pixmap,
			      GdkBitmap     *mask,
			      gint           x,
			      gint           y,
			      gint           width,
			      gint           height);
static void draw_row         (GtkUniList      *unilist,
			      GdkRectangle  *area,
			      gint           row,
			      GtkUniListRow   *unilist_row);
static void draw_rows        (GtkUniList      *unilist,
			      GdkRectangle  *area);
static void unilist_refresh    (GtkUniList      *unilist);
static void draw_drag_highlight (GtkUniList        *unilist,
				 GtkUniListRow     *dest_row,
				 gint             dest_row_number,
				 GtkUniListDragPos  drag_pos);
     
/* Size Allocation / Requisition */
static void size_allocate_title_buttons (GtkUniList *unilist);
static void size_allocate_columns       (GtkUniList *unilist,
					 gboolean  block_resize);
static gint list_requisition_width      (GtkUniList *unilist);

/* Memory Allocation/Distruction Routines */
static GtkUniListColumn *columns_new (GtkUniList      *unilist);
static void column_title_new       (GtkUniList      *unilist,
			            gint           column,
			            const gchar   *title);
static void columns_delete         (GtkUniList      *unilist);
static GtkUniListRow *row_new        (GtkUniList      *unilist);
static void row_delete             (GtkUniList      *unilist,
			            GtkUniListRow   *unilist_row);
static void set_cell_contents      (GtkUniList      *unilist,
			            GtkUniListRow   *unilist_row,
				    gint           column,
				    GtkCellType    type,
				    const gchar   *text,
				    guint8         spacing,
				    GdkPixmap     *pixmap,
				    GdkBitmap     *mask);
static gint real_insert_row        (GtkUniList      *unilist,
				    gint           row,
				    gchar         *text[]);
static void real_remove_row        (GtkUniList      *unilist,
				    gint           row);
static void real_clear             (GtkUniList      *unilist);

/* Sorting */
static gint default_compare        (GtkUniList      *unilist,
			            gconstpointer  row1,
			            gconstpointer  row2);
static void real_sort_list         (GtkUniList      *unilist);
static GList *gtk_unilist_merge      (GtkUniList      *unilist,
				    GList         *a,
				    GList         *b);
static GList *gtk_unilist_mergesort  (GtkUniList      *unilist,
				    GList         *list,
				    gint           num);
/* Misc */
static gboolean title_focus_in   (GtkUniList *unilist,
				  gint      dir);
static gboolean title_focus_move (GtkUniList *unilist,
				  gint      dir);

static void real_row_move             (GtkUniList  *unilist,
			               gint       source_row,
			               gint       dest_row);
static gint column_title_passive_func (GtkWidget *widget, 
				       GdkEvent  *event,
				       gpointer   data);
static void drag_dest_cell            (GtkUniList         *unilist,
				       gint              x,
				       gint              y,
				       GtkUniListDestInfo *dest_info);



static GtkContainerClass *parent_class = NULL;
static guint unilist_signals[LAST_SIGNAL] = {0};

static GtkTargetEntry unilist_target_table = { "gtk-unilist-drag-reorder", 0, 0};

GtkType
gtk_unilist_get_type (void)
{
  static GtkType unilist_type = 0;

  if (!unilist_type)
    {
      static const GtkTypeInfo unilist_info =
      {
	"GtkUniList",
	sizeof (GtkUniList),
	sizeof (GtkUniListClass),
	(GtkClassInitFunc) gtk_unilist_class_init,
	(GtkObjectInitFunc) gtk_unilist_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL,
      };

      unilist_type = gtk_type_unique (GTK_TYPE_CONTAINER, &unilist_info);
    }

  return unilist_type;
}

static void
gtk_unilist_class_init (GtkUniListClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;
  GtkContainerClass *container_class;
  GtkBindingSet *binding_set;

  gobject_class->constructor = gtk_unilist_constructor;

  object_class = (GtkObjectClass *) klass;
  widget_class = (GtkWidgetClass *) klass;
  container_class = (GtkContainerClass *) klass;

  parent_class = gtk_type_class (GTK_TYPE_CONTAINER);

  gobject_class->finalize = gtk_unilist_finalize;
  
  object_class->set_arg = gtk_unilist_set_arg;
  object_class->get_arg = gtk_unilist_get_arg;
  object_class->destroy = gtk_unilist_destroy;
  

  widget_class->realize = gtk_unilist_realize;
  widget_class->unrealize = gtk_unilist_unrealize;
  widget_class->map = gtk_unilist_map;
  widget_class->unmap = gtk_unilist_unmap;
  widget_class->button_press_event = gtk_unilist_button_press;
  widget_class->button_release_event = gtk_unilist_button_release;
  widget_class->motion_notify_event = gtk_unilist_motion;
  widget_class->expose_event = gtk_unilist_expose;
  widget_class->size_request = gtk_unilist_size_request;
  widget_class->size_allocate = gtk_unilist_size_allocate;
  widget_class->focus_in_event = gtk_unilist_focus_in;
  widget_class->focus_out_event = gtk_unilist_focus_out;
  widget_class->style_set = gtk_unilist_style_set;
  widget_class->drag_begin = gtk_unilist_drag_begin;
  widget_class->drag_end = gtk_unilist_drag_end;
  widget_class->drag_motion = gtk_unilist_drag_motion;
  widget_class->drag_leave = gtk_unilist_drag_leave;
  widget_class->drag_drop = gtk_unilist_drag_drop;
  widget_class->drag_data_get = gtk_unilist_drag_data_get;
  widget_class->drag_data_received = gtk_unilist_drag_data_received;
  widget_class->focus = gtk_unilist_focus;
  
  /* container_class->add = NULL; use the default GtkContainerClass warning */
  /* container_class->remove=NULL; use the default GtkContainerClass warning */

  container_class->forall = gtk_unilist_forall;
  container_class->set_focus_child = gtk_unilist_set_focus_child;

  klass->set_scroll_adjustments = gtk_unilist_set_scroll_adjustments;
  klass->refresh = unilist_refresh;
  klass->select_row = real_select_row;
  klass->unselect_row = real_unselect_row;
  klass->row_move = real_row_move;
  klass->undo_selection = real_undo_selection;
  klass->resync_selection = resync_selection;
  klass->selection_find = selection_find;
  klass->selection_changed = selection_changed;
  klass->click_column = NULL;
  klass->resize_column = real_resize_column;
  klass->draw_row = draw_row;
  klass->draw_drag_highlight = draw_drag_highlight;
  klass->insert_row = real_insert_row;
  klass->remove_row = real_remove_row;
  klass->clear = real_clear;
  klass->sort_list = real_sort_list;
  klass->select_all = real_select_all;
  klass->unselect_all = real_unselect_all;
  klass->fake_unselect_all = fake_unselect_all;
  klass->scroll_horizontal = scroll_horizontal;
  klass->scroll_vertical = scroll_vertical;
  klass->extend_selection = extend_selection;
  klass->toggle_focus_row = toggle_focus_row;
  klass->toggle_add_mode = toggle_add_mode;
  klass->start_selection = start_selection;
  klass->end_selection = end_selection;
  klass->abort_column_resize = abort_column_resize;
  klass->set_cell_contents = set_cell_contents;
  klass->cell_size_request = cell_size_request;

  gtk_object_add_arg_type ("GtkUniList::n_columns",
			   GTK_TYPE_UINT,
			   GTK_ARG_READWRITE | GTK_ARG_CONSTRUCT_ONLY,
			   ARG_N_COLUMNS);
  gtk_object_add_arg_type ("GtkUniList::shadow_type",
			   GTK_TYPE_SHADOW_TYPE,
			   GTK_ARG_READWRITE,
			   ARG_SHADOW_TYPE);
  gtk_object_add_arg_type ("GtkUniList::selection_mode",
			   GTK_TYPE_SELECTION_MODE,
			   GTK_ARG_READWRITE,
			   ARG_SELECTION_MODE);
  gtk_object_add_arg_type ("GtkUniList::row_height",
			   GTK_TYPE_UINT,
			   GTK_ARG_READWRITE,
			   ARG_ROW_HEIGHT);
  gtk_object_add_arg_type ("GtkUniList::reorderable",
			   GTK_TYPE_BOOL,
			   GTK_ARG_READWRITE,
			   ARG_REORDERABLE);
  gtk_object_add_arg_type ("GtkUniList::titles_active",
			   GTK_TYPE_BOOL,
			   GTK_ARG_READWRITE,
			   ARG_TITLES_ACTIVE);
  gtk_object_add_arg_type ("GtkUniList::use_drag_icons",
			   GTK_TYPE_BOOL,
			   GTK_ARG_READWRITE,
			   ARG_USE_DRAG_ICONS);
  gtk_object_add_arg_type ("GtkUniList::sort_type",
			   GTK_TYPE_SORT_TYPE,
			   GTK_ARG_READWRITE,
			   ARG_SORT_TYPE);  

  widget_class->set_scroll_adjustments_signal =
    gtk_signal_new ("set_scroll_adjustments",
		    GTK_RUN_LAST,
		    GTK_CLASS_TYPE (object_class),
		    GTK_SIGNAL_OFFSET (GtkUniListClass, set_scroll_adjustments),
		    _gtk_marshal_VOID__OBJECT_OBJECT,
		    GTK_TYPE_NONE, 2, GTK_TYPE_ADJUSTMENT, GTK_TYPE_ADJUSTMENT);

  unilist_signals[SELECT_ROW] =
    gtk_signal_new ("select_row",
		    GTK_RUN_FIRST,
		    GTK_CLASS_TYPE (object_class),
		    GTK_SIGNAL_OFFSET (GtkUniListClass, select_row),
		    _gtk_marshal_VOID__INT_INT_BOXED,
		    GTK_TYPE_NONE, 3,
		    GTK_TYPE_INT,
		    GTK_TYPE_INT,
		    GDK_TYPE_EVENT | G_SIGNAL_TYPE_STATIC_SCOPE);
  unilist_signals[UNSELECT_ROW] =
    gtk_signal_new ("unselect_row",
		    GTK_RUN_FIRST,
		    GTK_CLASS_TYPE (object_class),
		    GTK_SIGNAL_OFFSET (GtkUniListClass, unselect_row),
		    _gtk_marshal_VOID__INT_INT_BOXED,
		    GTK_TYPE_NONE, 3, GTK_TYPE_INT,
		    GTK_TYPE_INT, GDK_TYPE_EVENT);
  unilist_signals[ROW_MOVE] =
    gtk_signal_new ("row_move",
		    GTK_RUN_LAST,
		    GTK_CLASS_TYPE (object_class),
		    GTK_SIGNAL_OFFSET (GtkUniListClass, row_move),
		    _gtk_marshal_VOID__INT_INT,
		    GTK_TYPE_NONE, 2, GTK_TYPE_INT, GTK_TYPE_INT);
  unilist_signals[CLICK_COLUMN] =
    gtk_signal_new ("click_column",
		    GTK_RUN_FIRST,
		    GTK_CLASS_TYPE (object_class),
		    GTK_SIGNAL_OFFSET (GtkUniListClass, click_column),
		    _gtk_marshal_VOID__INT,
		    GTK_TYPE_NONE, 1, GTK_TYPE_INT);
  unilist_signals[RESIZE_COLUMN] =
    gtk_signal_new ("resize_column",
		    GTK_RUN_LAST,
		    GTK_CLASS_TYPE (object_class),
		    GTK_SIGNAL_OFFSET (GtkUniListClass, resize_column),
		    _gtk_marshal_VOID__INT_INT,
		    GTK_TYPE_NONE, 2, GTK_TYPE_INT, GTK_TYPE_INT);

  unilist_signals[TOGGLE_FOCUS_ROW] =
    gtk_signal_new ("toggle_focus_row",
                    GTK_RUN_LAST | GTK_RUN_ACTION,
                    GTK_CLASS_TYPE (object_class),
                    GTK_SIGNAL_OFFSET (GtkUniListClass, toggle_focus_row),
                    _gtk_marshal_VOID__VOID,
                    GTK_TYPE_NONE, 0);
  unilist_signals[SELECTION_CHANGED] =
    gtk_signal_new ("selection_changed",
                    GTK_RUN_LAST | GTK_RUN_ACTION,
                    GTK_CLASS_TYPE (object_class),
                    GTK_SIGNAL_OFFSET (GtkUniListClass, selection_changed),
                    _gtk_marshal_VOID__VOID,
                    GTK_TYPE_NONE, 0);
  unilist_signals[SELECT_ALL] =
    gtk_signal_new ("select_all",
                    GTK_RUN_LAST | GTK_RUN_ACTION,
                    GTK_CLASS_TYPE (object_class),
                    GTK_SIGNAL_OFFSET (GtkUniListClass, select_all),
                    _gtk_marshal_VOID__VOID,
                    GTK_TYPE_NONE, 0);
  unilist_signals[UNSELECT_ALL] =
    gtk_signal_new ("unselect_all",
                    GTK_RUN_LAST | GTK_RUN_ACTION,
                    GTK_CLASS_TYPE (object_class),
                    GTK_SIGNAL_OFFSET (GtkUniListClass, unselect_all),
                    _gtk_marshal_VOID__VOID,
                    GTK_TYPE_NONE, 0);
  unilist_signals[UNDO_SELECTION] =
    gtk_signal_new ("undo_selection",
		    GTK_RUN_LAST | GTK_RUN_ACTION,
		    GTK_CLASS_TYPE (object_class),
		    GTK_SIGNAL_OFFSET (GtkUniListClass, undo_selection),
		    _gtk_marshal_VOID__VOID,
		    GTK_TYPE_NONE, 0);
  unilist_signals[START_SELECTION] =
    gtk_signal_new ("start_selection",
		    GTK_RUN_LAST | GTK_RUN_ACTION,
		    GTK_CLASS_TYPE (object_class),
		    GTK_SIGNAL_OFFSET (GtkUniListClass, start_selection),
		    _gtk_marshal_VOID__VOID,
		    GTK_TYPE_NONE, 0);
  unilist_signals[END_SELECTION] =
    gtk_signal_new ("end_selection",
		    GTK_RUN_LAST | GTK_RUN_ACTION,
		    GTK_CLASS_TYPE (object_class),
		    GTK_SIGNAL_OFFSET (GtkUniListClass, end_selection),
		    _gtk_marshal_VOID__VOID,
		    GTK_TYPE_NONE, 0);
  unilist_signals[TOGGLE_ADD_MODE] =
    gtk_signal_new ("toggle_add_mode",
		    GTK_RUN_LAST | GTK_RUN_ACTION,
		    GTK_CLASS_TYPE (object_class),
		    GTK_SIGNAL_OFFSET (GtkUniListClass, toggle_add_mode),
		    _gtk_marshal_VOID__VOID,
		    GTK_TYPE_NONE, 0);
  unilist_signals[EXTEND_SELECTION] =
    gtk_signal_new ("extend_selection",
                    GTK_RUN_LAST | GTK_RUN_ACTION,
                    GTK_CLASS_TYPE (object_class),
                    GTK_SIGNAL_OFFSET (GtkUniListClass, extend_selection),
                    _gtk_marshal_VOID__ENUM_FLOAT_BOOLEAN,
                    GTK_TYPE_NONE, 3,
		    GTK_TYPE_SCROLL_TYPE, GTK_TYPE_FLOAT, GTK_TYPE_BOOL);
  unilist_signals[SCROLL_VERTICAL] =
    gtk_signal_new ("scroll_vertical",
                    GTK_RUN_LAST | GTK_RUN_ACTION,
                    GTK_CLASS_TYPE (object_class),
                    GTK_SIGNAL_OFFSET (GtkUniListClass, scroll_vertical),
                    _gtk_marshal_VOID__ENUM_FLOAT,
                    GTK_TYPE_NONE, 2, GTK_TYPE_SCROLL_TYPE, GTK_TYPE_FLOAT);
  unilist_signals[SCROLL_HORIZONTAL] =
    gtk_signal_new ("scroll_horizontal",
                    GTK_RUN_LAST | GTK_RUN_ACTION,
                    GTK_CLASS_TYPE (object_class),
                    GTK_SIGNAL_OFFSET (GtkUniListClass, scroll_horizontal),
                    _gtk_marshal_VOID__ENUM_FLOAT,
                    GTK_TYPE_NONE, 2, GTK_TYPE_SCROLL_TYPE, GTK_TYPE_FLOAT);
  unilist_signals[ABORT_COLUMN_RESIZE] =
    gtk_signal_new ("abort_column_resize",
                    GTK_RUN_LAST | GTK_RUN_ACTION,
                    GTK_CLASS_TYPE (object_class),
                    GTK_SIGNAL_OFFSET (GtkUniListClass, abort_column_resize),
                    _gtk_marshal_VOID__VOID,
                    GTK_TYPE_NONE, 0);

  binding_set = gtk_binding_set_by_class (klass);
  
  gtk_binding_entry_add_signal (binding_set, GDK_Up, 0,
				"scroll_vertical", 2,
				GTK_TYPE_ENUM, GTK_SCROLL_STEP_BACKWARD,
				GTK_TYPE_FLOAT, 0.0);
  gtk_binding_entry_add_signal (binding_set, GDK_KP_Up, 0,
				"scroll_vertical", 2,
				GTK_TYPE_ENUM, GTK_SCROLL_STEP_BACKWARD,
				GTK_TYPE_FLOAT, 0.0);
  gtk_binding_entry_add_signal (binding_set, GDK_Down, 0,
				"scroll_vertical", 2,
				GTK_TYPE_ENUM, GTK_SCROLL_STEP_FORWARD,
				GTK_TYPE_FLOAT, 0.0);
  gtk_binding_entry_add_signal (binding_set, GDK_KP_Down, 0,
				"scroll_vertical", 2,
				GTK_TYPE_ENUM, GTK_SCROLL_STEP_FORWARD,
				GTK_TYPE_FLOAT, 0.0);
  gtk_binding_entry_add_signal (binding_set, GDK_Page_Up, 0,
				"scroll_vertical", 2,
				GTK_TYPE_ENUM, GTK_SCROLL_PAGE_BACKWARD,
				GTK_TYPE_FLOAT, 0.0);
  gtk_binding_entry_add_signal (binding_set, GDK_KP_Page_Up, 0,
				"scroll_vertical", 2,
				GTK_TYPE_ENUM, GTK_SCROLL_PAGE_BACKWARD,
				GTK_TYPE_FLOAT, 0.0);
  gtk_binding_entry_add_signal (binding_set, GDK_Page_Down, 0,
				"scroll_vertical", 2,
				GTK_TYPE_ENUM, GTK_SCROLL_PAGE_FORWARD,
				GTK_TYPE_FLOAT, 0.0);
  gtk_binding_entry_add_signal (binding_set, GDK_KP_Page_Down, 0,
				"scroll_vertical", 2,
				GTK_TYPE_ENUM, GTK_SCROLL_PAGE_FORWARD,
				GTK_TYPE_FLOAT, 0.0);
  gtk_binding_entry_add_signal (binding_set, GDK_Home, GDK_CONTROL_MASK,
				"scroll_vertical", 2,
				GTK_TYPE_ENUM, GTK_SCROLL_JUMP,
				GTK_TYPE_FLOAT, 0.0);
  gtk_binding_entry_add_signal (binding_set, GDK_KP_Home, GDK_CONTROL_MASK,
				"scroll_vertical", 2,
				GTK_TYPE_ENUM, GTK_SCROLL_JUMP,
				GTK_TYPE_FLOAT, 0.0);
  gtk_binding_entry_add_signal (binding_set, GDK_End, GDK_CONTROL_MASK,
				"scroll_vertical", 2,
				GTK_TYPE_ENUM, GTK_SCROLL_JUMP,
				GTK_TYPE_FLOAT, 1.0);
  gtk_binding_entry_add_signal (binding_set, GDK_KP_End, GDK_CONTROL_MASK,
				"scroll_vertical", 2,
				GTK_TYPE_ENUM, GTK_SCROLL_JUMP,
				GTK_TYPE_FLOAT, 1.0);
  
  gtk_binding_entry_add_signal (binding_set, GDK_Up, GDK_SHIFT_MASK,
				"extend_selection", 3,
				GTK_TYPE_ENUM, GTK_SCROLL_STEP_BACKWARD,
				GTK_TYPE_FLOAT, 0.0, GTK_TYPE_BOOL, TRUE);
  gtk_binding_entry_add_signal (binding_set, GDK_KP_Up, GDK_SHIFT_MASK,
				"extend_selection", 3,
				GTK_TYPE_ENUM, GTK_SCROLL_STEP_BACKWARD,
				GTK_TYPE_FLOAT, 0.0, GTK_TYPE_BOOL, TRUE);
  gtk_binding_entry_add_signal (binding_set, GDK_Down, GDK_SHIFT_MASK,
				"extend_selection", 3,
				GTK_TYPE_ENUM, GTK_SCROLL_STEP_FORWARD,
				GTK_TYPE_FLOAT, 0.0, GTK_TYPE_BOOL, TRUE);
  gtk_binding_entry_add_signal (binding_set, GDK_KP_Down, GDK_SHIFT_MASK,
				"extend_selection", 3,
				GTK_TYPE_ENUM, GTK_SCROLL_STEP_FORWARD,
				GTK_TYPE_FLOAT, 0.0, GTK_TYPE_BOOL, TRUE);
  gtk_binding_entry_add_signal (binding_set, GDK_Page_Up, GDK_SHIFT_MASK,
				"extend_selection", 3,
				GTK_TYPE_ENUM, GTK_SCROLL_PAGE_BACKWARD,
				GTK_TYPE_FLOAT, 0.0, GTK_TYPE_BOOL, TRUE);
  gtk_binding_entry_add_signal (binding_set, GDK_KP_Page_Up, GDK_SHIFT_MASK,
				"extend_selection", 3,
				GTK_TYPE_ENUM, GTK_SCROLL_PAGE_BACKWARD,
				GTK_TYPE_FLOAT, 0.0, GTK_TYPE_BOOL, TRUE);
  gtk_binding_entry_add_signal (binding_set, GDK_Page_Down, GDK_SHIFT_MASK,
				"extend_selection", 3,
				GTK_TYPE_ENUM, GTK_SCROLL_PAGE_FORWARD,
				GTK_TYPE_FLOAT, 0.0, GTK_TYPE_BOOL, TRUE);
  gtk_binding_entry_add_signal (binding_set, GDK_KP_Page_Down, GDK_SHIFT_MASK,
				"extend_selection", 3,
				GTK_TYPE_ENUM, GTK_SCROLL_PAGE_FORWARD,
				GTK_TYPE_FLOAT, 0.0, GTK_TYPE_BOOL, TRUE);
  gtk_binding_entry_add_signal (binding_set, GDK_Home,
				GDK_SHIFT_MASK | GDK_CONTROL_MASK,
				"extend_selection", 3,
				GTK_TYPE_ENUM, GTK_SCROLL_JUMP,
				GTK_TYPE_FLOAT, 0.0, GTK_TYPE_BOOL, TRUE);
  gtk_binding_entry_add_signal (binding_set, GDK_KP_Home,
                                GDK_SHIFT_MASK | GDK_CONTROL_MASK,
				"extend_selection", 3,
				GTK_TYPE_ENUM, GTK_SCROLL_JUMP,
				GTK_TYPE_FLOAT, 0.0, GTK_TYPE_BOOL, TRUE);
  gtk_binding_entry_add_signal (binding_set, GDK_End,
				GDK_SHIFT_MASK | GDK_CONTROL_MASK,
				"extend_selection", 3,
				GTK_TYPE_ENUM, GTK_SCROLL_JUMP,
				GTK_TYPE_FLOAT, 1.0, GTK_TYPE_BOOL, TRUE);
  gtk_binding_entry_add_signal (binding_set, GDK_KP_End,
				GDK_SHIFT_MASK | GDK_CONTROL_MASK,
				"extend_selection", 3,
				GTK_TYPE_ENUM, GTK_SCROLL_JUMP,
				GTK_TYPE_FLOAT, 1.0, GTK_TYPE_BOOL, TRUE);


  gtk_binding_entry_add_signal (binding_set, GDK_Left, 0,
				"scroll_horizontal", 2,
				GTK_TYPE_ENUM, GTK_SCROLL_STEP_BACKWARD,
				GTK_TYPE_FLOAT, 0.0);
  gtk_binding_entry_add_signal (binding_set, GDK_KP_Left, 0,
				"scroll_horizontal", 2,
				GTK_TYPE_ENUM, GTK_SCROLL_STEP_BACKWARD,
				GTK_TYPE_FLOAT, 0.0);

  gtk_binding_entry_add_signal (binding_set, GDK_Right, 0,
				"scroll_horizontal", 2,
				GTK_TYPE_ENUM, GTK_SCROLL_STEP_FORWARD,
				GTK_TYPE_FLOAT, 0.0);
  gtk_binding_entry_add_signal (binding_set, GDK_KP_Right, 0,
				"scroll_horizontal", 2,
				GTK_TYPE_ENUM, GTK_SCROLL_STEP_FORWARD,
				GTK_TYPE_FLOAT, 0.0);

  gtk_binding_entry_add_signal (binding_set, GDK_Home, 0,
				"scroll_horizontal", 2,
				GTK_TYPE_ENUM, GTK_SCROLL_JUMP,
				GTK_TYPE_FLOAT, 0.0);
  gtk_binding_entry_add_signal (binding_set, GDK_KP_Home, 0,
				"scroll_horizontal", 2,
				GTK_TYPE_ENUM, GTK_SCROLL_JUMP,
				GTK_TYPE_FLOAT, 0.0);
  
  gtk_binding_entry_add_signal (binding_set, GDK_End, 0,
				"scroll_horizontal", 2,
				GTK_TYPE_ENUM, GTK_SCROLL_JUMP,
				GTK_TYPE_FLOAT, 1.0);
  gtk_binding_entry_add_signal (binding_set, GDK_KP_End, 0,
				"scroll_horizontal", 2,
				GTK_TYPE_ENUM, GTK_SCROLL_JUMP,
				GTK_TYPE_FLOAT, 1.0);

//  gtk_binding_entry_add_signal (binding_set, GDK_Insert, 0,
//				"extend_selection", 0);
//  gtk_binding_entry_add_signal (binding_set, GDK_space, 0,
//				"end_selection", 0);
  gtk_binding_entry_add_signal (binding_set, GDK_Escape, 0,
				"undo_selection", 0);
  gtk_binding_entry_add_signal (binding_set, GDK_Escape, 0,
				"abort_column_resize", 0);
  gtk_binding_entry_add_signal (binding_set, GDK_space, 0,
				"toggle_focus_row", 0);
/*  gtk_binding_entry_add_signal (binding_set, GDK_KP_Space, 0,
				"toggle_focus_row", 0);  
  gtk_binding_entry_add_signal (binding_set, GDK_space, GDK_CONTROL_MASK,
				"toggle_add_mode", 0);
  gtk_binding_entry_add_signal (binding_set, GDK_KP_Space, GDK_CONTROL_MASK,
				"toggle_add_mode", 0);
  gtk_binding_entry_add_signal (binding_set, GDK_slash, GDK_CONTROL_MASK,
				"select_all", 0);
  gtk_binding_entry_add_signal (binding_set, GDK_KP_Divide, GDK_CONTROL_MASK,
				"select_all", 0);
  gtk_binding_entry_add_signal (binding_set, '\\', GDK_CONTROL_MASK,
				"unselect_all", 0);
*/

  gtk_binding_entry_add_signal (binding_set, GDK_Shift_L,
				GDK_RELEASE_MASK | GDK_SHIFT_MASK,
				"end_selection", 0);
  gtk_binding_entry_add_signal (binding_set, GDK_Shift_R,
				GDK_RELEASE_MASK | GDK_SHIFT_MASK,
				"end_selection", 0);
  gtk_binding_entry_add_signal (binding_set, GDK_Shift_L,
				GDK_RELEASE_MASK | GDK_SHIFT_MASK |
				GDK_CONTROL_MASK,
				"end_selection", 0);
  gtk_binding_entry_add_signal (binding_set, GDK_Shift_R,
				GDK_RELEASE_MASK | GDK_SHIFT_MASK |
				GDK_CONTROL_MASK,
				"end_selection", 0);

}

static void
gtk_unilist_set_arg (GtkObject      *object,
		   GtkArg         *arg,
		   guint           arg_id)
{
  GtkUniList *unilist;

  unilist = GTK_UNILIST (object);

  switch (arg_id)
    {
    case ARG_N_COLUMNS: /* only set at construction time */
      unilist->columns = MAX (1, GTK_VALUE_UINT (*arg));
      break;
    case ARG_SHADOW_TYPE:
      gtk_unilist_set_shadow_type (unilist, GTK_VALUE_ENUM (*arg));
      break;
    case ARG_SELECTION_MODE:
      gtk_unilist_set_selection_mode (unilist, GTK_VALUE_ENUM (*arg));
      break;
    case ARG_ROW_HEIGHT:
      gtk_unilist_set_row_height (unilist, GTK_VALUE_UINT (*arg));
      break;
    case ARG_REORDERABLE:
      gtk_unilist_set_reorderable (unilist, GTK_VALUE_BOOL (*arg));
      break;
    case ARG_TITLES_ACTIVE:
      if (GTK_VALUE_BOOL (*arg))
	gtk_unilist_column_titles_active (unilist);
      else
	gtk_unilist_column_titles_passive (unilist);
      break;
    case ARG_USE_DRAG_ICONS:
      gtk_unilist_set_use_drag_icons (unilist, GTK_VALUE_BOOL (*arg));
      break;
    case ARG_SORT_TYPE:
      gtk_unilist_set_sort_type (unilist, GTK_VALUE_ENUM (*arg));
      break;
    }
}

static void
gtk_unilist_get_arg (GtkObject      *object,
		   GtkArg         *arg,
		   guint           arg_id)
{
  GtkUniList *unilist;

  unilist = GTK_UNILIST (object);

  switch (arg_id)
    {
      guint i;

    case ARG_N_COLUMNS:
      GTK_VALUE_UINT (*arg) = unilist->columns;
      break;
    case ARG_SHADOW_TYPE:
      GTK_VALUE_ENUM (*arg) = unilist->shadow_type;
      break;
    case ARG_SELECTION_MODE:
      GTK_VALUE_ENUM (*arg) = unilist->selection_mode;
      break;
    case ARG_ROW_HEIGHT:
      GTK_VALUE_UINT (*arg) = GTK_UNILIST_ROW_HEIGHT_SET(unilist) ? unilist->row_height : 0;
      break;
    case ARG_REORDERABLE:
      GTK_VALUE_BOOL (*arg) = GTK_UNILIST_REORDERABLE (unilist);
      break;
    case ARG_TITLES_ACTIVE:
      GTK_VALUE_BOOL (*arg) = TRUE;
      for (i = 0; i < unilist->columns; i++)
	if (unilist->column[i].button &&
	    !GTK_WIDGET_SENSITIVE (unilist->column[i].button))
	  {
	    GTK_VALUE_BOOL (*arg) = FALSE;
	    break;
	  }
      break;
    case ARG_USE_DRAG_ICONS:
      GTK_VALUE_BOOL (*arg) = GTK_UNILIST_USE_DRAG_ICONS (unilist);
      break;
    case ARG_SORT_TYPE:
      GTK_VALUE_ENUM (*arg) = unilist->sort_type;
      break;
    default:
      arg->type = GTK_TYPE_INVALID;
      break;
    }
}

static void
gtk_unilist_init (GtkUniList *unilist)
{
  unilist->flags = 0;

  GTK_WIDGET_UNSET_FLAGS (unilist, GTK_NO_WINDOW);
  GTK_WIDGET_SET_FLAGS (unilist, GTK_CAN_FOCUS);
  GTK_UNILIST_SET_FLAG (unilist, UNILIST_DRAW_DRAG_LINE);
  GTK_UNILIST_SET_FLAG (unilist, UNILIST_USE_DRAG_ICONS);

  unilist->row_mem_chunk = NULL;
  unilist->cell_mem_chunk = NULL;

  unilist->freeze_count = 0;

  unilist->rows = 0;
  unilist->row_height = 0;
  unilist->row_list = NULL;
  unilist->row_list_end = NULL;

  unilist->columns = 0;

  unilist->title_window = NULL;
  unilist->column_title_area.x = 0;
  unilist->column_title_area.y = 0;
  unilist->column_title_area.width = 1;
  unilist->column_title_area.height = 1;

  unilist->unilist_window = NULL;
  unilist->unilist_window_width = 1;
  unilist->unilist_window_height = 1;

  unilist->hoffset = 0;
  unilist->voffset = 0;

  unilist->shadow_type = GTK_SHADOW_IN;
  unilist->vadjustment = NULL;
  unilist->hadjustment = NULL;

  unilist->button_actions[0] = GTK_BUTTON_SELECTS | GTK_BUTTON_DRAGS;
  unilist->button_actions[1] = GTK_BUTTON_IGNORED;
  unilist->button_actions[2] = GTK_BUTTON_IGNORED;
  unilist->button_actions[3] = GTK_BUTTON_IGNORED;
  unilist->button_actions[4] = GTK_BUTTON_IGNORED;

  unilist->cursor_drag = NULL;
  unilist->xor_gc = NULL;
  unilist->fg_gc = NULL;
  unilist->bg_gc = NULL;
  unilist->x_drag = 0;

  unilist->selection_mode = GTK_SELECTION_SINGLE;
  unilist->selection = NULL;
  unilist->selection_end = NULL;
  unilist->undo_selection = NULL;
  unilist->undo_unselection = NULL;

  unilist->focus_row = -1;
  unilist->focus_header_column = -1;
  unilist->undo_anchor = -1;

  unilist->anchor = -1;
  unilist->anchor_state = GTK_STATE_SELECTED;
  unilist->drag_pos = -1;
  unilist->htimer = 0;
  unilist->vtimer = 0;

  unilist->click_cell.row = -1;
  unilist->click_cell.column = -1;

  unilist->compare = default_compare;
  unilist->sort_type = GTK_SORT_ASCENDING;
  unilist->sort_column = 0;

  unilist->drag_highlight_row = -1;
}

/* Constructor */
static GObject*
gtk_unilist_constructor (GType                  type,
		       guint                  n_construct_properties,
		       GObjectConstructParam *construct_properties)
{
  GObject *object = G_OBJECT_CLASS (parent_class)->constructor (type,
								n_construct_properties,
								construct_properties);
  GtkUniList *unilist = GTK_UNILIST (object);
  
  /* initalize memory chunks, if this has not been done by any
   * possibly derived widget
   */
  if (!unilist->row_mem_chunk)
    unilist->row_mem_chunk = g_mem_chunk_new ("unilist row mem chunk",
					    sizeof (GtkUniListRow),
					    sizeof (GtkUniListRow) *
					    UNILIST_OPTIMUM_SIZE, 
					    G_ALLOC_AND_FREE);
  
  if (!unilist->cell_mem_chunk)
    unilist->cell_mem_chunk = g_mem_chunk_new ("unilist cell mem chunk",
					     sizeof (GtkCell) * unilist->columns,
					     sizeof (GtkCell) * unilist->columns *
					     UNILIST_OPTIMUM_SIZE, 
					     G_ALLOC_AND_FREE);
  
  /* allocate memory for columns */
  unilist->column = columns_new (unilist);
  
  /* there needs to be at least one column button 
   * because there is alot of code that will break if it
   * isn't there
   */
  column_button_create (unilist, 0);
  
  return object;
}

/* GTKUNILIST PUBLIC INTERFACE
 *   gtk_unilist_new
 *   gtk_unilist_new_with_titles
 *   gtk_unilist_set_hadjustment
 *   gtk_unilist_set_vadjustment
 *   gtk_unilist_get_hadjustment
 *   gtk_unilist_get_vadjustment
 *   gtk_unilist_set_shadow_type
 *   gtk_unilist_set_selection_mode
 *   gtk_unilist_freeze
 *   gtk_unilist_thaw
 */
GtkWidget*
gtk_unilist_new (gint columns)
{
  return gtk_unilist_new_with_titles (columns, NULL);
}
 
GtkWidget*
gtk_unilist_new_with_titles (gint   columns,
			   gchar *titles[])
{
  GtkUniList *unilist;

  unilist = g_object_new (GTK_TYPE_UNILIST,
			"n_columns", columns,
			NULL);
  if (titles)
    {
      guint i;

      for (i = 0; i < unilist->columns; i++)
	gtk_unilist_set_column_title (unilist, i, titles[i]);
      gtk_unilist_column_titles_show (unilist);
    }
  else
    gtk_unilist_column_titles_hide (unilist);
  
  return GTK_WIDGET (unilist);
}

void
gtk_unilist_set_hadjustment (GtkUniList      *unilist,
			   GtkAdjustment *adjustment)
{
  GtkAdjustment *old_adjustment;

  g_return_if_fail (GTK_IS_UNILIST (unilist));
  if (adjustment)
    g_return_if_fail (GTK_IS_ADJUSTMENT (adjustment));
  
  if (unilist->hadjustment == adjustment)
    return;
  
  old_adjustment = unilist->hadjustment;

  if (unilist->hadjustment)
    {
      gtk_signal_disconnect_by_data (GTK_OBJECT (unilist->hadjustment), unilist);
      gtk_object_unref (GTK_OBJECT (unilist->hadjustment));
    }

  unilist->hadjustment = adjustment;

  if (unilist->hadjustment)
    {
      gtk_object_ref (GTK_OBJECT (unilist->hadjustment));
      gtk_object_sink (GTK_OBJECT (unilist->hadjustment));

      gtk_signal_connect (GTK_OBJECT (unilist->hadjustment), "changed",
			  (GtkSignalFunc) hadjustment_changed,
			  (gpointer) unilist);
      gtk_signal_connect (GTK_OBJECT (unilist->hadjustment), "value_changed",
			  (GtkSignalFunc) hadjustment_value_changed,
			  (gpointer) unilist);
    }

  if (!unilist->hadjustment || !old_adjustment)
    gtk_widget_queue_resize (GTK_WIDGET (unilist));
}

GtkAdjustment *
gtk_unilist_get_hadjustment (GtkUniList *unilist)
{
  g_return_val_if_fail (GTK_IS_UNILIST (unilist), NULL);

  return unilist->hadjustment;
}

void
gtk_unilist_set_vadjustment (GtkUniList      *unilist,
			   GtkAdjustment *adjustment)
{
  GtkAdjustment *old_adjustment;

  g_return_if_fail (GTK_IS_UNILIST (unilist));
  if (adjustment)
    g_return_if_fail (GTK_IS_ADJUSTMENT (adjustment));

  if (unilist->vadjustment == adjustment)
    return;
  
  old_adjustment = unilist->vadjustment;

  if (unilist->vadjustment)
    {
      gtk_signal_disconnect_by_data (GTK_OBJECT (unilist->vadjustment), unilist);
      gtk_object_unref (GTK_OBJECT (unilist->vadjustment));
    }

  unilist->vadjustment = adjustment;

  if (unilist->vadjustment)
    {
      gtk_object_ref (GTK_OBJECT (unilist->vadjustment));
      gtk_object_sink (GTK_OBJECT (unilist->vadjustment));

      gtk_signal_connect (GTK_OBJECT (unilist->vadjustment), "changed",
			  (GtkSignalFunc) vadjustment_changed,
			  (gpointer) unilist);
      gtk_signal_connect (GTK_OBJECT (unilist->vadjustment), "value_changed",
			  (GtkSignalFunc) vadjustment_value_changed,
			  (gpointer) unilist);
    }

  if (!unilist->vadjustment || !old_adjustment)
    gtk_widget_queue_resize (GTK_WIDGET (unilist));
}

GtkAdjustment *
gtk_unilist_get_vadjustment (GtkUniList *unilist)
{
  g_return_val_if_fail (GTK_IS_UNILIST (unilist), NULL);

  return unilist->vadjustment;
}

static void
gtk_unilist_set_scroll_adjustments (GtkUniList      *unilist,
				  GtkAdjustment *hadjustment,
				  GtkAdjustment *vadjustment)
{
  if (unilist->hadjustment != hadjustment)
    gtk_unilist_set_hadjustment (unilist, hadjustment);
  if (unilist->vadjustment != vadjustment)
    gtk_unilist_set_vadjustment (unilist, vadjustment);
}

void
gtk_unilist_set_shadow_type (GtkUniList      *unilist,
			   GtkShadowType  type)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  unilist->shadow_type = type;

  if (GTK_WIDGET_VISIBLE (unilist))
    gtk_widget_queue_resize (GTK_WIDGET (unilist));
}

void
gtk_unilist_set_selection_mode (GtkUniList         *unilist,
			      GtkSelectionMode  mode)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));
  g_return_if_fail (mode != GTK_SELECTION_NONE);

  if (mode == unilist->selection_mode)
    return;

  unilist->selection_mode = mode;
  unilist->anchor = -1;
  unilist->anchor_state = GTK_STATE_SELECTED;
  unilist->drag_pos = -1;
  unilist->undo_anchor = unilist->focus_row;

  g_list_free (unilist->undo_selection);
  g_list_free (unilist->undo_unselection);
  unilist->undo_selection = NULL;
  unilist->undo_unselection = NULL;

  switch (mode)
    {
    case GTK_SELECTION_MULTIPLE:
      return;
    case GTK_SELECTION_BROWSE:
    case GTK_SELECTION_SINGLE:
      gtk_unilist_unselect_all (unilist);
      break;
    default:
      /* Someone set it by hand */
      g_assert_not_reached ();
    }
}

void
gtk_unilist_freeze (GtkUniList *unilist)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  unilist->freeze_count++;
}

void
gtk_unilist_thaw (GtkUniList *unilist)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (unilist->freeze_count)
    {
      unilist->freeze_count--;
      UNILIST_REFRESH (unilist);
      gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[SELECTION_CHANGED], NULL);
    }
}

/* PUBLIC COLUMN FUNCTIONS
 *   gtk_unilist_column_titles_show
 *   gtk_unilist_column_titles_hide
 *   gtk_unilist_column_title_active
 *   gtk_unilist_column_title_passive
 *   gtk_unilist_column_titles_active
 *   gtk_unilist_column_titles_passive
 *   gtk_unilist_set_column_title
 *   gtk_unilist_get_column_title
 *   gtk_unilist_set_column_widget
 *   gtk_unilist_set_column_justification
 *   gtk_unilist_set_column_visibility
 *   gtk_unilist_set_column_resizeable
 *   gtk_unilist_set_column_auto_resize
 *   gtk_unilist_optimal_column_width
 *   gtk_unilist_set_column_width
 *   gtk_unilist_set_column_min_width
 *   gtk_unilist_set_column_max_width
 */
void
gtk_unilist_column_titles_show (GtkUniList *unilist)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (!GTK_UNILIST_SHOW_TITLES(unilist))
    {
      GTK_UNILIST_SET_FLAG (unilist, UNILIST_SHOW_TITLES);
      if (unilist->title_window)
	gdk_window_show (unilist->title_window);
      gtk_widget_queue_resize (GTK_WIDGET (unilist));
    }
}

void 
gtk_unilist_column_titles_hide (GtkUniList *unilist)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (GTK_UNILIST_SHOW_TITLES(unilist))
    {
      GTK_UNILIST_UNSET_FLAG (unilist, UNILIST_SHOW_TITLES);
      if (unilist->title_window)
	gdk_window_hide (unilist->title_window);
      gtk_widget_queue_resize (GTK_WIDGET (unilist));
    }
}

void
gtk_unilist_column_title_active (GtkUniList *unilist,
			       gint      column)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (column < 0 || column >= unilist->columns)
    return;
  if (!unilist->column[column].button || !unilist->column[column].button_passive)
    return;

  unilist->column[column].button_passive = FALSE;

  gtk_signal_disconnect_by_func (GTK_OBJECT (unilist->column[column].button),
				 (GtkSignalFunc) column_title_passive_func,
				 NULL);

  GTK_WIDGET_SET_FLAGS (unilist->column[column].button, GTK_CAN_FOCUS);
  if (GTK_WIDGET_VISIBLE (unilist))
    gtk_widget_queue_draw (unilist->column[column].button);
}

void
gtk_unilist_column_title_passive (GtkUniList *unilist,
				gint      column)
{
  GtkButton *button;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (column < 0 || column >= unilist->columns)
    return;
  if (!unilist->column[column].button || unilist->column[column].button_passive)
    return;

  button = GTK_BUTTON (unilist->column[column].button);

  unilist->column[column].button_passive = TRUE;

  if (button->button_down)
    gtk_button_released (button);
  if (button->in_button)
    gtk_button_leave (button);

  gtk_signal_connect (GTK_OBJECT (unilist->column[column].button), "event",
		      (GtkSignalFunc) column_title_passive_func, NULL);

  GTK_WIDGET_UNSET_FLAGS (unilist->column[column].button, GTK_CAN_FOCUS);
  if (GTK_WIDGET_VISIBLE (unilist))
    gtk_widget_queue_draw (unilist->column[column].button);
}

void
gtk_unilist_column_titles_active (GtkUniList *unilist)
{
  gint i;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  for (i = 0; i < unilist->columns; i++)
    gtk_unilist_column_title_active (unilist, i);
}

void
gtk_unilist_column_titles_passive (GtkUniList *unilist)
{
  gint i;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  for (i = 0; i < unilist->columns; i++)
    gtk_unilist_column_title_passive (unilist, i);
}

void
gtk_unilist_set_column_title (GtkUniList    *unilist,
			    gint         column,
			    const gchar *title)
{
  gint new_button = 0;
  GtkWidget *old_widget;
  GtkWidget *alignment = NULL;
  GtkWidget *label;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (column < 0 || column >= unilist->columns)
    return;

  /* if the column button doesn't currently exist,
   * it has to be created first */
  if (!unilist->column[column].button)
    {
      column_button_create (unilist, column);
      new_button = 1;
    }

  column_title_new (unilist, column, title);

  /* remove and destroy the old widget */
  old_widget = GTK_BIN (unilist->column[column].button)->child;
  if (old_widget)
    gtk_container_remove (GTK_CONTAINER (unilist->column[column].button), old_widget);

  /* create new alignment based no column justification */
  switch (unilist->column[column].justification)
    {
    case GTK_JUSTIFY_LEFT:
      alignment = gtk_alignment_new (0.0, 0.5, 0.0, 0.0);
      break;

    case GTK_JUSTIFY_RIGHT:
      alignment = gtk_alignment_new (1.0, 0.5, 0.0, 0.0);
      break;

    case GTK_JUSTIFY_CENTER:
      alignment = gtk_alignment_new (0.5, 0.5, 0.0, 0.0);
      break;

    case GTK_JUSTIFY_FILL:
      alignment = gtk_alignment_new (0.5, 0.5, 0.0, 0.0);
      break;
    }

  gtk_widget_push_composite_child ();
  label = gtk_label_new (unilist->column[column].title);
  gtk_widget_pop_composite_child ();
  gtk_container_add (GTK_CONTAINER (alignment), label);
  gtk_container_add (GTK_CONTAINER (unilist->column[column].button), alignment);
  gtk_widget_show (label);
  gtk_widget_show (alignment);

  /* if this button didn't previously exist, then the
   * column button positions have to be re-computed */
  if (GTK_WIDGET_VISIBLE (unilist) && new_button)
    size_allocate_title_buttons (unilist);
}

gchar *
gtk_unilist_get_column_title (GtkUniList *unilist,
			    gint      column)
{
  g_return_val_if_fail (GTK_IS_UNILIST (unilist), NULL);

  if (column < 0 || column >= unilist->columns)
    return NULL;

  return unilist->column[column].title;
}

void
gtk_unilist_set_column_widget (GtkUniList  *unilist,
			     gint       column,
			     GtkWidget *widget)
{
  gint new_button = 0;
  GtkWidget *old_widget;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (column < 0 || column >= unilist->columns)
    return;

  /* if the column button doesn't currently exist,
   * it has to be created first */
  if (!unilist->column[column].button)
    {
      column_button_create (unilist, column);
      new_button = 1;
    }

  column_title_new (unilist, column, NULL);

  /* remove and destroy the old widget */
  old_widget = GTK_BIN (unilist->column[column].button)->child;
  if (old_widget)
    gtk_container_remove (GTK_CONTAINER (unilist->column[column].button),
			  old_widget);

  /* add and show the widget */
  if (widget)
    {
      gtk_container_add (GTK_CONTAINER (unilist->column[column].button), widget);
      gtk_widget_show (widget);
    }

  /* if this button didn't previously exist, then the
   * column button positions have to be re-computed */
  if (GTK_WIDGET_VISIBLE (unilist) && new_button)
    size_allocate_title_buttons (unilist);
}

GtkWidget *
gtk_unilist_get_column_widget (GtkUniList *unilist,
			     gint      column)
{
  g_return_val_if_fail (GTK_IS_UNILIST (unilist), NULL);

  if (column < 0 || column >= unilist->columns)
    return NULL;

  if (unilist->column[column].button)
    return GTK_BIN (unilist->column[column].button)->child;

  return NULL;
}

void
gtk_unilist_set_column_justification (GtkUniList         *unilist,
				    gint              column,
				    GtkJustification  justification)
{
  GtkWidget *alignment;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (column < 0 || column >= unilist->columns)
    return;

  unilist->column[column].justification = justification;

  /* change the alinment of the button title if it's not a
   * custom widget */
  if (unilist->column[column].title)
    {
      alignment = GTK_BIN (unilist->column[column].button)->child;

      switch (unilist->column[column].justification)
	{
	case GTK_JUSTIFY_LEFT:
	  gtk_alignment_set (GTK_ALIGNMENT (alignment), 0.0, 0.5, 0.0, 0.0);
	  break;

	case GTK_JUSTIFY_RIGHT:
	  gtk_alignment_set (GTK_ALIGNMENT (alignment), 1.0, 0.5, 0.0, 0.0);
	  break;

	case GTK_JUSTIFY_CENTER:
	  gtk_alignment_set (GTK_ALIGNMENT (alignment), 0.5, 0.5, 0.0, 0.0);
	  break;

	case GTK_JUSTIFY_FILL:
	  gtk_alignment_set (GTK_ALIGNMENT (alignment), 0.5, 0.5, 0.0, 0.0);
	  break;

	default:
	  break;
	}
    }

  if (UNILIST_UNFROZEN (unilist))
    draw_rows (unilist, NULL);
}

void
gtk_unilist_set_column_visibility (GtkUniList *unilist,
				 gint      column,
				 gboolean  visible)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (column < 0 || column >= unilist->columns)
    return;
  if (unilist->column[column].visible == visible)
    return;

  /* don't hide last visible column */
  if (!visible)
    {
      gint i;
      gint vis_columns = 0;

      for (i = 0, vis_columns = 0; i < unilist->columns && vis_columns < 2; i++)
	if (unilist->column[i].visible)
	  vis_columns++;

      if (vis_columns < 2)
	return;
    }

  unilist->column[column].visible = visible;

  if (unilist->column[column].button)
    {
      if (visible)
	gtk_widget_show (unilist->column[column].button);
      else
	gtk_widget_hide (unilist->column[column].button);
    }
  
  gtk_widget_queue_resize (GTK_WIDGET(unilist));
}

void
gtk_unilist_set_column_resizeable (GtkUniList *unilist,
				 gint      column,
				 gboolean  resizeable)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (column < 0 || column >= unilist->columns)
    return;
  if (unilist->column[column].resizeable == resizeable)
    return;

  unilist->column[column].resizeable = resizeable;
  if (resizeable)
    unilist->column[column].auto_resize = FALSE;

  if (GTK_WIDGET_VISIBLE (unilist))
    size_allocate_title_buttons (unilist);
}

void
gtk_unilist_set_column_auto_resize (GtkUniList *unilist,
				  gint      column,
				  gboolean  auto_resize)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (column < 0 || column >= unilist->columns)
    return;
  if (unilist->column[column].auto_resize == auto_resize)
    return;

  unilist->column[column].auto_resize = auto_resize;
  if (auto_resize)
    {
      unilist->column[column].resizeable = FALSE;
      if (!GTK_UNILIST_AUTO_RESIZE_BLOCKED (unilist))
	{
	  gint width;

	  width = gtk_unilist_optimal_column_width (unilist, column);
	  gtk_unilist_set_column_width (unilist, column, width);
	}
    }

  if (GTK_WIDGET_VISIBLE (unilist))
    size_allocate_title_buttons (unilist);
}

gint
gtk_unilist_columns_autosize (GtkUniList *unilist)
{
  gint i;
  gint width;

  g_return_val_if_fail (GTK_IS_UNILIST (unilist), 0);

  gtk_unilist_freeze (unilist);
  width = 0;
  for (i = 0; i < unilist->columns; i++)
    {
      gtk_unilist_set_column_width (unilist, i,
				  gtk_unilist_optimal_column_width (unilist, i));

      width += unilist->column[i].width;
    }

  gtk_unilist_thaw (unilist);
  return width;
}

gint
gtk_unilist_optimal_column_width (GtkUniList *unilist,
				gint      column)
{
  GtkRequisition requisition;
  GList *list;
  gint width;

  g_return_val_if_fail (GTK_UNILIST (unilist), 0);

  if (column < 0 || column >= unilist->columns)
    return 0;

  if (GTK_UNILIST_SHOW_TITLES(unilist) && unilist->column[column].button)
    width = (unilist->column[column].button->requisition.width)
#if 0
	     (CELL_SPACING + (2 * COLUMN_INSET)))
#endif
		;
  else
    width = 0;

  for (list = unilist->row_list; list; list = list->next)
    {
  GTK_UNILIST_GET_CLASS (unilist)->cell_size_request
	(unilist, GTK_UNILIST_ROW (list), column, &requisition);
      width = MAX (width, requisition.width);
    }

  return width;
}

void
gtk_unilist_set_column_width (GtkUniList *unilist,
			    gint      column,
			    gint      width)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (column < 0 || column >= unilist->columns)
    return;

  gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[RESIZE_COLUMN],
		   column, width);
}

void
gtk_unilist_set_column_min_width (GtkUniList *unilist,
				gint      column,
				gint      min_width)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (column < 0 || column >= unilist->columns)
    return;
  if (unilist->column[column].min_width == min_width)
    return;

  if (unilist->column[column].max_width >= 0  &&
      unilist->column[column].max_width < min_width)
    unilist->column[column].min_width = unilist->column[column].max_width;
  else
    unilist->column[column].min_width = min_width;

  if (unilist->column[column].area.width < unilist->column[column].min_width)
    gtk_unilist_set_column_width (unilist, column,unilist->column[column].min_width);
}

void
gtk_unilist_set_column_max_width (GtkUniList *unilist,
				gint      column,
				gint      max_width)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (column < 0 || column >= unilist->columns)
    return;
  if (unilist->column[column].max_width == max_width)
    return;

  if (unilist->column[column].min_width >= 0 && max_width >= 0 &&
      unilist->column[column].min_width > max_width)
    unilist->column[column].max_width = unilist->column[column].min_width;
  else
    unilist->column[column].max_width = max_width;
  
  if (unilist->column[column].area.width > unilist->column[column].max_width)
    gtk_unilist_set_column_width (unilist, column,unilist->column[column].max_width);
}

/* PRIVATE COLUMN FUNCTIONS
 *   column_auto_resize
 *   real_resize_column
 *   abort_column_resize
 *   size_allocate_title_buttons
 *   size_allocate_columns
 *   list_requisition_width
 *   new_column_width
 *   column_button_create
 *   column_button_clicked
 *   column_title_passive_func
 */
static void
column_auto_resize (GtkUniList    *unilist,
		    GtkUniListRow *unilist_row,
		    gint         column,
		    gint         old_width)
{
  /* resize column if needed for auto_resize */
  GtkRequisition requisition;

  if (!unilist->column[column].auto_resize ||
      GTK_UNILIST_AUTO_RESIZE_BLOCKED (unilist))
    return;

  if (unilist_row)
    GTK_UNILIST_GET_CLASS (unilist)->cell_size_request (unilist, unilist_row,
						   column, &requisition);
  else
    requisition.width = 0;

  if (requisition.width > unilist->column[column].width)
    gtk_unilist_set_column_width (unilist, column, requisition.width);
  else if (requisition.width < old_width &&
	   old_width == unilist->column[column].width)
    {
      GList *list;
      gint new_width = 0;

      /* run a "gtk_unilist_optimal_column_width" but break, if
       * the column doesn't shrink */
      if (GTK_UNILIST_SHOW_TITLES(unilist) && unilist->column[column].button)
	new_width = (unilist->column[column].button->requisition.width -
		     (CELL_SPACING + (2 * COLUMN_INSET)));
      else
	new_width = 0;

      for (list = unilist->row_list; list; list = list->next)
	{
	  GTK_UNILIST_GET_CLASS (unilist)->cell_size_request
	    (unilist, GTK_UNILIST_ROW (list), column, &requisition);
	  new_width = MAX (new_width, requisition.width);
	  if (new_width == unilist->column[column].width)
	    break;
	}
      if (new_width < unilist->column[column].width)
	gtk_unilist_set_column_width
	  (unilist, column, MAX (new_width, unilist->column[column].min_width));
    }
}

static void
real_resize_column (GtkUniList *unilist,
		    gint      column,
		    gint      width)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (column < 0 || column >= unilist->columns)
    return;
  
  if (width < MAX (COLUMN_MIN_WIDTH, unilist->column[column].min_width))
    width = MAX (COLUMN_MIN_WIDTH, unilist->column[column].min_width);
  if (unilist->column[column].max_width >= 0 &&
      width > unilist->column[column].max_width)
    width = unilist->column[column].max_width;

  unilist->column[column].width = width;
  unilist->column[column].width_set = TRUE;

  /* FIXME: this is quite expensive to do if the widget hasn't
   *        been size_allocated yet, and pointless. Should
   *        a flag be kept
   */
  size_allocate_columns (unilist, TRUE);
  size_allocate_title_buttons (unilist);

  UNILIST_REFRESH (unilist);
}

static void
abort_column_resize (GtkUniList *unilist)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (!GTK_UNILIST_IN_DRAG(unilist))
    return;

  GTK_UNILIST_UNSET_FLAG (unilist, UNILIST_IN_DRAG);
  gtk_grab_remove (GTK_WIDGET (unilist));
  gdk_pointer_ungrab (GDK_CURRENT_TIME);
  unilist->drag_pos = -1;

  if (unilist->x_drag >= 0 && unilist->x_drag <= unilist->unilist_window_width - 1)
    draw_xor_line (unilist);

  if (GTK_UNILIST_ADD_MODE(unilist))
    {
      gdk_gc_set_line_attributes (unilist->xor_gc, 1, GDK_LINE_ON_OFF_DASH, 0,0);
      gdk_gc_set_dashes (unilist->xor_gc, 0, "\4\4", 2);
    }
}

static void
size_allocate_title_buttons (GtkUniList *unilist)
{
  GtkAllocation button_allocation;
  gint last_column;
  gint last_button = 0;
  gint i;

  if (!GTK_WIDGET_REALIZED (unilist))
    return;

  button_allocation.x = unilist->hoffset;
  button_allocation.y = 0;
  button_allocation.width = 0;
  button_allocation.height = unilist->column_title_area.height;

  /* find last visible column */
  for (last_column = unilist->columns - 1; last_column >= 0; last_column--)
    if (unilist->column[last_column].visible)
      break;

  for (i = 0; i < last_column; i++)
    {
      if (!unilist->column[i].visible)
	{
	  last_button = i + 1;
	  gdk_window_hide (unilist->column[i].window);
	  continue;
	}

      button_allocation.width += (unilist->column[i].area.width +
				  CELL_SPACING + 2 * COLUMN_INSET);

      if (!unilist->column[i + 1].button)
	{
	  gdk_window_hide (unilist->column[i].window);
	  continue;
	}

      gtk_widget_size_allocate (unilist->column[last_button].button,
				&button_allocation);
      button_allocation.x += button_allocation.width;
      button_allocation.width = 0;

      if (unilist->column[last_button].resizeable)
	{
	  gdk_window_show (unilist->column[last_button].window);
	  gdk_window_move_resize (unilist->column[last_button].window,
				  button_allocation.x - (DRAG_WIDTH / 2), 
				  0, DRAG_WIDTH,
				  unilist->column_title_area.height);
	}
      else
	gdk_window_hide (unilist->column[last_button].window);

      last_button = i + 1;
    }

  button_allocation.width += (unilist->column[last_column].area.width +
			      2 * (CELL_SPACING + COLUMN_INSET));
  gtk_widget_size_allocate (unilist->column[last_button].button,
			    &button_allocation);

  if (unilist->column[last_button].resizeable)
    {
      button_allocation.x += button_allocation.width;

      gdk_window_show (unilist->column[last_button].window);
      gdk_window_move_resize (unilist->column[last_button].window,
			      button_allocation.x - (DRAG_WIDTH / 2), 
			      0, DRAG_WIDTH, unilist->column_title_area.height);
    }
  else
    gdk_window_hide (unilist->column[last_button].window);
}

static void
size_allocate_columns (GtkUniList *unilist,
		       gboolean  block_resize)
{
  gint xoffset = CELL_SPACING + COLUMN_INSET;
  gint last_column;
  gint i;

  /* find last visible column and calculate correct column width */
  for (last_column = unilist->columns - 1;
       last_column >= 0 && !unilist->column[last_column].visible; last_column--);

  if (last_column < 0)
    return;

  for (i = 0; i <= last_column; i++)
    {
      if (!unilist->column[i].visible)
	continue;
      unilist->column[i].area.x = xoffset;
      if (unilist->column[i].width_set)
	{
	  if (!block_resize && GTK_UNILIST_SHOW_TITLES(unilist) &&
	      unilist->column[i].auto_resize && unilist->column[i].button)
	    {
	      gint width;

	      width = (unilist->column[i].button->requisition.width -
		       (CELL_SPACING + (2 * COLUMN_INSET)));

	      if (width > unilist->column[i].width)
		gtk_unilist_set_column_width (unilist, i, width);
	    }

	  unilist->column[i].area.width = unilist->column[i].width;
	  xoffset += unilist->column[i].width + CELL_SPACING + (2* COLUMN_INSET);
	}
      else if (GTK_UNILIST_SHOW_TITLES(unilist) && unilist->column[i].button)
	{
	  unilist->column[i].area.width =
	    unilist->column[i].button->requisition.width -
	    (CELL_SPACING + (2 * COLUMN_INSET));
	  xoffset += unilist->column[i].button->requisition.width;
	}
    }

  unilist->column[last_column].area.width = unilist->column[last_column].area.width
    + MAX (0, unilist->unilist_window_width + COLUMN_INSET - xoffset);
}

static gint
list_requisition_width (GtkUniList *unilist) 
{
  gint width = CELL_SPACING;
  gint i;

  for (i = unilist->columns - 1; i >= 0; i--)
    {
      if (!unilist->column[i].visible)
	continue;

      if (unilist->column[i].width_set)
	width += unilist->column[i].width + CELL_SPACING + (2 * COLUMN_INSET);
      else if (GTK_UNILIST_SHOW_TITLES(unilist) && unilist->column[i].button)
	width += unilist->column[i].button->requisition.width;
    }

  return width;
}

/* this function returns the new width of the column being resized given
 * the column and x position of the cursor; the x cursor position is passed
 * in as a pointer and automagicly corrected if it's beyond min/max limits */
static gint
new_column_width (GtkUniList *unilist,
		  gint      column,
		  gint     *x)
{
  gint xthickness = GTK_WIDGET (unilist)->style->xthickness;
  gint width;
  gint cx;
  gint dx;
  gint last_column;

  /* first translate the x position from widget->window
   * to unilist->unilist_window */
  cx = *x - xthickness;

  for (last_column = unilist->columns - 1;
       last_column >= 0 && !unilist->column[last_column].visible; last_column--);

  /* calculate new column width making sure it doesn't end up
   * less than the minimum width */
  dx = (COLUMN_LEFT_XPIXEL (unilist, column) + COLUMN_INSET +
	(column < last_column) * CELL_SPACING);
  width = cx - dx;

  if (width < MAX (COLUMN_MIN_WIDTH, unilist->column[column].min_width))
    {
      width = MAX (COLUMN_MIN_WIDTH, unilist->column[column].min_width);
      cx = dx + width;
      *x = cx + xthickness;
    }
  else if (unilist->column[column].max_width >= COLUMN_MIN_WIDTH &&
	   width > unilist->column[column].max_width)
    {
      width = unilist->column[column].max_width;
      cx = dx + unilist->column[column].max_width;
      *x = cx + xthickness;
    }      

  if (cx < 0 || cx > unilist->unilist_window_width)
    *x = -1;

  return width;
}

static void
column_button_create (GtkUniList *unilist,
		      gint      column)
{
  GtkWidget *button;

  gtk_widget_push_composite_child ();
  button = unilist->column[column].button = gtk_button_new ();
  gtk_widget_pop_composite_child ();

  if (GTK_WIDGET_REALIZED (unilist) && unilist->title_window)
    gtk_widget_set_parent_window (unilist->column[column].button,
				  unilist->title_window);
  gtk_widget_set_parent (button, GTK_WIDGET (unilist));

  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      (GtkSignalFunc) column_button_clicked,
		      (gpointer) unilist);
  gtk_widget_show (button);
}

static void
column_button_clicked (GtkWidget *widget,
		       gpointer   data)
{
  gint i;
  GtkUniList *unilist;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_UNILIST (data));

  unilist = GTK_UNILIST (data);

  /* find the column who's button was pressed */
  for (i = 0; i < unilist->columns; i++)
    if (unilist->column[i].button == widget)
      break;

  gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[CLICK_COLUMN], i);
}

static gint
column_title_passive_func (GtkWidget *widget, 
			   GdkEvent  *event,
			   gpointer   data)
{
  g_return_val_if_fail (event != NULL, FALSE);
  
  switch (event->type)
    {
    case GDK_MOTION_NOTIFY:
    case GDK_BUTTON_PRESS:
    case GDK_2BUTTON_PRESS:
    case GDK_3BUTTON_PRESS:
    case GDK_BUTTON_RELEASE:
    case GDK_ENTER_NOTIFY:
    case GDK_LEAVE_NOTIFY:
      return TRUE;
    default:
      break;
    }
  return FALSE;
}


/* PUBLIC CELL FUNCTIONS
 *   gtk_unilist_get_cell_type
 *   gtk_unilist_set_text
 *   gtk_unilist_get_text
 *   gtk_unilist_set_pixmap
 *   gtk_unilist_get_pixmap
 *   gtk_unilist_set_pixtext
 *   gtk_unilist_get_pixtext
 *   gtk_unilist_set_shift
 */
GtkCellType 
gtk_unilist_get_cell_type (GtkUniList *unilist,
			 gint      row,
			 gint      column)
{
  GtkUniListRow *unilist_row;

  g_return_val_if_fail (GTK_IS_UNILIST (unilist), -1);

  if (row < 0 || row >= unilist->rows)
    return -1;
  if (column < 0 || column >= unilist->columns)
    return -1;

  unilist_row = ROW_ELEMENT (unilist, row)->data;

  return unilist_row->cell[column].type;
}

void
gtk_unilist_set_text (GtkUniList    *unilist,
		    gint         row,
		    gint         column,
		    const gchar *text)
{
  GtkUniListRow *unilist_row;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (row < 0 || row >= unilist->rows)
    return;
  if (column < 0 || column >= unilist->columns)
    return;

  unilist_row = ROW_ELEMENT (unilist, row)->data;

  /* if text is null, then the cell is empty */
  GTK_UNILIST_GET_CLASS (unilist)->set_cell_contents
    (unilist, unilist_row, column, GTK_CELL_TEXT, text, 0, NULL, NULL);

  /* redraw the list if it's not frozen */
  if (UNILIST_UNFROZEN (unilist))
    {
      if (gtk_unilist_row_is_visible (unilist, row) != GTK_VISIBILITY_NONE)
	GTK_UNILIST_GET_CLASS (unilist)->draw_row (unilist, NULL, row, unilist_row);
    }
}

gint
gtk_unilist_get_text (GtkUniList  *unilist,
		    gint       row,
		    gint       column,
		    gchar    **text)
{
  GtkUniListRow *unilist_row;

  g_return_val_if_fail (GTK_IS_UNILIST (unilist), 0);

  if (row < 0 || row >= unilist->rows)
    return 0;
  if (column < 0 || column >= unilist->columns)
    return 0;

  unilist_row = ROW_ELEMENT (unilist, row)->data;

  if (unilist_row->cell[column].type != GTK_CELL_TEXT)
    return 0;

  if (text)
    *text = GTK_CELL_TEXT (unilist_row->cell[column])->text;

  return 1;
}

void
gtk_unilist_set_pixmap (GtkUniList  *unilist,
		      gint       row,
		      gint       column,
		      GdkPixmap *pixmap,
		      GdkBitmap *mask)
{
  GtkUniListRow *unilist_row;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (row < 0 || row >= unilist->rows)
    return;
  if (column < 0 || column >= unilist->columns)
    return;

  unilist_row = ROW_ELEMENT (unilist, row)->data;
  
  gdk_pixmap_ref (pixmap);
  
  if (mask) gdk_pixmap_ref (mask);
  
  GTK_UNILIST_GET_CLASS (unilist)->set_cell_contents
    (unilist, unilist_row, column, GTK_CELL_PIXMAP, NULL, 0, pixmap, mask);

  /* redraw the list if it's not frozen */
  if (UNILIST_UNFROZEN (unilist))
    {
      if (gtk_unilist_row_is_visible (unilist, row) != GTK_VISIBILITY_NONE)
	GTK_UNILIST_GET_CLASS (unilist)->draw_row (unilist, NULL, row, unilist_row);
    }
}

gint
gtk_unilist_get_pixmap (GtkUniList   *unilist,
		      gint        row,
		      gint        column,
		      GdkPixmap **pixmap,
		      GdkBitmap **mask)
{
  GtkUniListRow *unilist_row;

  g_return_val_if_fail (GTK_IS_UNILIST (unilist), 0);

  if (row < 0 || row >= unilist->rows)
    return 0;
  if (column < 0 || column >= unilist->columns)
    return 0;

  unilist_row = ROW_ELEMENT (unilist, row)->data;

  if (unilist_row->cell[column].type != GTK_CELL_PIXMAP)
    return 0;

  if (pixmap)
  {
    *pixmap = GTK_CELL_PIXMAP (unilist_row->cell[column])->pixmap;
    /* mask can be NULL */
    *mask = GTK_CELL_PIXMAP (unilist_row->cell[column])->mask;
  }

  return 1;
}

void
gtk_unilist_set_pixtext (GtkUniList    *unilist,
		       gint         row,
		       gint         column,
		       const gchar *text,
		       guint8       spacing,
		       GdkPixmap   *pixmap,
		       GdkBitmap   *mask)
{
  GtkUniListRow *unilist_row;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (row < 0 || row >= unilist->rows)
    return;
  if (column < 0 || column >= unilist->columns)
    return;

  unilist_row = ROW_ELEMENT (unilist, row)->data;
  
  gdk_pixmap_ref (pixmap);
  if (mask) gdk_pixmap_ref (mask);
  GTK_UNILIST_GET_CLASS (unilist)->set_cell_contents
    (unilist, unilist_row, column, GTK_CELL_PIXTEXT, text, spacing, pixmap, mask);

  /* redraw the list if it's not frozen */
  if (UNILIST_UNFROZEN (unilist))
    {
      if (gtk_unilist_row_is_visible (unilist, row) != GTK_VISIBILITY_NONE)
	GTK_UNILIST_GET_CLASS (unilist)->draw_row (unilist, NULL, row, unilist_row);
    }
}

gint
gtk_unilist_get_pixtext (GtkUniList   *unilist,
		       gint        row,
		       gint        column,
		       gchar     **text,
		       guint8     *spacing,
		       GdkPixmap **pixmap,
		       GdkBitmap **mask)
{
  GtkUniListRow *unilist_row;

  g_return_val_if_fail (GTK_IS_UNILIST (unilist), 0);

  if (row < 0 || row >= unilist->rows)
    return 0;
  if (column < 0 || column >= unilist->columns)
    return 0;

  unilist_row = ROW_ELEMENT (unilist, row)->data;

  if (unilist_row->cell[column].type != GTK_CELL_PIXTEXT)
    return 0;

  if (text)
    *text = GTK_CELL_PIXTEXT (unilist_row->cell[column])->text;
  if (spacing)
    *spacing = GTK_CELL_PIXTEXT (unilist_row->cell[column])->spacing;
  if (pixmap)
    *pixmap = GTK_CELL_PIXTEXT (unilist_row->cell[column])->pixmap;

  /* mask can be NULL */
  if (mask)
    *mask = GTK_CELL_PIXTEXT (unilist_row->cell[column])->mask;

  return 1;
}

void
gtk_unilist_set_shift (GtkUniList *unilist,
		     gint      row,
		     gint      column,
		     gint      vertical,
		     gint      horizontal)
{
  GtkRequisition requisition = { 0 };
  GtkUniListRow *unilist_row;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (row < 0 || row >= unilist->rows)
    return;
  if (column < 0 || column >= unilist->columns)
    return;

  unilist_row = ROW_ELEMENT (unilist, row)->data;

  if (unilist->column[column].auto_resize &&
      !GTK_UNILIST_AUTO_RESIZE_BLOCKED (unilist))
    GTK_UNILIST_GET_CLASS (unilist)->cell_size_request (unilist, unilist_row,
						   column, &requisition);

  unilist_row->cell[column].vertical = vertical;
  unilist_row->cell[column].horizontal = horizontal;

  column_auto_resize (unilist, unilist_row, column, requisition.width);

  if (UNILIST_UNFROZEN (unilist) && gtk_unilist_row_is_visible (unilist, row) != GTK_VISIBILITY_NONE)
    GTK_UNILIST_GET_CLASS (unilist)->draw_row (unilist, NULL, row, unilist_row);
}

/* PRIVATE CELL FUNCTIONS
 *   set_cell_contents
 *   cell_size_request
 */
static void
set_cell_contents (GtkUniList    *unilist,
		   GtkUniListRow *unilist_row,
		   gint         column,
		   GtkCellType  type,
		   const gchar *text,
		   guint8       spacing,
		   GdkPixmap   *pixmap,
		   GdkBitmap   *mask)
{
  GtkRequisition requisition;
  gchar *old_text = NULL;
  GdkPixmap *old_pixmap = NULL;
  GdkBitmap *old_mask = NULL;
  
  g_return_if_fail (GTK_IS_UNILIST (unilist));
  g_return_if_fail (unilist_row != NULL);

  if (unilist->column[column].auto_resize &&
      !GTK_UNILIST_AUTO_RESIZE_BLOCKED (unilist))
    GTK_UNILIST_GET_CLASS (unilist)->cell_size_request (unilist, unilist_row,
						   column, &requisition);

  switch (unilist_row->cell[column].type)
    {
    case GTK_CELL_EMPTY:
      break;
    case GTK_CELL_TEXT:
      old_text = GTK_CELL_TEXT (unilist_row->cell[column])->text;
      break;
    case GTK_CELL_PIXMAP:
      old_pixmap = GTK_CELL_PIXMAP (unilist_row->cell[column])->pixmap;
      old_mask = GTK_CELL_PIXMAP (unilist_row->cell[column])->mask;
      break;
    case GTK_CELL_PIXTEXT:
      old_text = GTK_CELL_PIXTEXT (unilist_row->cell[column])->text;
      old_pixmap = GTK_CELL_PIXTEXT (unilist_row->cell[column])->pixmap;
      old_mask = GTK_CELL_PIXTEXT (unilist_row->cell[column])->mask;
      break;
    case GTK_CELL_WIDGET:
      /* unimplimented */
      break;
    default:
      break;
    }

  unilist_row->cell[column].type = GTK_CELL_EMPTY;

  /* Note that pixmap and mask were already ref'ed by the caller
   */
  switch (type)
    {
    case GTK_CELL_TEXT:
      if (text)
	{
	  unilist_row->cell[column].type = GTK_CELL_TEXT;
	  GTK_CELL_TEXT (unilist_row->cell[column])->text = g_strdup (text);
	}
      break;
    case GTK_CELL_PIXMAP:
      if (pixmap)
	{
	  unilist_row->cell[column].type = GTK_CELL_PIXMAP;
	  GTK_CELL_PIXMAP (unilist_row->cell[column])->pixmap = pixmap;
	  /* We set the mask even if it is NULL */
	  GTK_CELL_PIXMAP (unilist_row->cell[column])->mask = mask;
	}
      break;
    case GTK_CELL_PIXTEXT:
      if (text && pixmap)
	{
	  unilist_row->cell[column].type = GTK_CELL_PIXTEXT;
	  GTK_CELL_PIXTEXT (unilist_row->cell[column])->text = g_strdup (text);
	  GTK_CELL_PIXTEXT (unilist_row->cell[column])->spacing = spacing;
	  GTK_CELL_PIXTEXT (unilist_row->cell[column])->pixmap = pixmap;
	  GTK_CELL_PIXTEXT (unilist_row->cell[column])->mask = mask;
	}
      break;
    default:
      break;
    }

  if (unilist->column[column].auto_resize &&
      !GTK_UNILIST_AUTO_RESIZE_BLOCKED (unilist))
    column_auto_resize (unilist, unilist_row, column, requisition.width);

  if (old_text)
    g_free (old_text);
  if (old_pixmap)
    gdk_pixmap_unref (old_pixmap);
  if (old_mask)
    gdk_pixmap_unref (old_mask);
}

PangoLayout *
_gtk_unilist_create_cell_layout (GtkUniList       *unilist,
			       GtkUniListRow    *unilist_row,
			       gint            column)
{
  PangoLayout *layout;
  GtkStyle *style;
  GtkCell *cell;
  gchar *text;
  
  get_cell_style (unilist, unilist_row, GTK_STATE_NORMAL, column, &style,
		  NULL, NULL);


  cell = &unilist_row->cell[column];
  switch (cell->type)
    {
    case GTK_CELL_TEXT:
    case GTK_CELL_PIXTEXT:
      text = ((cell->type == GTK_CELL_PIXTEXT) ?
	      GTK_CELL_PIXTEXT (*cell)->text :
	      GTK_CELL_TEXT (*cell)->text);

      if (!text)
	return NULL;
      
      layout = gtk_widget_create_pango_layout (GTK_WIDGET (unilist),
					       ((cell->type == GTK_CELL_PIXTEXT) ?
						GTK_CELL_PIXTEXT (*cell)->text :
						GTK_CELL_TEXT (*cell)->text));
      pango_layout_set_font_description (layout, style->font_desc);
      
      return layout;
      
    default:
      return NULL;
    }
}

static void
cell_size_request (GtkUniList       *unilist,
		   GtkUniListRow    *unilist_row,
		   gint            column,
		   GtkRequisition *requisition)
{
  gint width;
  gint height;
  PangoLayout *layout;
  PangoRectangle logical_rect;

  g_return_if_fail (GTK_IS_UNILIST (unilist));
  g_return_if_fail (requisition != NULL);

  layout = _gtk_unilist_create_cell_layout (unilist, unilist_row, column);
  if (layout)
    {
      pango_layout_get_pixel_extents (layout, NULL, &logical_rect);
      
      requisition->width = logical_rect.width;
      requisition->height = logical_rect.height;
      
      g_object_unref (G_OBJECT (layout));
    }
  else
    {
      requisition->width  = 0;
      requisition->height = 0;
    }

  if (layout && unilist_row->cell[column].type == GTK_CELL_PIXTEXT)
    requisition->width += GTK_CELL_PIXTEXT (unilist_row->cell[column])->spacing;

  switch (unilist_row->cell[column].type)
    {
    case GTK_CELL_PIXTEXT:
      gdk_window_get_size (GTK_CELL_PIXTEXT (unilist_row->cell[column])->pixmap,
			   &width, &height);
      requisition->width += width;
      requisition->height = MAX (requisition->height, height);      
      break;
    case GTK_CELL_PIXMAP:
      gdk_window_get_size (GTK_CELL_PIXMAP (unilist_row->cell[column])->pixmap,
			   &width, &height);
      requisition->width += width;
      requisition->height = MAX (requisition->height, height);
      break;
      
    default:
      break;
    }

  requisition->width  += unilist_row->cell[column].horizontal;
  requisition->height += unilist_row->cell[column].vertical;
}

/* PUBLIC INSERT/REMOVE ROW FUNCTIONS
 *   gtk_unilist_prepend
 *   gtk_unilist_append
 *   gtk_unilist_insert
 *   gtk_unilist_remove
 *   gtk_unilist_clear
 */
gint
gtk_unilist_prepend (GtkUniList    *unilist,
		   gchar       *text[])
{
  g_return_val_if_fail (GTK_IS_UNILIST (unilist), -1);
  g_return_val_if_fail (text != NULL, -1);

  return GTK_UNILIST_GET_CLASS (unilist)->insert_row (unilist, 0, text);
}

gint
gtk_unilist_append (GtkUniList    *unilist,
		  gchar       *text[])
{
  g_return_val_if_fail (GTK_IS_UNILIST (unilist), -1);
  g_return_val_if_fail (text != NULL, -1);

  return GTK_UNILIST_GET_CLASS (unilist)->insert_row (unilist, unilist->rows, text);
}

gint
gtk_unilist_insert (GtkUniList    *unilist,
		  gint         row,
		  gchar       *text[])
{
  g_return_val_if_fail (GTK_IS_UNILIST (unilist), -1);
  g_return_val_if_fail (text != NULL, -1);

  if (row < 0 || row > unilist->rows)
    row = unilist->rows;

  return GTK_UNILIST_GET_CLASS (unilist)->insert_row (unilist, row, text);
}

void
gtk_unilist_remove (GtkUniList *unilist,
		  gint      row)
{
  GTK_UNILIST_GET_CLASS (unilist)->remove_row (unilist, row);
}

void
gtk_unilist_clear (GtkUniList *unilist)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));
  
  GTK_UNILIST_GET_CLASS (unilist)->clear (unilist);
}

/* PRIVATE INSERT/REMOVE ROW FUNCTIONS
 *   real_insert_row
 *   real_remove_row
 *   real_clear
 *   real_row_move
 */
static gint
real_insert_row (GtkUniList *unilist,
		 gint      row,
		 gchar    *text[])
{
  gint i;
  GtkUniListRow *unilist_row;

  g_return_val_if_fail (GTK_IS_UNILIST (unilist), -1);
  g_return_val_if_fail (text != NULL, -1);

  /* return if out of bounds */
  if (row < 0 || row > unilist->rows)
    return -1;

  /* create the row */
  unilist_row = row_new (unilist);

  /* set the text in the row's columns */
  for (i = 0; i < unilist->columns; i++)
    if (text[i])
      GTK_UNILIST_GET_CLASS (unilist)->set_cell_contents
	(unilist, unilist_row, i, GTK_CELL_TEXT, text[i], 0, NULL ,NULL);

  if (!unilist->rows)
    {
      unilist->row_list = g_list_append (unilist->row_list, unilist_row);
      unilist->row_list_end = unilist->row_list;
    }
  else
    {
      if (GTK_UNILIST_AUTO_SORT(unilist))   /* override insertion pos */
	{
	  GList *work;
	  
	  row = 0;
	  work = unilist->row_list;
	  
	  if (unilist->sort_type == GTK_SORT_ASCENDING)
	    {
	      while (row < unilist->rows &&
		     unilist->compare (unilist, unilist_row,
				     GTK_UNILIST_ROW (work)) > 0)
		{
		  row++;
		  work = work->next;
		}
	    }
	  else
	    {
	      while (row < unilist->rows &&
		     unilist->compare (unilist, unilist_row,
				     GTK_UNILIST_ROW (work)) < 0)
		{
		  row++;
		  work = work->next;
		}
	    }
	}
      
      /* reset the row end pointer if we're inserting at the end of the list */
      if (row == unilist->rows)
	unilist->row_list_end = (g_list_append (unilist->row_list_end,
					      unilist_row))->next;
      else
	unilist->row_list = g_list_insert (unilist->row_list, unilist_row, row);

    }
  unilist->rows++;

  if (row < ROW_FROM_YPIXEL (unilist, 0))
    unilist->voffset -= (unilist->row_height + CELL_SPACING);

  /* syncronize the selection list */
  sync_selection (unilist, row, SYNC_INSERT);

  if (unilist->rows == 1)
    {
      unilist->focus_row = 0;
      if (unilist->selection_mode == GTK_SELECTION_BROWSE)
	gtk_unilist_select_row (unilist, 0, -1);
    }

  /* redraw the list if it isn't frozen */
  if (UNILIST_UNFROZEN (unilist))
    {
      adjust_adjustments (unilist, FALSE);

      if (gtk_unilist_row_is_visible (unilist, row) != GTK_VISIBILITY_NONE)
	draw_rows (unilist, NULL);
    }

  return row;
}

static void
real_remove_row (GtkUniList *unilist,
		 gint      row)
{
  gint was_visible, was_selected;
  GList *list;
  GtkUniListRow *unilist_row;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  /* return if out of bounds */
  if (row < 0 || row > (unilist->rows - 1))
    return;

  was_visible = (gtk_unilist_row_is_visible (unilist, row) != GTK_VISIBILITY_NONE);
  was_selected = 0;

  /* get the row we're going to delete */
  list = ROW_ELEMENT (unilist, row);
  g_assert (list != NULL);
  unilist_row = list->data;

  /* if we're removing a selected row, we have to make sure
   * it's properly unselected, and then sync up the unilist->selected
   * list to reflect the deincrimented indexies of rows after the
   * removal */
  if (unilist_row->state == GTK_STATE_SELECTED)
    gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[UNSELECT_ROW],
		     row, -1, NULL);

  sync_selection (unilist, row, SYNC_REMOVE);

  /* reset the row end pointer if we're removing at the end of the list */
  unilist->rows--;
  if (unilist->row_list == list)
    unilist->row_list = g_list_next (list);
  if (unilist->row_list_end == list)
    unilist->row_list_end = g_list_previous (list);
  g_list_remove (list, unilist_row);

  /*if (unilist->focus_row >=0 &&
      (row <= unilist->focus_row || unilist->focus_row >= unilist->rows))
      unilist->focus_row--;*/

  if (row < ROW_FROM_YPIXEL (unilist, 0))
    unilist->voffset += unilist->row_height + CELL_SPACING;

  if (unilist->selection_mode == GTK_SELECTION_BROWSE && !unilist->selection &&
      unilist->focus_row >= 0)
    gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[SELECT_ROW],
		     unilist->focus_row, -1, NULL);

  /* toast the row */
  row_delete (unilist, unilist_row);

  /* redraw the row if it isn't frozen */
  if (UNILIST_UNFROZEN (unilist))
    {
      adjust_adjustments (unilist, FALSE);

      if (was_visible)
	draw_rows (unilist, NULL);
    }
}

static void
real_clear (GtkUniList *unilist)
{
  GList *list;
  GList *free_list;
  gint i;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  /* free up the selection list */
  g_list_free (unilist->selection);
  g_list_free (unilist->undo_selection);
  g_list_free (unilist->undo_unselection);

  unilist->selection = NULL;
  unilist->selection_end = NULL;
  unilist->undo_selection = NULL;
  unilist->undo_unselection = NULL;
  unilist->voffset = 0;
  unilist->focus_row = -1;
  unilist->anchor = -1;
  unilist->undo_anchor = -1;
  unilist->anchor_state = GTK_STATE_SELECTED;
  unilist->drag_pos = -1;

  /* remove all the rows */
  GTK_UNILIST_SET_FLAG (unilist, UNILIST_AUTO_RESIZE_BLOCKED);
  free_list = unilist->row_list;
  unilist->row_list = NULL;
  unilist->row_list_end = NULL;
  unilist->rows = 0;
  for (list = free_list; list; list = list->next)
    row_delete (unilist, GTK_UNILIST_ROW (list));
  g_list_free (free_list);
  GTK_UNILIST_UNSET_FLAG (unilist, UNILIST_AUTO_RESIZE_BLOCKED);
  for (i = 0; i < unilist->columns; i++)
    if (unilist->column[i].auto_resize)
      {
	if (GTK_UNILIST_SHOW_TITLES(unilist) && unilist->column[i].button)
	  gtk_unilist_set_column_width
	    (unilist, i, (unilist->column[i].button->requisition.width -
			(CELL_SPACING + (2 * COLUMN_INSET))));
	else
	  gtk_unilist_set_column_width (unilist, i, 0);
      }
  /* zero-out the scrollbars */
  if (unilist->vadjustment)
    {
      gtk_adjustment_set_value (unilist->vadjustment, 0.0);
      UNILIST_REFRESH (unilist);
    }
  else
    gtk_widget_queue_resize (GTK_WIDGET (unilist));
}

static void
real_row_move (GtkUniList *unilist,
	       gint      source_row,
	       gint      dest_row)
{
  GtkUniListRow *unilist_row;
  GList *list;
  gint first, last;
  gint d;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (GTK_UNILIST_AUTO_SORT(unilist))
    return;

  if (source_row < 0 || source_row >= unilist->rows ||
      dest_row   < 0 || dest_row   >= unilist->rows ||
      source_row == dest_row)
    return;

  gtk_unilist_freeze (unilist);

  /* unlink source row */
  unilist_row = ROW_ELEMENT (unilist, source_row)->data;
  if (source_row == unilist->rows - 1)
    unilist->row_list_end = unilist->row_list_end->prev;
  unilist->row_list = g_list_remove (unilist->row_list, unilist_row);
  unilist->rows--;

  /* relink source row */
  unilist->row_list = g_list_insert (unilist->row_list, unilist_row, dest_row);
  if (dest_row == unilist->rows)
    unilist->row_list_end = unilist->row_list_end->next;
  unilist->rows++;

  /* sync selection */
  if (source_row > dest_row)
    {
      first = dest_row;
      last  = source_row;
      d = 1;
    }
  else
    {
      first = source_row;
      last  = dest_row;
      d = -1;
    }

  for (list = unilist->selection; list; list = list->next)
    {
      if (list->data == GINT_TO_POINTER (source_row))
	list->data = GINT_TO_POINTER (dest_row);
      else if (first <= GPOINTER_TO_INT (list->data) &&
	       last >= GPOINTER_TO_INT (list->data))
	list->data = GINT_TO_POINTER (GPOINTER_TO_INT (list->data) + d);
    }
  
  if (unilist->focus_row == source_row)
    unilist->focus_row = dest_row;
  else if (unilist->focus_row > first)
    unilist->focus_row += d;

  gtk_unilist_thaw (unilist);
}

/* PUBLIC ROW FUNCTIONS
 *   gtk_unilist_moveto
 *   gtk_unilist_set_row_height
 *   gtk_unilist_set_row_data
 *   gtk_unilist_set_row_data_full
 *   gtk_unilist_get_row_data
 *   gtk_unilist_find_row_from_data
 *   gtk_unilist_swap_rows
 *   gtk_unilist_row_move
 *   gtk_unilist_row_is_visible
 *   gtk_unilist_set_foreground
 *   gtk_unilist_set_background
 */
void
gtk_unilist_moveto (GtkUniList *unilist,
		  gint      row,
		  gint      column,
		  gfloat    row_align,
		  gfloat    col_align)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (row < -1 || row >= unilist->rows)
    return;
  if (column < -1 || column >= unilist->columns)
    return;

  row_align = CLAMP (row_align, 0, 1);
  col_align = CLAMP (col_align, 0, 1);

  /* adjust horizontal scrollbar */
  if (unilist->hadjustment && column >= 0)
    {
      gint x;

      x = (COLUMN_LEFT (unilist, column) - CELL_SPACING - COLUMN_INSET -
	   (col_align * (unilist->unilist_window_width - 2 * COLUMN_INSET -
			 CELL_SPACING - unilist->column[column].area.width)));
      if (x < 0)
	gtk_adjustment_set_value (unilist->hadjustment, 0.0);
      else if (x > LIST_WIDTH (unilist) - unilist->unilist_window_width)
	gtk_adjustment_set_value 
	  (unilist->hadjustment, LIST_WIDTH (unilist) - unilist->unilist_window_width);
      else
	gtk_adjustment_set_value (unilist->hadjustment, x);
    }

  /* adjust vertical scrollbar */
  if (unilist->vadjustment && row >= 0)
    move_vertical (unilist, row, row_align);
}

void
gtk_unilist_set_row_height (GtkUniList *unilist,
			  guint     height)
{
  GtkWidget *widget;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  widget = GTK_WIDGET (unilist);

  if (height > 0)
    {
      unilist->row_height = height;
      GTK_UNILIST_SET_FLAG (unilist, UNILIST_ROW_HEIGHT_SET);
    }
  else
    {
      GTK_UNILIST_UNSET_FLAG (unilist, UNILIST_ROW_HEIGHT_SET);
      unilist->row_height = 0;
    }

  if (widget->style->font_desc)
    {
      PangoContext *context = gtk_widget_get_pango_context (widget);
      PangoFontMetrics *metrics;

      metrics = pango_context_get_metrics (context,
					   widget->style->font_desc,
					   pango_context_get_language (context));
      
      if (!GTK_UNILIST_ROW_HEIGHT_SET(unilist))
	{
	  unilist->row_height = (pango_font_metrics_get_ascent (metrics) +
			       pango_font_metrics_get_descent (metrics));
	  unilist->row_height = PANGO_PIXELS (unilist->row_height);
	}

      pango_font_metrics_unref (metrics);
    }
      
  UNILIST_REFRESH (unilist);
}

void
gtk_unilist_set_row_data (GtkUniList *unilist,
			gint      row,
			gpointer  data)
{
  gtk_unilist_set_row_data_full (unilist, row, data, NULL);
}

void
gtk_unilist_set_row_data_full (GtkUniList         *unilist,
			     gint              row,
			     gpointer          data,
			     GtkDestroyNotify  destroy)
{
  GtkUniListRow *unilist_row;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (row < 0 || row > (unilist->rows - 1))
    return;

  unilist_row = ROW_ELEMENT (unilist, row)->data;

  if (unilist_row->destroy)
    unilist_row->destroy (unilist_row->data);
  
  unilist_row->data = data;
  unilist_row->destroy = destroy;
}

gpointer
gtk_unilist_get_row_data (GtkUniList *unilist,
			gint      row)
{
  GtkUniListRow *unilist_row;

  g_return_val_if_fail (GTK_IS_UNILIST (unilist), NULL);

  if (row < 0 || row > (unilist->rows - 1))
    return NULL;

  unilist_row = ROW_ELEMENT (unilist, row)->data;
  return unilist_row->data;
}

gint
gtk_unilist_find_row_from_data (GtkUniList *unilist,
			      gpointer  data)
{
  GList *list;
  gint n;

  g_return_val_if_fail (GTK_IS_UNILIST (unilist), -1);

  for (n = 0, list = unilist->row_list; list; n++, list = list->next)
    if (GTK_UNILIST_ROW (list)->data == data)
      return n;

  return -1;
}

void 
gtk_unilist_swap_rows (GtkUniList *unilist,
		     gint      row1, 
		     gint      row2)
{
  gint first, last;

  g_return_if_fail (GTK_IS_UNILIST (unilist));
  g_return_if_fail (row1 != row2);

  if (GTK_UNILIST_AUTO_SORT(unilist))
    return;

  gtk_unilist_freeze (unilist);

  first = MIN (row1, row2);
  last  = MAX (row1, row2);

  gtk_unilist_row_move (unilist, last, first);
  gtk_unilist_row_move (unilist, first + 1, last);
  
  gtk_unilist_thaw (unilist);
}

void
gtk_unilist_row_move (GtkUniList *unilist,
		    gint      source_row,
		    gint      dest_row)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (GTK_UNILIST_AUTO_SORT(unilist))
    return;

  if (source_row < 0 || source_row >= unilist->rows ||
      dest_row   < 0 || dest_row   >= unilist->rows ||
      source_row == dest_row)
    return;

  gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[ROW_MOVE],
		   source_row, dest_row);
}

GtkVisibility
gtk_unilist_row_is_visible (GtkUniList *unilist,
			  gint      row)
{
  gint top;

  g_return_val_if_fail (GTK_IS_UNILIST (unilist), 0);

  if (row < 0 || row >= unilist->rows)
    return GTK_VISIBILITY_NONE;

  if (unilist->row_height == 0)
    return GTK_VISIBILITY_NONE;

  if (row < ROW_FROM_YPIXEL (unilist, 0))
    return GTK_VISIBILITY_NONE;

  if (row > ROW_FROM_YPIXEL (unilist, unilist->unilist_window_height))
    return GTK_VISIBILITY_NONE;

  top = ROW_TOP_YPIXEL (unilist, row);

  if ((top < 0)
      || ((top + unilist->row_height) >= unilist->unilist_window_height))
    return GTK_VISIBILITY_PARTIAL;

  return GTK_VISIBILITY_FULL;
}

void
gtk_unilist_set_foreground (GtkUniList *unilist,
			  gint      row,
			  GdkColor *color)
{
  GtkUniListRow *unilist_row;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (row < 0 || row >= unilist->rows)
    return;

  unilist_row = ROW_ELEMENT (unilist, row)->data;

  if (color)
    {
      unilist_row->foreground = *color;
      unilist_row->fg_set = TRUE;
      if (GTK_WIDGET_REALIZED (unilist))
	gdk_color_alloc (gtk_widget_get_colormap (GTK_WIDGET (unilist)),
			 &unilist_row->foreground);
    }
  else
    unilist_row->fg_set = FALSE;

  if (UNILIST_UNFROZEN (unilist) && gtk_unilist_row_is_visible (unilist, row) != GTK_VISIBILITY_NONE)
    GTK_UNILIST_GET_CLASS (unilist)->draw_row (unilist, NULL, row, unilist_row);
}

void
gtk_unilist_set_background (GtkUniList *unilist,
			  gint      row,
			  GdkColor *color)
{
  GtkUniListRow *unilist_row;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (row < 0 || row >= unilist->rows)
    return;

  unilist_row = ROW_ELEMENT (unilist, row)->data;

  if (color)
    {
      unilist_row->background = *color;
      unilist_row->bg_set = TRUE;
      if (GTK_WIDGET_REALIZED (unilist))
	gdk_color_alloc (gtk_widget_get_colormap (GTK_WIDGET (unilist)),
			 &unilist_row->background);
    }
  else
    unilist_row->bg_set = FALSE;

  if (UNILIST_UNFROZEN (unilist)
      && (gtk_unilist_row_is_visible (unilist, row) != GTK_VISIBILITY_NONE))
    GTK_UNILIST_GET_CLASS (unilist)->draw_row (unilist, NULL, row, unilist_row);
}

/* PUBLIC ROW/CELL STYLE FUNCTIONS
 *   gtk_unilist_set_cell_style
 *   gtk_unilist_get_cell_style
 *   gtk_unilist_set_row_style
 *   gtk_unilist_get_row_style
 */
void
gtk_unilist_set_cell_style (GtkUniList *unilist,
			  gint      row,
			  gint      column,
			  GtkStyle *style)
{
  GtkRequisition requisition = { 0 };
  GtkUniListRow *unilist_row;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (row < 0 || row >= unilist->rows)
    return;
  if (column < 0 || column >= unilist->columns)
    return;

  unilist_row = ROW_ELEMENT (unilist, row)->data;

  if (unilist_row->cell[column].style == style)
    return;

  if (unilist->column[column].auto_resize &&
      !GTK_UNILIST_AUTO_RESIZE_BLOCKED (unilist))
    GTK_UNILIST_GET_CLASS (unilist)->cell_size_request (unilist, unilist_row,
						   column, &requisition);

  if (unilist_row->cell[column].style)
    {
      if (GTK_WIDGET_REALIZED (unilist))
        gtk_style_detach (unilist_row->cell[column].style);
      g_object_unref (unilist_row->cell[column].style);
    }

  unilist_row->cell[column].style = style;

  if (unilist_row->cell[column].style)
    {
      g_object_ref (unilist_row->cell[column].style);
      
      if (GTK_WIDGET_REALIZED (unilist))
        unilist_row->cell[column].style =
	  gtk_style_attach (unilist_row->cell[column].style,
			    unilist->unilist_window);
    }

  column_auto_resize (unilist, unilist_row, column, requisition.width);

  /* redraw the list if it's not frozen */
  if (UNILIST_UNFROZEN (unilist))
    {
      if (gtk_unilist_row_is_visible (unilist, row) != GTK_VISIBILITY_NONE)
	GTK_UNILIST_GET_CLASS (unilist)->draw_row (unilist, NULL, row, unilist_row);
    }
}

GtkStyle *
gtk_unilist_get_cell_style (GtkUniList *unilist,
			  gint      row,
			  gint      column)
{
  GtkUniListRow *unilist_row;

  g_return_val_if_fail (GTK_IS_UNILIST (unilist), NULL);

  if (row < 0 || row >= unilist->rows || column < 0 || column >= unilist->columns)
    return NULL;

  unilist_row = ROW_ELEMENT (unilist, row)->data;

  return unilist_row->cell[column].style;
}

void
gtk_unilist_set_row_style (GtkUniList *unilist,
			 gint      row,
			 GtkStyle *style)
{
  GtkRequisition requisition;
  GtkUniListRow *unilist_row;
  gint *old_width;
  gint i;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (row < 0 || row >= unilist->rows)
    return;

  unilist_row = ROW_ELEMENT (unilist, row)->data;

  if (unilist_row->style == style)
    return;

  old_width = g_new (gint, unilist->columns);

  if (!GTK_UNILIST_AUTO_RESIZE_BLOCKED (unilist))
    {
      for (i = 0; i < unilist->columns; i++)
	if (unilist->column[i].auto_resize)
	  {
	    GTK_UNILIST_GET_CLASS (unilist)->cell_size_request (unilist, unilist_row,
							   i, &requisition);
	    old_width[i] = requisition.width;
	  }
    }

  if (unilist_row->style)
    {
      if (GTK_WIDGET_REALIZED (unilist))
        gtk_style_detach (unilist_row->style);
      g_object_unref (unilist_row->style);
    }

  unilist_row->style = style;

  if (unilist_row->style)
    {
      g_object_ref (unilist_row->style);
      
      if (GTK_WIDGET_REALIZED (unilist))
        unilist_row->style = gtk_style_attach (unilist_row->style,
					     unilist->unilist_window);
    }

  if (GTK_UNILIST_AUTO_RESIZE_BLOCKED (unilist))
    for (i = 0; i < unilist->columns; i++)
      column_auto_resize (unilist, unilist_row, i, old_width[i]);

  g_free (old_width);

  /* redraw the list if it's not frozen */
  if (UNILIST_UNFROZEN (unilist))
    {
      if (gtk_unilist_row_is_visible (unilist, row) != GTK_VISIBILITY_NONE)
	GTK_UNILIST_GET_CLASS (unilist)->draw_row (unilist, NULL, row, unilist_row);
    }
}

GtkStyle *
gtk_unilist_get_row_style (GtkUniList *unilist,
			 gint      row)
{
  GtkUniListRow *unilist_row;

  g_return_val_if_fail (GTK_IS_UNILIST (unilist), NULL);

  if (row < 0 || row >= unilist->rows)
    return NULL;

  unilist_row = ROW_ELEMENT (unilist, row)->data;

  return unilist_row->style;
}

/* PUBLIC SELECTION FUNCTIONS
 *   gtk_unilist_set_selectable
 *   gtk_unilist_get_selectable
 *   gtk_unilist_select_row
 *   gtk_unilist_unselect_row
 *   gtk_unilist_select_all
 *   gtk_unilist_unselect_all
 *   gtk_unilist_undo_selection
 */
void
gtk_unilist_set_selectable (GtkUniList *unilist,
			  gint      row,
			  gboolean  selectable)
{
  GtkUniListRow *unilist_row;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (row < 0 || row >= unilist->rows)
    return;

  unilist_row = ROW_ELEMENT (unilist, row)->data;

  if (selectable == unilist_row->selectable)
    return;

  unilist_row->selectable = selectable;

  if (!selectable && unilist_row->state == GTK_STATE_SELECTED)
    {
      if (unilist->anchor >= 0 &&
	  unilist->selection_mode == GTK_SELECTION_MULTIPLE)
	{
	  unilist->drag_button = 0;
	  remove_grab (unilist);
	  GTK_UNILIST_GET_CLASS (unilist)->resync_selection (unilist, NULL);
	}
      gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[UNSELECT_ROW],
		       row, -1, NULL);
    }      
}

gboolean
gtk_unilist_get_selectable (GtkUniList *unilist,
			  gint      row)
{
  g_return_val_if_fail (GTK_IS_UNILIST (unilist), FALSE);

  if (row < 0 || row >= unilist->rows)
    return FALSE;

  return GTK_UNILIST_ROW (ROW_ELEMENT (unilist, row))->selectable;
}

void
gtk_unilist_select_row (GtkUniList *unilist,
		      gint      row,
		      gint      column)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (row < 0 || row >= unilist->rows)
    return;
  if (column < -1 || column >= unilist->columns)
    return;

  gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[SELECT_ROW],
		   row, column, NULL);
//    set_anchor (unilist, GTK_UNILIST_ADD_MODE(unilist), unilist->focus_row,
//		unilist->focus_row);
//    end_selection(unilist);
}

void
gtk_unilist_unselect_row (GtkUniList *unilist,
			gint      row,
			gint      column)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (row < 0 || row >= unilist->rows)
    return;
  if (column < -1 || column >= unilist->columns)
    return;

  gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[UNSELECT_ROW],
		   row, column, NULL);
}

void
gtk_unilist_select_all (GtkUniList *unilist)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  GTK_UNILIST_GET_CLASS (unilist)->select_all (unilist);
}

void
gtk_unilist_unselect_all (GtkUniList *unilist)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  GTK_UNILIST_GET_CLASS (unilist)->unselect_all (unilist);
}

void
gtk_unilist_undo_selection (GtkUniList *unilist)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (unilist->selection_mode == GTK_SELECTION_MULTIPLE &&
      (unilist->undo_selection || unilist->undo_unselection))
    gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[UNDO_SELECTION]);
}

/* PRIVATE SELECTION FUNCTIONS
 *   selection_find
 *   toggle_row
 *   fake_toggle_row
 *   toggle_focus_row
 *   toggle_add_mode
 *   real_select_row
 *   real_unselect_row
 *   real_select_all
 *   real_unselect_all
 *   fake_unselect_all
 *   real_undo_selection
 *   set_anchor
 *   resync_selection
 *   update_extended_selection
 *   start_selection
 *   end_selection
 *   extend_selection
 *   sync_selection
 */
static GList *
selection_find (GtkUniList *unilist,
		gint      row_number,
		GList    *row_list_element)
{
  return g_list_find (unilist->selection, GINT_TO_POINTER (row_number));
}

static void
toggle_row (GtkUniList *unilist,
	    gint      row,
	    gint      column,
	    GdkEvent *event)
{
  GtkUniListRow *unilist_row;

  switch (unilist->selection_mode)
    {
    case GTK_SELECTION_MULTIPLE:
    case GTK_SELECTION_SINGLE:
      unilist_row = ROW_ELEMENT (unilist, row)->data;

      if (!unilist_row)
	return;

      if (unilist_row->state == GTK_STATE_SELECTED)
	{
	  gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[UNSELECT_ROW],
			   row, column, event);
	  return;
	}
    case GTK_SELECTION_BROWSE:
      gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[SELECT_ROW],
		       row, column, event);
      break;
    default:
      g_assert_not_reached ();
    }
}

gboolean gtk_unilist_row_is_selected (GtkUniList *unilist,
				   gint      row)
{
	  GList *work;

  work = ROW_ELEMENT (unilist, row);

  if (!work || !GTK_UNILIST_ROW (work)->selectable)
    return FALSE;
	
	return (GTK_UNILIST_ROW (work)->state == GTK_STATE_SELECTED);
}

void
gtk_unilist_real_select_row (GtkUniList *unilist,
		 gint      row)
{
	  GList *work;

  work = ROW_ELEMENT (unilist, row);

  if (!work || !GTK_UNILIST_ROW (work)->selectable)
    return;
  
    unilist->anchor_state = GTK_UNILIST_ROW (work)->state = GTK_STATE_SELECTED;

  if (UNILIST_UNFROZEN (unilist) &&
      gtk_unilist_row_is_visible (unilist, row) != GTK_VISIBILITY_NONE)
    GTK_UNILIST_GET_CLASS (unilist)->draw_row (unilist, NULL, row,
					  GTK_UNILIST_ROW (work));
}
void
gtk_unilist_real_unselect_row (GtkUniList *unilist,
		 gint      row)
{
	  GList *work;

  work = ROW_ELEMENT (unilist, row);

  if (!work || !GTK_UNILIST_ROW (work)->selectable)
    return;
  
    unilist->anchor_state = GTK_UNILIST_ROW (work)->state = GTK_STATE_NORMAL;

  if (UNILIST_UNFROZEN (unilist) &&
      gtk_unilist_row_is_visible (unilist, row) != GTK_VISIBILITY_NONE)
    GTK_UNILIST_GET_CLASS (unilist)->draw_row (unilist, NULL, row,
					  GTK_UNILIST_ROW (work));

}
void gtk_unilist_toggle_row (GtkUniList *unilist,
		 gint      row)
{
	fake_toggle_row(unilist,row);
}
static void
fake_toggle_row (GtkUniList *unilist,
		 gint      row)
{
  GList *work;

  work = ROW_ELEMENT (unilist, row);

  if (!work || !GTK_UNILIST_ROW (work)->selectable)
    return;
  
  if (GTK_UNILIST_ROW (work)->state == GTK_STATE_NORMAL)
    unilist->anchor_state = GTK_UNILIST_ROW (work)->state = GTK_STATE_SELECTED;
  else
    unilist->anchor_state = GTK_UNILIST_ROW (work)->state = GTK_STATE_NORMAL;
  
  if (UNILIST_UNFROZEN (unilist) &&
      gtk_unilist_row_is_visible (unilist, row) != GTK_VISIBILITY_NONE){
    GTK_UNILIST_GET_CLASS (unilist)->draw_row (unilist, NULL, row,
					  GTK_UNILIST_ROW (work));
    gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[SELECTION_CHANGED], NULL);
	  }

}

static void
toggle_focus_row (GtkUniList *unilist)
{
  g_return_if_fail (unilist != 0);
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if ((gdk_pointer_is_grabbed () && GTK_WIDGET_HAS_GRAB (unilist)) ||
      unilist->focus_row < 0 || unilist->focus_row >= unilist->rows)
    return;

  switch (unilist->selection_mode)
    {
    case  GTK_SELECTION_SINGLE:
      toggle_row (unilist, unilist->focus_row, 0, NULL);
      break;
    case GTK_SELECTION_MULTIPLE:
      g_list_free (unilist->undo_selection);
      g_list_free (unilist->undo_unselection);
      unilist->undo_selection = NULL;
      unilist->undo_unselection = NULL;

	if (unilist->anchor == -1)fake_toggle_row (unilist, unilist->focus_row);
	else gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[SELECTION_CHANGED], NULL);
	

//      unilist->anchor = unilist->focus_row;
//      unilist->drag_pos = unilist->focus_row;
//      unilist->undo_anchor = unilist->focus_row;
      unilist->anchor = -1;
      unilist->drag_pos = -1;
      unilist->undo_anchor = -1;
      
///*
//	if (GTK_UNILIST_ADD_MODE(unilist))
//	fake_toggle_row (unilist, unilist->focus_row);
//      else
//	GTK_UNILIST_GET_CLASS (unilist)->fake_unselect_all (unilist,unilist->focus_row);
//*/
      GTK_UNILIST_GET_CLASS (unilist)->resync_selection (unilist, NULL);
//      gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[SELECTION_CHANGED], NULL);

      break;
    default:
      break;
    }
}

static void
toggle_add_mode (GtkUniList *unilist)
{
  g_return_if_fail (unilist != 0);
  g_return_if_fail (GTK_IS_UNILIST (unilist));
  
  if ((gdk_pointer_is_grabbed () && GTK_WIDGET_HAS_GRAB (unilist)) ||
      unilist->selection_mode != GTK_SELECTION_MULTIPLE)
    return;

  gtk_unilist_draw_focus (GTK_WIDGET (unilist));
  if (!GTK_UNILIST_ADD_MODE(unilist))
    {
      GTK_UNILIST_SET_FLAG (unilist, UNILIST_ADD_MODE);
      gdk_gc_set_line_attributes (unilist->xor_gc, 1,
				  GDK_LINE_ON_OFF_DASH, 0, 0);
      gdk_gc_set_dashes (unilist->xor_gc, 0, "\4\4", 2);
    }
  else
    {
      GTK_UNILIST_UNSET_FLAG (unilist, UNILIST_ADD_MODE);
      gdk_gc_set_line_attributes (unilist->xor_gc, 1, GDK_LINE_SOLID, 0, 0);
      unilist->anchor_state = GTK_STATE_SELECTED;
    }
  gtk_unilist_draw_focus (GTK_WIDGET (unilist));
}

static void
real_select_row (GtkUniList *unilist,
		 gint      row,
		 gint      column,
		 GdkEvent *event)
{
  GtkUniListRow *unilist_row;
  GList *list;
  gint sel_row;
  gboolean row_selected;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (row < 0 || row > (unilist->rows - 1))
    return;

  switch (unilist->selection_mode)
    {
    case GTK_SELECTION_SINGLE:
    case GTK_SELECTION_BROWSE:

      row_selected = FALSE;
      list = unilist->selection;

      while (list)
	{
	  sel_row = GPOINTER_TO_INT (list->data);
	  list = list->next;

	  if (row == sel_row)
	    row_selected = TRUE;
	  else
	    gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[UNSELECT_ROW], 
			     sel_row, column, event);
	}

      if (row_selected)
	return;
      
    default:
      break;
    }

  unilist_row = ROW_ELEMENT (unilist, row)->data;

  if (unilist_row->state != GTK_STATE_NORMAL || !unilist_row->selectable)
    return;

  unilist_row->state = GTK_STATE_SELECTED;
  if (!unilist->selection)
    {
      unilist->selection = g_list_append (unilist->selection,
					GINT_TO_POINTER (row));
      unilist->selection_end = unilist->selection;
    }
  else
    unilist->selection_end = 
      g_list_append (unilist->selection_end, GINT_TO_POINTER (row))->next;
  
  if (UNILIST_UNFROZEN (unilist)
      && (gtk_unilist_row_is_visible (unilist, row) != GTK_VISIBILITY_NONE))
    GTK_UNILIST_GET_CLASS (unilist)->draw_row (unilist, NULL, row, unilist_row);
}

static void
real_unselect_row (GtkUniList *unilist,
		   gint      row,
		   gint      column,
		   GdkEvent *event)
{
  GtkUniListRow *unilist_row;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (row < 0 || row > (unilist->rows - 1))
    return;

  unilist_row = ROW_ELEMENT (unilist, row)->data;

  if (unilist_row->state == GTK_STATE_SELECTED)
    {
      unilist_row->state = GTK_STATE_NORMAL;

      if (unilist->selection_end && 
	  unilist->selection_end->data == GINT_TO_POINTER (row))
	unilist->selection_end = unilist->selection_end->prev;

      unilist->selection = g_list_remove (unilist->selection,
					GINT_TO_POINTER (row));
      
      if (UNILIST_UNFROZEN (unilist)
	  && (gtk_unilist_row_is_visible (unilist, row) != GTK_VISIBILITY_NONE))
	GTK_UNILIST_GET_CLASS (unilist)->draw_row (unilist, NULL, row, unilist_row);
    }
}

static void
real_select_all (GtkUniList *unilist)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (gdk_pointer_is_grabbed () && GTK_WIDGET_HAS_GRAB (unilist))
    return;

  switch (unilist->selection_mode)
    {
    case GTK_SELECTION_SINGLE:
    case GTK_SELECTION_BROWSE:
      return;

    case GTK_SELECTION_MULTIPLE:
      g_list_free (unilist->undo_selection);
      g_list_free (unilist->undo_unselection);
      unilist->undo_selection = NULL;
      unilist->undo_unselection = NULL;
	  
      if (unilist->rows &&
	  ((GtkUniListRow *) (unilist->row_list->data))->state !=
	  GTK_STATE_SELECTED)
	fake_toggle_row (unilist, 0);

      unilist->anchor_state =  GTK_STATE_SELECTED;
      unilist->anchor = 0;
      unilist->drag_pos = 0;
      unilist->undo_anchor = unilist->focus_row;
      update_extended_selection (unilist, unilist->rows);
      GTK_UNILIST_GET_CLASS (unilist)->resync_selection (unilist, NULL);
      return;
    default:
      g_assert_not_reached ();
    }
}

static void
real_unselect_all (GtkUniList *unilist)
{
  GList *list;
  gint i;
 
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (gdk_pointer_is_grabbed () && GTK_WIDGET_HAS_GRAB (unilist))
    return;

  switch (unilist->selection_mode)
    {
    case GTK_SELECTION_BROWSE:
      if (unilist->focus_row >= 0)
	{
	  gtk_signal_emit (GTK_OBJECT (unilist),
			   unilist_signals[SELECT_ROW],
			   unilist->focus_row, -1, NULL);
	  return;
	}
      break;
    case GTK_SELECTION_MULTIPLE:
      g_list_free (unilist->undo_selection);
      g_list_free (unilist->undo_unselection);
      unilist->undo_selection = NULL;
      unilist->undo_unselection = NULL;

      unilist->anchor = -1;
      unilist->drag_pos = -1;
      unilist->undo_anchor = unilist->focus_row;
      break;
    default:
      break;
    }

  list = unilist->selection;
  while (list)
    {
      i = GPOINTER_TO_INT (list->data);
      list = list->next;
      gtk_signal_emit (GTK_OBJECT (unilist),
		       unilist_signals[UNSELECT_ROW], i, -1, NULL);
    }
}

static void
fake_unselect_all (GtkUniList *unilist,
		   gint      row)
{
  GList *list;
  GList *work;
  gint i;

  if (row >= 0 && (work = ROW_ELEMENT (unilist, row)))
    {
      if (GTK_UNILIST_ROW (work)->state == GTK_STATE_NORMAL &&
	  GTK_UNILIST_ROW (work)->selectable)
	{
	  GTK_UNILIST_ROW (work)->state = GTK_STATE_SELECTED;
	  
	  if (UNILIST_UNFROZEN (unilist) &&
	      gtk_unilist_row_is_visible (unilist, row) != GTK_VISIBILITY_NONE)
	    GTK_UNILIST_GET_CLASS (unilist)->draw_row (unilist, NULL, row,
						  GTK_UNILIST_ROW (work));
	}  
    }

  unilist->undo_selection = unilist->selection;
  unilist->selection = NULL;
  unilist->selection_end = NULL;

  for (list = unilist->undo_selection; list; list = list->next)
    {
      if ((i = GPOINTER_TO_INT (list->data)) == row ||
	  !(work = g_list_nth (unilist->row_list, i)))
	continue;

      GTK_UNILIST_ROW (work)->state = GTK_STATE_NORMAL;
      if (UNILIST_UNFROZEN (unilist) &&
	  gtk_unilist_row_is_visible (unilist, i) != GTK_VISIBILITY_NONE)
	GTK_UNILIST_GET_CLASS (unilist)->draw_row (unilist, NULL, i,
					      GTK_UNILIST_ROW (work));
    }
}

static void
real_undo_selection (GtkUniList *unilist)
{
  GList *work;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if ((gdk_pointer_is_grabbed () && GTK_WIDGET_HAS_GRAB (unilist)) ||
      unilist->selection_mode != GTK_SELECTION_MULTIPLE)
    return;

  GTK_UNILIST_GET_CLASS (unilist)->resync_selection (unilist, NULL);

  if (!(unilist->undo_selection || unilist->undo_unselection))
    {
      gtk_unilist_unselect_all (unilist);
      return;
    }

  for (work = unilist->undo_selection; work; work = work->next)
    gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[SELECT_ROW],
		     GPOINTER_TO_INT (work->data), -1, NULL);

  for (work = unilist->undo_unselection; work; work = work->next)
    {
      /* g_print ("unselect %d\n",GPOINTER_TO_INT (work->data)); */
      gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[UNSELECT_ROW], 
		       GPOINTER_TO_INT (work->data), -1, NULL);
    }

  if (GTK_WIDGET_HAS_FOCUS(unilist) && unilist->focus_row != unilist->undo_anchor)
    {
      gtk_unilist_draw_focus (GTK_WIDGET (unilist));
      unilist->focus_row = unilist->undo_anchor;
      gtk_unilist_draw_focus (GTK_WIDGET (unilist));
    }
  else
    unilist->focus_row = unilist->undo_anchor;
  
  unilist->undo_anchor = -1;
 
  g_list_free (unilist->undo_selection);
  g_list_free (unilist->undo_unselection);
  unilist->undo_selection = NULL;
  unilist->undo_unselection = NULL;

  if (ROW_TOP_YPIXEL (unilist, unilist->focus_row) + unilist->row_height >
      unilist->unilist_window_height)
    gtk_unilist_moveto (unilist, unilist->focus_row, -1, 1, 0);
  else if (ROW_TOP_YPIXEL (unilist, unilist->focus_row) < 0)
    gtk_unilist_moveto (unilist, unilist->focus_row, -1, 0, 0);
  gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[SELECTION_CHANGED], NULL);

}

static void
set_anchor (GtkUniList *unilist,
	    gboolean  add_mode,
	    gint      anchor,
	    gint      undo_anchor)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));
  
  if (unilist->selection_mode != GTK_SELECTION_MULTIPLE || unilist->anchor >= 0)
    return;

  g_list_free (unilist->undo_selection);
  g_list_free (unilist->undo_unselection);
  unilist->undo_selection = NULL;
  unilist->undo_unselection = NULL;

  if (add_mode)
    fake_toggle_row (unilist, anchor);
  else
    {
      GTK_UNILIST_GET_CLASS (unilist)->fake_unselect_all (unilist, anchor);
      unilist->anchor_state = GTK_STATE_SELECTED;
    }

  unilist->anchor = anchor;
  unilist->drag_pos = anchor;
  unilist->undo_anchor = undo_anchor;
}

static void
resync_selection (GtkUniList *unilist,
		  GdkEvent *event)
{
  gint i;
  gint e;
  gint row;
  GList *list;
  GtkUniListRow *unilist_row;

  if (unilist->selection_mode != GTK_SELECTION_MULTIPLE)
    return;

  if (unilist->anchor < 0 || unilist->drag_pos < 0)
    return;

  gtk_unilist_freeze (unilist);

  i = MIN (unilist->anchor, unilist->drag_pos);
  e = MAX (unilist->anchor, unilist->drag_pos);

  if (unilist->undo_selection)
    {
      list = unilist->selection;
      unilist->selection = unilist->undo_selection;
      unilist->selection_end = g_list_last (unilist->selection);
      unilist->undo_selection = list;
      list = unilist->selection;
      while (list)
	{
	  row = GPOINTER_TO_INT (list->data);
	  list = list->next;
	  if (row < i || row > e)
	    {
	      unilist_row = g_list_nth (unilist->row_list, row)->data;
	      if (unilist_row->selectable)
		{
		  unilist_row->state = GTK_STATE_SELECTED;
		  gtk_signal_emit (GTK_OBJECT (unilist),
				   unilist_signals[UNSELECT_ROW],
				   row, -1, event);
		  unilist->undo_selection = g_list_prepend
		    (unilist->undo_selection, GINT_TO_POINTER (row));
		}
	    }
	}
    }    

  if (unilist->anchor < unilist->drag_pos)
    {
      for (list = g_list_nth (unilist->row_list, i); i <= e;
	   i++, list = list->next)
	if (GTK_UNILIST_ROW (list)->selectable)
	  {
	    if (g_list_find (unilist->selection, GINT_TO_POINTER(i)))
	      {
		if (GTK_UNILIST_ROW (list)->state == GTK_STATE_NORMAL)
		  {
		    GTK_UNILIST_ROW (list)->state = GTK_STATE_SELECTED;
		    gtk_signal_emit (GTK_OBJECT (unilist),
				     unilist_signals[UNSELECT_ROW],
				     i, -1, event);
		    unilist->undo_selection =
		      g_list_prepend (unilist->undo_selection,
				      GINT_TO_POINTER (i));
		  }
	      }
	    else if (GTK_UNILIST_ROW (list)->state == GTK_STATE_SELECTED)
	      {
		GTK_UNILIST_ROW (list)->state = GTK_STATE_NORMAL;
		unilist->undo_unselection =
		  g_list_prepend (unilist->undo_unselection,
				  GINT_TO_POINTER (i));
	      }
	  }
    }
  else
    {
      for (list = g_list_nth (unilist->row_list, e); i <= e;
	   e--, list = list->prev)
	if (GTK_UNILIST_ROW (list)->selectable)
	  {
	    if (g_list_find (unilist->selection, GINT_TO_POINTER(e)))
	      {
		if (GTK_UNILIST_ROW (list)->state == GTK_STATE_NORMAL)
		  {
		    GTK_UNILIST_ROW (list)->state = GTK_STATE_SELECTED;
		    gtk_signal_emit (GTK_OBJECT (unilist),
				     unilist_signals[UNSELECT_ROW],
				     e, -1, event);
		    unilist->undo_selection =
		      g_list_prepend (unilist->undo_selection,
				      GINT_TO_POINTER (e));
		  }
	      }
	    else if (GTK_UNILIST_ROW (list)->state == GTK_STATE_SELECTED)
	      {
		GTK_UNILIST_ROW (list)->state = GTK_STATE_NORMAL;
		unilist->undo_unselection =
		  g_list_prepend (unilist->undo_unselection,
				  GINT_TO_POINTER (e));
	      }
	  }
    }
  
  unilist->undo_unselection = g_list_reverse (unilist->undo_unselection);
  for (list = unilist->undo_unselection; list; list = list->next)
    gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[SELECT_ROW],
		     GPOINTER_TO_INT (list->data), -1, event);

  unilist->anchor = -1;
  unilist->drag_pos = -1;

  gtk_unilist_thaw (unilist);
}

static void
update_extended_selection (GtkUniList *unilist,
			   gint      row)
{
  gint i;
  GList *list;
  GdkRectangle area;
  gint s1 = -1;
  gint s2 = -1;
  gint e1 = -1;
  gint e2 = -1;
  gint y1 = unilist->unilist_window_height;
  gint y2 = unilist->unilist_window_height;
  gint h1 = 0;
  gint h2 = 0;
  gint top;

  if (unilist->selection_mode != GTK_SELECTION_MULTIPLE || unilist->anchor == -1)
    return;

  if (row < 0)
    row = 0;
  if (row >= unilist->rows)
    row = unilist->rows - 1;

  /* extending downwards */
  if (row > unilist->drag_pos && unilist->anchor <= unilist->drag_pos)
    {
      s2 = unilist->drag_pos + 1;
      e2 = row;
    }
  /* extending upwards */
  else if (row < unilist->drag_pos && unilist->anchor >= unilist->drag_pos)
    {
      s2 = row;
      e2 = unilist->drag_pos - 1;
    }
  else if (row < unilist->drag_pos && unilist->anchor < unilist->drag_pos)
    {
      e1 = unilist->drag_pos;
      /* row and drag_pos on different sides of anchor :
	 take back the selection between anchor and drag_pos,
         select between anchor and row */
      if (row < unilist->anchor)
	{
	  s1 = unilist->anchor + 1;
	  s2 = row;
	  e2 = unilist->anchor - 1;
	}
      /* take back the selection between anchor and drag_pos */
      else
	s1 = row + 1;
    }
  else if (row > unilist->drag_pos && unilist->anchor > unilist->drag_pos)
    {
      s1 = unilist->drag_pos;
      /* row and drag_pos on different sides of anchor :
	 take back the selection between anchor and drag_pos,
         select between anchor and row */
      if (row > unilist->anchor)
	{
	  e1 = unilist->anchor - 1;
	  s2 = unilist->anchor + 1;
	  e2 = row;
	}
      /* take back the selection between anchor and drag_pos */
      else
	e1 = row - 1;
    }

  unilist->drag_pos = row;

  area.x = 0;
  area.width = unilist->unilist_window_width;

  /* restore the elements between s1 and e1 */
  if (s1 >= 0)
    {
      for (i = s1, list = g_list_nth (unilist->row_list, i); i <= e1;
	   i++, list = list->next)
	if (GTK_UNILIST_ROW (list)->selectable)
	  {
	    if (GTK_UNILIST_GET_CLASS (unilist)->selection_find (unilist, i, list))
	      GTK_UNILIST_ROW (list)->state = GTK_STATE_SELECTED;
	    else
	      GTK_UNILIST_ROW (list)->state = GTK_STATE_NORMAL;
	  }

      top = ROW_TOP_YPIXEL (unilist, unilist->focus_row);

      if (top + unilist->row_height <= 0)
	{
	  area.y = 0;
	  area.height = ROW_TOP_YPIXEL (unilist, e1) + unilist->row_height;
	  draw_rows (unilist, &area);
	  gtk_unilist_moveto (unilist, unilist->focus_row, -1, 0, 0);
	}
      else if (top >= unilist->unilist_window_height)
	{
	  area.y = ROW_TOP_YPIXEL (unilist, s1) - 1;
	  area.height = unilist->unilist_window_height - area.y;
	  draw_rows (unilist, &area);
	  gtk_unilist_moveto (unilist, unilist->focus_row, -1, 1, 0);
	}
      else if (top < 0)
	gtk_unilist_moveto (unilist, unilist->focus_row, -1, 0, 0);
      else if (top + unilist->row_height > unilist->unilist_window_height)
	gtk_unilist_moveto (unilist, unilist->focus_row, -1, 1, 0);

      y1 = ROW_TOP_YPIXEL (unilist, s1) - 1;
      h1 = (e1 - s1 + 1) * (unilist->row_height + CELL_SPACING);
    }

  /* extend the selection between s2 and e2 */
  if (s2 >= 0)
    {
      for (i = s2, list = g_list_nth (unilist->row_list, i); i <= e2;
	   i++, list = list->next)
	if (GTK_UNILIST_ROW (list)->selectable &&
	    GTK_UNILIST_ROW (list)->state != unilist->anchor_state)
	  GTK_UNILIST_ROW (list)->state = unilist->anchor_state;

      top = ROW_TOP_YPIXEL (unilist, unilist->focus_row);

      if (top + unilist->row_height <= 0)
	{
	  area.y = 0;
	  area.height = ROW_TOP_YPIXEL (unilist, e2) + unilist->row_height;
	  draw_rows (unilist, &area);
	  gtk_unilist_moveto (unilist, unilist->focus_row, -1, 0, 0);
	}
      else if (top >= unilist->unilist_window_height)
	{
	  area.y = ROW_TOP_YPIXEL (unilist, s2) - 1;
	  area.height = unilist->unilist_window_height - area.y;
	  draw_rows (unilist, &area);
	  gtk_unilist_moveto (unilist, unilist->focus_row, -1, 1, 0);
	}
      else if (top < 0)
	gtk_unilist_moveto (unilist, unilist->focus_row, -1, 0, 0);
      else if (top + unilist->row_height > unilist->unilist_window_height)
	gtk_unilist_moveto (unilist, unilist->focus_row, -1, 1, 0);

      y2 = ROW_TOP_YPIXEL (unilist, s2) - 1;
      h2 = (e2 - s2 + 1) * (unilist->row_height + CELL_SPACING);
    }

  area.y = MAX (0, MIN (y1, y2));
  if (area.y > unilist->unilist_window_height)
    area.y = 0;
  area.height = MIN (unilist->unilist_window_height, h1 + h2);
  if (s1 >= 0 && s2 >= 0)
    area.height += (unilist->row_height + CELL_SPACING);
  draw_rows (unilist, &area);
}

static void
start_selection (GtkUniList *unilist)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (gdk_pointer_is_grabbed () && GTK_WIDGET_HAS_GRAB (unilist))
    return;

  set_anchor (unilist, GTK_UNILIST_ADD_MODE(unilist), unilist->focus_row,
	      unilist->focus_row);
}

static void
end_selection (GtkUniList *unilist)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (gdk_pointer_is_grabbed () && GTK_WIDGET_HAS_FOCUS(unilist))
    return;
      unilist->anchor = -1;
      unilist->drag_pos = -1;
      unilist->undo_anchor = -1;
  GTK_UNILIST_GET_CLASS (unilist)->resync_selection (unilist, NULL);
  gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[SELECTION_CHANGED], NULL);
}

static void
extend_selection (GtkUniList      *unilist,
		  GtkScrollType  scroll_type,
		  gfloat         position,
		  gboolean       auto_start_selection)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if ((gdk_pointer_is_grabbed () && GTK_WIDGET_HAS_GRAB (unilist)) ||
      unilist->selection_mode != GTK_SELECTION_MULTIPLE)
    return;

  if (auto_start_selection)
    set_anchor (unilist, GTK_UNILIST_ADD_MODE(unilist), unilist->focus_row,
		unilist->focus_row);
  else if (unilist->anchor == -1)
    return;

  move_focus_row (unilist, scroll_type, position);

  if (ROW_TOP_YPIXEL (unilist, unilist->focus_row) + unilist->row_height >
      unilist->unilist_window_height)
    gtk_unilist_moveto (unilist, unilist->focus_row, -1, 1, 0);
  else if (ROW_TOP_YPIXEL (unilist, unilist->focus_row) < 0)
    gtk_unilist_moveto (unilist, unilist->focus_row, -1, 0, 0);

  update_extended_selection (unilist, unilist->focus_row);
}

static void
sync_selection (GtkUniList *unilist,
		gint      row,
		gint      mode)
{
  GList *list;
  gint d;

  if (mode == SYNC_INSERT)
    d = 1;
  else
    d = -1;
      
  if (unilist->focus_row >= row)
    {
      if (d > 0 || unilist->focus_row > row)
	unilist->focus_row += d;
      if (unilist->focus_row == -1 && unilist->rows >= 1)
	unilist->focus_row = 0;
      else if (unilist->focus_row >= unilist->rows)
	unilist->focus_row = unilist->rows - 1;
    }

  GTK_UNILIST_GET_CLASS (unilist)->resync_selection (unilist, NULL);

  g_list_free (unilist->undo_selection);
  g_list_free (unilist->undo_unselection);
  unilist->undo_selection = NULL;
  unilist->undo_unselection = NULL;

  unilist->anchor = -1;
  unilist->drag_pos = -1;
  unilist->undo_anchor = unilist->focus_row;

  list = unilist->selection;

  while (list)
    {
      if (GPOINTER_TO_INT (list->data) >= row)
	list->data = ((gchar*) list->data) + d;
      list = list->next;
    }
}

/* GTKOBJECT
 *   gtk_unilist_destroy
 *   gtk_unilist_finalize
 */
static void
gtk_unilist_destroy (GtkObject *object)
{
  gint i;
  GtkUniList *unilist;

  g_return_if_fail (GTK_IS_UNILIST (object));

  unilist = GTK_UNILIST (object);

  /* freeze the list */
  unilist->freeze_count++;

  /* get rid of all the rows */
  gtk_unilist_clear (unilist);

  /* Since we don't have a _remove method, unparent the children
   * instead of destroying them so the focus will be unset properly.
   * (For other containers, the _remove method takes care of the
   * unparent) The destroy will happen when the refcount drops
   * to zero.
   */

  /* unref adjustments */
  if (unilist->hadjustment)
    {
      gtk_signal_disconnect_by_data (GTK_OBJECT (unilist->hadjustment), unilist);
      gtk_object_unref (GTK_OBJECT (unilist->hadjustment));
      unilist->hadjustment = NULL;
    }
  if (unilist->vadjustment)
    {
      gtk_signal_disconnect_by_data (GTK_OBJECT (unilist->vadjustment), unilist);
      gtk_object_unref (GTK_OBJECT (unilist->vadjustment));
      unilist->vadjustment = NULL;
    }

  remove_grab (unilist);

  /* destroy the column buttons */
  for (i = 0; i < unilist->columns; i++)
    if (unilist->column[i].button)
      {
	gtk_widget_unparent (unilist->column[i].button);
	unilist->column[i].button = NULL;
      }

  if (GTK_OBJECT_CLASS (parent_class)->destroy)
    (*GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

static void
gtk_unilist_finalize (GObject *object)
{
  GtkUniList *unilist;

  g_return_if_fail (GTK_IS_UNILIST (object));

  unilist = GTK_UNILIST (object);

  columns_delete (unilist);

  g_mem_chunk_destroy (unilist->cell_mem_chunk);
  g_mem_chunk_destroy (unilist->row_mem_chunk);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}

/* GTKWIDGET
 *   gtk_unilist_realize
 *   gtk_unilist_unrealize
 *   gtk_unilist_map
 *   gtk_unilist_unmap
 *   gtk_unilist_expose
 *   gtk_unilist_style_set
 *   gtk_unilist_button_press
 *   gtk_unilist_button_release
 *   gtk_unilist_motion
 *   gtk_unilist_size_request
 *   gtk_unilist_size_allocate
 */
static void
gtk_unilist_realize (GtkWidget *widget)
{
  GtkUniList *unilist;
  GdkWindowAttr attributes;
  GdkGCValues values;
  GtkUniListRow *unilist_row;
  GList *list;
  gint attributes_mask;
  gint border_width;
  gint i;
  gint j;

  g_return_if_fail (GTK_IS_UNILIST (widget));

  unilist = GTK_UNILIST (widget);

  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);

  border_width = GTK_CONTAINER (widget)->border_width;
  
  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.x = widget->allocation.x + border_width;
  attributes.y = widget->allocation.y + border_width;
  attributes.width = widget->allocation.width - border_width * 2;
  attributes.height = widget->allocation.height - border_width * 2;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.visual = gtk_widget_get_visual (widget);
  attributes.colormap = gtk_widget_get_colormap (widget);
  attributes.event_mask = gtk_widget_get_events (widget);
  attributes.event_mask |= (GDK_EXPOSURE_MASK |
			    GDK_BUTTON_PRESS_MASK |
			    GDK_BUTTON_RELEASE_MASK |
			    GDK_KEY_RELEASE_MASK);
  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

  /* main window */
  widget->window = gdk_window_new (gtk_widget_get_parent_window (widget),
				   &attributes, attributes_mask);
  gdk_window_set_user_data (widget->window, unilist);

  widget->style = gtk_style_attach (widget->style, widget->window);

  gtk_style_set_background (widget->style, widget->window, GTK_STATE_NORMAL);

  /* column-title window */

  attributes.x = unilist->column_title_area.x;
  attributes.y = unilist->column_title_area.y;
  attributes.width = unilist->column_title_area.width;
  attributes.height = unilist->column_title_area.height;
  
  unilist->title_window = gdk_window_new (widget->window, &attributes,
					attributes_mask);
  gdk_window_set_user_data (unilist->title_window, unilist);

  gtk_style_set_background (widget->style, unilist->title_window,
			    GTK_STATE_NORMAL);
  gdk_window_show (unilist->title_window);

  /* set things up so column buttons are drawn in title window */
  for (i = 0; i < unilist->columns; i++)
    if (unilist->column[i].button)
      gtk_widget_set_parent_window (unilist->column[i].button,
				    unilist->title_window);

  /* unilist-window */
  attributes.x = (unilist->internal_allocation.x +
		  widget->style->xthickness);
  attributes.y = (unilist->internal_allocation.y +
		  widget->style->ythickness +
		  unilist->column_title_area.height);
  attributes.width = unilist->unilist_window_width;
  attributes.height = unilist->unilist_window_height;
  
  unilist->unilist_window = gdk_window_new (widget->window, &attributes,
					attributes_mask);
  gdk_window_set_user_data (unilist->unilist_window, unilist);

  gdk_window_set_background (unilist->unilist_window,
			     &widget->style->base[GTK_STATE_NORMAL]);
  gdk_window_show (unilist->unilist_window);
  gdk_window_get_size (unilist->unilist_window, &unilist->unilist_window_width,
		       &unilist->unilist_window_height);

  /* create resize windows */
  attributes.wclass = GDK_INPUT_ONLY;
  attributes.event_mask = (GDK_BUTTON_PRESS_MASK |
			   GDK_BUTTON_RELEASE_MASK |
			   GDK_POINTER_MOTION_MASK |
			   GDK_POINTER_MOTION_HINT_MASK);
  attributes_mask = GDK_WA_CURSOR;
  attributes.cursor = gdk_cursor_new (GDK_SB_H_DOUBLE_ARROW);
  unilist->cursor_drag = attributes.cursor;

  attributes.x =  LIST_WIDTH (unilist) + 1;
  attributes.y = 0;
  attributes.width = 0;
  attributes.height = 0;

  for (i = 0; i < unilist->columns; i++)
    {
      unilist->column[i].window = gdk_window_new (unilist->title_window,
						&attributes, attributes_mask);
      gdk_window_set_user_data (unilist->column[i].window, unilist);
    }

  /* This is slightly less efficient than creating them with the
   * right size to begin with, but easier
   */
  size_allocate_title_buttons (unilist);

  /* GCs */
  unilist->fg_gc = gdk_gc_new (widget->window);
  unilist->bg_gc = gdk_gc_new (widget->window);
  
  /* We'll use this gc to do scrolling as well */
  gdk_gc_set_exposures (unilist->fg_gc, TRUE);

  values.foreground = (widget->style->white.pixel==0 ?
		       widget->style->black:widget->style->white);
  values.function = GDK_XOR;
  values.subwindow_mode = GDK_INCLUDE_INFERIORS;
  unilist->xor_gc = gdk_gc_new_with_values (widget->window,
					  &values,
					  GDK_GC_FOREGROUND |
					  GDK_GC_FUNCTION |
					  GDK_GC_SUBWINDOW);

  /* attach optional row/cell styles, allocate foreground/background colors */
  list = unilist->row_list;
  for (i = 0; i < unilist->rows; i++)
    {
      unilist_row = list->data;
      list = list->next;

      if (unilist_row->style)
	unilist_row->style = gtk_style_attach (unilist_row->style,
					     unilist->unilist_window);

      if (unilist_row->fg_set || unilist_row->bg_set)
	{
	  GdkColormap *colormap;

	  colormap = gtk_widget_get_colormap (widget);
	  if (unilist_row->fg_set)
	    gdk_color_alloc (colormap, &unilist_row->foreground);
	  if (unilist_row->bg_set)
	    gdk_color_alloc (colormap, &unilist_row->background);
	}
      
      for (j = 0; j < unilist->columns; j++)
	if  (unilist_row->cell[j].style)
	  unilist_row->cell[j].style =
	    gtk_style_attach (unilist_row->cell[j].style, unilist->unilist_window);
    }
}

static void
gtk_unilist_unrealize (GtkWidget *widget)
{
  gint i;
  GtkUniList *unilist;

  g_return_if_fail (GTK_IS_UNILIST (widget));

  unilist = GTK_UNILIST (widget);

  /* freeze the list */
  unilist->freeze_count++;

  if (GTK_WIDGET_MAPPED (widget))
    gtk_unilist_unmap (widget);

  GTK_WIDGET_UNSET_FLAGS (widget, GTK_MAPPED);

  /* detach optional row/cell styles */
  if (GTK_WIDGET_REALIZED (widget))
    {
      GtkUniListRow *unilist_row;
      GList *list;
      gint j;

      list = unilist->row_list;
      for (i = 0; i < unilist->rows; i++)
	{
	  unilist_row = list->data;
	  list = list->next;

	  if (unilist_row->style)
	    gtk_style_detach (unilist_row->style);
	  for (j = 0; j < unilist->columns; j++)
	    if  (unilist_row->cell[j].style)
	      gtk_style_detach (unilist_row->cell[j].style);
	}
    }

  gdk_cursor_destroy (unilist->cursor_drag);
  gdk_gc_destroy (unilist->xor_gc);
  gdk_gc_destroy (unilist->fg_gc);
  gdk_gc_destroy (unilist->bg_gc);

  for (i = 0; i < unilist->columns; i++)
    {
      if (unilist->column[i].button)
	gtk_widget_unrealize (unilist->column[i].button);
      if (unilist->column[i].window)
	{
	  gdk_window_set_user_data (unilist->column[i].window, NULL);
	  gdk_window_destroy (unilist->column[i].window);
	  unilist->column[i].window = NULL;
	}
    }

  gdk_window_set_user_data (unilist->unilist_window, NULL);
  gdk_window_destroy (unilist->unilist_window);
  unilist->unilist_window = NULL;

  gdk_window_set_user_data (unilist->title_window, NULL);
  gdk_window_destroy (unilist->title_window);
  unilist->title_window = NULL;

  unilist->cursor_drag = NULL;
  unilist->xor_gc = NULL;
  unilist->fg_gc = NULL;
  unilist->bg_gc = NULL;

  if (GTK_WIDGET_CLASS (parent_class)->unrealize)
    (* GTK_WIDGET_CLASS (parent_class)->unrealize) (widget);
}

static void
gtk_unilist_map (GtkWidget *widget)
{
  gint i;
  GtkUniList *unilist;

  g_return_if_fail (GTK_IS_UNILIST (widget));

  unilist = GTK_UNILIST (widget);

  if (!GTK_WIDGET_MAPPED (widget))
    {
      GTK_WIDGET_SET_FLAGS (widget, GTK_MAPPED);

      /* map column buttons */
      for (i = 0; i < unilist->columns; i++)
	{
	  if (unilist->column[i].button &&
	      GTK_WIDGET_VISIBLE (unilist->column[i].button) &&
	      !GTK_WIDGET_MAPPED (unilist->column[i].button))
	    gtk_widget_map (unilist->column[i].button);
	}
      
      for (i = 0; i < unilist->columns; i++)
	if (unilist->column[i].window && unilist->column[i].button)
	  {
	    gdk_window_raise (unilist->column[i].window);
	    gdk_window_show (unilist->column[i].window);
	  }

      gdk_window_show (unilist->title_window);
      gdk_window_show (unilist->unilist_window);
      gdk_window_show (widget->window);

      /* unfreeze the list */
      unilist->freeze_count = 0;
    }
}

static void
gtk_unilist_unmap (GtkWidget *widget)
{
  gint i;
  GtkUniList *unilist;

  g_return_if_fail (GTK_IS_UNILIST (widget));

  unilist = GTK_UNILIST (widget);

  if (GTK_WIDGET_MAPPED (widget))
    {
      GTK_WIDGET_UNSET_FLAGS (widget, GTK_MAPPED);

      if (gdk_pointer_is_grabbed () && GTK_WIDGET_HAS_GRAB (unilist))
	{
	  remove_grab (unilist);

	  GTK_UNILIST_GET_CLASS (widget)->resync_selection (unilist, NULL);

	  unilist->click_cell.row = -1;
	  unilist->click_cell.column = -1;
	  unilist->drag_button = 0;

	  if (GTK_UNILIST_IN_DRAG(unilist))
	    {
	      gpointer drag_data;

	      GTK_UNILIST_UNSET_FLAG (unilist, UNILIST_IN_DRAG);
	      drag_data = gtk_object_get_data (GTK_OBJECT (unilist),
					       "gtk-site-data");
	      if (drag_data)
		gtk_signal_handler_unblock_by_data (GTK_OBJECT (unilist),
						    drag_data);
	    }
	}

      for (i = 0; i < unilist->columns; i++)
	if (unilist->column[i].window)
	  gdk_window_hide (unilist->column[i].window);

      gdk_window_hide (unilist->unilist_window);
      gdk_window_hide (unilist->title_window);
      gdk_window_hide (widget->window);

      /* unmap column buttons */
      for (i = 0; i < unilist->columns; i++)
	if (unilist->column[i].button &&
	    GTK_WIDGET_MAPPED (unilist->column[i].button))
	  gtk_widget_unmap (unilist->column[i].button);

      /* freeze the list */
      unilist->freeze_count++;
    }
}

static gint
gtk_unilist_expose (GtkWidget      *widget,
		  GdkEventExpose *event)
{
  GtkUniList *unilist;

  g_return_val_if_fail (GTK_IS_UNILIST (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  if (GTK_WIDGET_DRAWABLE (widget))
    {
      unilist = GTK_UNILIST (widget);

      /* draw border */
      if (event->window == widget->window)
	gtk_draw_shadow (widget->style, widget->window,
			 GTK_STATE_NORMAL, unilist->shadow_type,
			 0, 0,
			 unilist->unilist_window_width +
			 (2 * widget->style->xthickness),
			 unilist->unilist_window_height +
			 (2 * widget->style->ythickness) +
			 unilist->column_title_area.height);

      /* exposure events on the list */
      if (event->window == unilist->unilist_window)
	draw_rows (unilist, &event->area);

      if (event->window == unilist->unilist_window &&
	  unilist->drag_highlight_row >= 0)
	GTK_UNILIST_GET_CLASS (unilist)->draw_drag_highlight
	  (unilist, g_list_nth (unilist->row_list,
			      unilist->drag_highlight_row)->data,
	   unilist->drag_highlight_row, unilist->drag_highlight_pos);

      if (event->window == unilist->title_window)
	{
	  gint i;
	  
	  for (i = 0; i < unilist->columns; i++)
	    {
	      if (unilist->column[i].button)
		gtk_container_propagate_expose (GTK_CONTAINER (unilist),
						unilist->column[i].button,
						event);
	    }
	}
    }

  return FALSE;
}

static void
gtk_unilist_style_set (GtkWidget *widget,
		     GtkStyle  *previous_style)
{
  GtkUniList *unilist;

  g_return_if_fail (GTK_IS_UNILIST (widget));

  if (GTK_WIDGET_CLASS (parent_class)->style_set)
    (*GTK_WIDGET_CLASS (parent_class)->style_set) (widget, previous_style);

  unilist = GTK_UNILIST (widget);

  if (GTK_WIDGET_REALIZED (widget))
    {
      gtk_style_set_background (widget->style, widget->window, widget->state);
      gtk_style_set_background (widget->style, unilist->title_window, GTK_STATE_SELECTED);
      gdk_window_set_background (unilist->unilist_window, &widget->style->base[GTK_STATE_NORMAL]);
    }

  /* Fill in data after widget has correct style */

  /* text properties */
  if (!GTK_UNILIST_ROW_HEIGHT_SET(unilist))
    /* Reset unilist->row_height */
    gtk_unilist_set_row_height (unilist, 0);

  /* Column widths */
  if (!GTK_UNILIST_AUTO_RESIZE_BLOCKED (unilist))
    {
      gint width;
      gint i;

      for (i = 0; i < unilist->columns; i++)
	if (unilist->column[i].auto_resize)
	  {
	    width = gtk_unilist_optimal_column_width (unilist, i);
	    if (width != unilist->column[i].width)
	      gtk_unilist_set_column_width (unilist, i, width);
	  }
    }
}

static gint
gtk_unilist_button_press (GtkWidget      *widget,
			GdkEventButton *event)
{
  gint i;
  GtkUniList *unilist;
  gint x;
  gint y;
  gint row;
  gint column;
  gint button_actions;

  g_return_val_if_fail (GTK_IS_UNILIST (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  unilist = GTK_UNILIST (widget);

  button_actions = unilist->button_actions[event->button - 1];

  if (button_actions == GTK_BUTTON_IGNORED)
    return FALSE;

  /* selections on the list */
  if (event->window == unilist->unilist_window)
    {
      x = event->x;
      y = event->y;

      if (get_selection_info (unilist, x, y, &row, &column))
	{
	  gint old_row = unilist->focus_row;

	  if (unilist->focus_row == -1)
	    old_row = row;

	  if (event->type == GDK_BUTTON_PRESS)
	    {
	      GdkEventMask mask = ((1 << (4 + event->button)) |
				   GDK_POINTER_MOTION_HINT_MASK |
				   GDK_BUTTON_RELEASE_MASK);

	      if (gdk_pointer_grab (unilist->unilist_window, FALSE, mask,
				    NULL, NULL, event->time))
		return FALSE;
	      gtk_grab_add (widget);

	      unilist->click_cell.row = row;
	      unilist->click_cell.column = column;
	      unilist->drag_button = event->button;
	    }
	  else
	    {
	      unilist->click_cell.row = -1;
	      unilist->click_cell.column = -1;

	      unilist->drag_button = 0;
	      remove_grab (unilist);
	    }

	if (button_actions & GTK_BUTTON_DRAGS){
	  if (GTK_WIDGET_HAS_FOCUS(widget))
		{
		gtk_unilist_draw_focus (widget);
		gdk_gc_set_line_attributes (unilist->xor_gc, 1,
							  GDK_LINE_SOLID, 0, 0);
		unilist->focus_row = row;
		gtk_unilist_draw_focus (widget);
	  }else{
		gdk_gc_set_line_attributes (unilist->xor_gc, 1,
						  GDK_LINE_SOLID, 0, 0);
		unilist->focus_row = row;
	    gtk_widget_grab_focus (widget);
	  }
	  return FALSE;
	}
	  if (button_actions & GTK_BUTTON_SELECTS)
	    {
	      if (GTK_UNILIST_ADD_MODE(unilist))
		{
		  GTK_UNILIST_UNSET_FLAG (unilist, UNILIST_ADD_MODE);
		  if (GTK_WIDGET_HAS_FOCUS(widget))
		    {
		      gtk_unilist_draw_focus (widget);
		      gdk_gc_set_line_attributes (unilist->xor_gc, 1,
						  GDK_LINE_SOLID, 0, 0);
		      unilist->focus_row = row;
		      gtk_unilist_draw_focus (widget);
		    }
		  else
		    {
		      gdk_gc_set_line_attributes (unilist->xor_gc, 1,
						  GDK_LINE_SOLID, 0, 0);
		      unilist->focus_row = row;
		    }
		}
	      else if (row != unilist->focus_row)
		{
		  if (GTK_WIDGET_HAS_FOCUS(widget))
		    {
		      gtk_unilist_draw_focus (widget);
		      unilist->focus_row = row;
		      gtk_unilist_draw_focus (widget);
		    }
		  else
		    unilist->focus_row = row;
		}
	    }

	  if (!GTK_WIDGET_HAS_FOCUS(widget))
	    gtk_widget_grab_focus (widget);

	  if (button_actions & GTK_BUTTON_SELECTS)
	    {
	      switch (unilist->selection_mode)
		{
		case GTK_SELECTION_SINGLE:
		  if (event->type != GDK_BUTTON_PRESS)
		    {
		      gtk_signal_emit (GTK_OBJECT (unilist),
				       unilist_signals[SELECT_ROW],
				       row, column, event);
		      unilist->anchor = -1;
		    }
		  else
		    unilist->anchor = row;
		  break;
		case GTK_SELECTION_BROWSE:
		  gtk_signal_emit (GTK_OBJECT (unilist),
				   unilist_signals[SELECT_ROW],
				   row, column, event);
		  break;
		case GTK_SELECTION_MULTIPLE:
		  if (event->type != GDK_BUTTON_PRESS)
		    {
		      if (unilist->anchor != -1)
			{
			  update_extended_selection (unilist, unilist->focus_row);
			  GTK_UNILIST_GET_CLASS (unilist)->resync_selection
			    (unilist, (GdkEvent *) event);
			}
		      gtk_signal_emit (GTK_OBJECT (unilist),
				       unilist_signals[SELECT_ROW],
				       row, column, event);
		      break;
		    }
	      
		  if (event->state & GDK_CONTROL_MASK)
		    {
		      if (event->state & GDK_SHIFT_MASK)
			{
			  if (unilist->anchor < 0)
			    {
			      g_list_free (unilist->undo_selection);
			      g_list_free (unilist->undo_unselection);
			      unilist->undo_selection = NULL;
			      unilist->undo_unselection = NULL;
			      unilist->anchor = old_row;
			      unilist->drag_pos = old_row;
			      unilist->undo_anchor = old_row;
			    }
			  update_extended_selection (unilist, unilist->focus_row);
			}
		      else
			{
			  if (unilist->anchor == -1)
			    set_anchor (unilist, TRUE, row, old_row);
			  else
			    update_extended_selection (unilist,
						       unilist->focus_row);
			}
		      break;
		    }

		  if (event->state & GDK_SHIFT_MASK)
		    {
		      set_anchor (unilist, FALSE, old_row, old_row);
		      update_extended_selection (unilist, unilist->focus_row);
		      break;
		    }

		  if (unilist->anchor == -1)
		    set_anchor (unilist, FALSE, row, old_row);
		  else
		    update_extended_selection (unilist, unilist->focus_row);
		  break;
		default:
		  break;
		}
	    }

	}
      return TRUE;
    }

  /* press on resize windows */
  for (i = 0; i < unilist->columns; i++)
    if (unilist->column[i].resizeable && unilist->column[i].window &&
	event->window == unilist->column[i].window)
      {
	gpointer drag_data;

	if (gdk_pointer_grab (unilist->column[i].window, FALSE,
			      GDK_POINTER_MOTION_HINT_MASK |
			      GDK_BUTTON1_MOTION_MASK |
			      GDK_BUTTON_RELEASE_MASK,
			      NULL, NULL, event->time))
	  return FALSE;

	gtk_grab_add (widget);
	GTK_UNILIST_SET_FLAG (unilist, UNILIST_IN_DRAG);

	/* block attached dnd signal handler */
	drag_data = gtk_object_get_data (GTK_OBJECT (unilist), "gtk-site-data");
	if (drag_data)
	  gtk_signal_handler_block_by_data (GTK_OBJECT (unilist), drag_data);

	if (!GTK_WIDGET_HAS_FOCUS(widget))
	  gtk_widget_grab_focus (widget);

	unilist->drag_pos = i;
	unilist->x_drag = (COLUMN_LEFT_XPIXEL(unilist, i) + COLUMN_INSET +
			 unilist->column[i].area.width + CELL_SPACING);

	if (GTK_UNILIST_ADD_MODE(unilist))
	  gdk_gc_set_line_attributes (unilist->xor_gc, 1, GDK_LINE_SOLID, 0, 0);
	draw_xor_line (unilist);

        return TRUE;
      }

  return FALSE;
}

static gint
gtk_unilist_button_release (GtkWidget      *widget,
			  GdkEventButton *event)
{
  GtkUniList *unilist;
  gint button_actions;

  g_return_val_if_fail (GTK_IS_UNILIST (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  unilist = GTK_UNILIST (widget);

  button_actions = unilist->button_actions[event->button - 1];
  if (button_actions == GTK_BUTTON_IGNORED)
    return FALSE;

  /* release on resize windows */
  if (GTK_UNILIST_IN_DRAG(unilist))
    {
      gpointer drag_data;
      gint width;
      gint x;
      gint i;

      i = unilist->drag_pos;
      unilist->drag_pos = -1;

      /* unblock attached dnd signal handler */
      drag_data = gtk_object_get_data (GTK_OBJECT (unilist), "gtk-site-data");
      if (drag_data)
	gtk_signal_handler_unblock_by_data (GTK_OBJECT (unilist), drag_data);

      GTK_UNILIST_UNSET_FLAG (unilist, UNILIST_IN_DRAG);
      gtk_widget_get_pointer (widget, &x, NULL);
      gtk_grab_remove (widget);
      gdk_pointer_ungrab (event->time);

      if (unilist->x_drag >= 0)
	draw_xor_line (unilist);

      if (GTK_UNILIST_ADD_MODE(unilist))
	{
	  gdk_gc_set_line_attributes (unilist->xor_gc, 1,
				      GDK_LINE_ON_OFF_DASH, 0, 0);
	  gdk_gc_set_dashes (unilist->xor_gc, 0, "\4\4", 2);
	}

      width = new_column_width (unilist, i, &x);
      gtk_unilist_set_column_width (unilist, i, width);


      return TRUE;
    }

  if (unilist->drag_button == event->button)
    {
      gint row;
      gint column;

      unilist->drag_button = 0;
      unilist->click_cell.row = -1;
      unilist->click_cell.column = -1;

      remove_grab (unilist);

      if (button_actions & GTK_BUTTON_SELECTS)
	{
	  switch (unilist->selection_mode)
	    {
	    case GTK_SELECTION_MULTIPLE:
	      if (!(event->state & GDK_SHIFT_MASK) ||
		  !GTK_WIDGET_CAN_FOCUS (widget) ||
		  event->x < 0 || event->x >= unilist->unilist_window_width ||
		  event->y < 0 || event->y >= unilist->unilist_window_height)
		  end_selection(unilist);
//		GTK_UNILIST_GET_CLASS (unilist)->resync_selection
//		  (unilist, (GdkEvent *) event);
	      break;
	    case GTK_SELECTION_SINGLE:
	      if (get_selection_info (unilist, event->x, event->y,
				      &row, &column))
		{
		  if (row >= 0 && row < unilist->rows && unilist->anchor == row)
		    toggle_row (unilist, row, column, (GdkEvent *) event);
		}
	      unilist->anchor = -1;
	      break;
	    default:
	      break;
	    }
	}

      return TRUE;
    }
  
  return FALSE;
}

static gint
gtk_unilist_motion (GtkWidget      *widget,
		  GdkEventMotion *event)
{
  GtkUniList *unilist;
  gint x;
  gint y;
  gint row;
  gint new_width;
  gint button_actions = 0;

  g_return_val_if_fail (GTK_IS_UNILIST (widget), FALSE);

  unilist = GTK_UNILIST (widget);
  if (!(gdk_pointer_is_grabbed () && GTK_WIDGET_HAS_GRAB (unilist)))
    return FALSE;

  if (unilist->drag_button > 0)
    button_actions = unilist->button_actions[unilist->drag_button - 1];

  if (GTK_UNILIST_IN_DRAG(unilist))
    {
      if (event->is_hint || event->window != widget->window)
	gtk_widget_get_pointer (widget, &x, NULL);
      else
	x = event->x;
      
      new_width = new_column_width (unilist, unilist->drag_pos, &x);
      if (x != unilist->x_drag)
	{
	  /* x_drag < 0 indicates that the xor line is already invisible */
	  if (unilist->x_drag >= 0)
	    draw_xor_line (unilist);

	  unilist->x_drag = x;

	  if (unilist->x_drag >= 0)
	    draw_xor_line (unilist);
	}

      if (new_width <= MAX (COLUMN_MIN_WIDTH + 1,
			    unilist->column[unilist->drag_pos].min_width + 1))
	{
	  if (COLUMN_LEFT_XPIXEL (unilist, unilist->drag_pos) < 0 && x < 0)
	    gtk_unilist_moveto (unilist, -1, unilist->drag_pos, 0, 0);
	  return FALSE;
	}
      if (unilist->column[unilist->drag_pos].max_width >= COLUMN_MIN_WIDTH &&
	  new_width >= unilist->column[unilist->drag_pos].max_width)
	{
	  if (COLUMN_LEFT_XPIXEL (unilist, unilist->drag_pos) + new_width >
	      unilist->unilist_window_width && x < 0)
	    move_horizontal (unilist,
			     COLUMN_LEFT_XPIXEL (unilist, unilist->drag_pos) +
			     new_width - unilist->unilist_window_width +
			     COLUMN_INSET + CELL_SPACING);
	  return FALSE;
	}
    }

  if (event->is_hint || event->window != unilist->unilist_window)
    gdk_window_get_pointer (unilist->unilist_window, &x, &y, NULL);
  else
    {
      x = event->x;
      y = event->y;
    }

  if (GTK_UNILIST_REORDERABLE(unilist) && button_actions & GTK_BUTTON_DRAGS)
    {
      /* delayed drag start */
      if (event->window == unilist->unilist_window &&
	  unilist->click_cell.row >= 0 && unilist->click_cell.column >= 0 &&
	  (y < 0 || y >= unilist->unilist_window_height ||
	   x < 0 || x >= unilist->unilist_window_width  ||
	   y < ROW_TOP_YPIXEL (unilist, unilist->click_cell.row) ||
	   y >= (ROW_TOP_YPIXEL (unilist, unilist->click_cell.row) +
		 unilist->row_height) ||
	   x < COLUMN_LEFT_XPIXEL (unilist, unilist->click_cell.column) ||
	   x >= (COLUMN_LEFT_XPIXEL(unilist, unilist->click_cell.column) + 
		 unilist->column[unilist->click_cell.column].area.width)))
	{
	  GtkTargetList  *target_list;

	  target_list = gtk_target_list_new (&unilist_target_table, 1);
	  gtk_drag_begin (widget, target_list, GDK_ACTION_MOVE,
			  unilist->drag_button, (GdkEvent *)event);

	}
      return TRUE;
    }

  /* horizontal autoscrolling */
  if (unilist->hadjustment && LIST_WIDTH (unilist) > unilist->unilist_window_width &&
      (x < 0 || x >= unilist->unilist_window_width))
    {
      if (unilist->htimer)
	return FALSE;

      unilist->htimer = gtk_timeout_add
	(SCROLL_TIME, (GtkFunction) horizontal_timeout, unilist);

      if (!((x < 0 && unilist->hadjustment->value == 0) ||
	    (x >= unilist->unilist_window_width &&
	     unilist->hadjustment->value ==
	     LIST_WIDTH (unilist) - unilist->unilist_window_width)))
	{
	  if (x < 0)
	    move_horizontal (unilist, -1 + (x/2));
	  else
	    move_horizontal (unilist, 1 + (x - unilist->unilist_window_width) / 2);
	}
    }

  if (GTK_UNILIST_IN_DRAG(unilist))
    return FALSE;

  /* vertical autoscrolling */
  row = ROW_FROM_YPIXEL (unilist, y);

  /* don't scroll on last pixel row if it's a cell spacing */
  if (y == unilist->unilist_window_height - 1 &&
      y == ROW_TOP_YPIXEL (unilist, row-1) + unilist->row_height)
    return FALSE;

  if (LIST_HEIGHT (unilist) > unilist->unilist_window_height &&
      (y < 0 || y >= unilist->unilist_window_height))
    {
      if (unilist->vtimer)
	return FALSE;

      unilist->vtimer = gtk_timeout_add (SCROLL_TIME,
				       (GtkFunction) vertical_timeout, unilist);

      if (unilist->drag_button &&
	  ((y < 0 && unilist->focus_row == 0) ||
	   (y >= unilist->unilist_window_height &&
	    unilist->focus_row == unilist->rows - 1)))
	return FALSE;
    }

  row = CLAMP (row, 0, unilist->rows - 1);

  if (button_actions & GTK_BUTTON_SELECTS &
      !gtk_object_get_data (GTK_OBJECT (widget), "gtk-site-data"))
    {
      if (row == unilist->focus_row)
	return FALSE;

      gtk_unilist_draw_focus (widget);
      unilist->focus_row = row;
      gtk_unilist_draw_focus (widget);

      switch (unilist->selection_mode)
	{
	case GTK_SELECTION_BROWSE:
	  gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[SELECT_ROW],
			   unilist->focus_row, -1, event);
	  break;
	case GTK_SELECTION_MULTIPLE:
	  update_extended_selection (unilist, unilist->focus_row);
	  break;
	default:
	  break;
	}
    }
  
  if (ROW_TOP_YPIXEL(unilist, row) < 0)
    move_vertical (unilist, row, 0);
  else if (ROW_TOP_YPIXEL(unilist, row) + unilist->row_height >
	   unilist->unilist_window_height)
    move_vertical (unilist, row, 1);

  return FALSE;
}

static void
gtk_unilist_size_request (GtkWidget      *widget,
			GtkRequisition *requisition)
{
  GtkUniList *unilist;
  gint i;

  g_return_if_fail (GTK_IS_UNILIST (widget));
  g_return_if_fail (requisition != NULL);

  unilist = GTK_UNILIST (widget);

  requisition->width = 0;
  requisition->height = 0;

  /* compute the size of the column title (title) area */
  unilist->column_title_area.height = 0;
  if (GTK_UNILIST_SHOW_TITLES(unilist))
    for (i = 0; i < unilist->columns; i++)
      if (unilist->column[i].button)
	{
	  GtkRequisition child_requisition;
	  
	  gtk_widget_size_request (unilist->column[i].button,
				   &child_requisition);
	  unilist->column_title_area.height =
	    MAX (unilist->column_title_area.height,
		 child_requisition.height);
	}
  
  requisition->width += (widget->style->xthickness +
			 GTK_CONTAINER (widget)->border_width) * 2;
  requisition->height += (unilist->column_title_area.height +
			  (widget->style->ythickness +
			   GTK_CONTAINER (widget)->border_width) * 2);

  /* if (!unilist->hadjustment) */
  requisition->width += list_requisition_width (unilist);
  /* if (!unilist->vadjustment) */
  requisition->height += LIST_HEIGHT (unilist);
}

static void
gtk_unilist_size_allocate (GtkWidget     *widget,
			 GtkAllocation *allocation)
{
  GtkUniList *unilist;
  GtkAllocation unilist_allocation;
  gint border_width;

  g_return_if_fail (GTK_IS_UNILIST (widget));
  g_return_if_fail (allocation != NULL);

  unilist = GTK_UNILIST (widget);
  widget->allocation = *allocation;
  border_width = GTK_CONTAINER (widget)->border_width;

  if (GTK_WIDGET_REALIZED (widget))
    {
      gdk_window_move_resize (widget->window,
			      allocation->x + border_width,
			      allocation->y + border_width,
			      allocation->width - border_width * 2,
			      allocation->height - border_width * 2);
    }

  /* use internal allocation structure for all the math
   * because it's easier than always subtracting the container
   * border width */
  unilist->internal_allocation.x = 0;
  unilist->internal_allocation.y = 0;
  unilist->internal_allocation.width = MAX (1, (gint)allocation->width -
					  border_width * 2);
  unilist->internal_allocation.height = MAX (1, (gint)allocation->height -
					   border_width * 2);
	
  /* allocate unilist window assuming no scrollbars */
  unilist_allocation.x = (unilist->internal_allocation.x +
			widget->style->xthickness);
  unilist_allocation.y = (unilist->internal_allocation.y +
			widget->style->ythickness +
			unilist->column_title_area.height);
  unilist_allocation.width = MAX (1, (gint)unilist->internal_allocation.width - 
				(2 * (gint)widget->style->xthickness));
  unilist_allocation.height = MAX (1, (gint)unilist->internal_allocation.height -
				 (2 * (gint)widget->style->ythickness) -
				 (gint)unilist->column_title_area.height);
  
  unilist->unilist_window_width = unilist_allocation.width;
  unilist->unilist_window_height = unilist_allocation.height;
  
  if (GTK_WIDGET_REALIZED (widget))
    {
      gdk_window_move_resize (unilist->unilist_window,
			      unilist_allocation.x,
			      unilist_allocation.y,
			      unilist_allocation.width,
			      unilist_allocation.height);
    }
  
  /* position the window which holds the column title buttons */
  unilist->column_title_area.x = widget->style->xthickness;
  unilist->column_title_area.y = widget->style->ythickness;
  unilist->column_title_area.width = unilist_allocation.width;
  
  if (GTK_WIDGET_REALIZED (widget))
    {
      gdk_window_move_resize (unilist->title_window,
			      unilist->column_title_area.x,
			      unilist->column_title_area.y,
			      unilist->column_title_area.width,
			      unilist->column_title_area.height);
    }
  
  /* column button allocation */
  size_allocate_columns (unilist, FALSE);
  size_allocate_title_buttons (unilist);

  adjust_adjustments (unilist, TRUE);
}

/* GTKCONTAINER
 *   gtk_unilist_forall
 */
static void
gtk_unilist_forall (GtkContainer *container,
		  gboolean      include_internals,
		  GtkCallback   callback,
		  gpointer      callback_data)
{
  GtkUniList *unilist;
  guint i;

  g_return_if_fail (GTK_IS_UNILIST (container));
  g_return_if_fail (callback != NULL);

  if (!include_internals)
    return;

  unilist = GTK_UNILIST (container);
      
  /* callback for the column buttons */
  for (i = 0; i < unilist->columns; i++)
    if (unilist->column[i].button)
      (*callback) (unilist->column[i].button, callback_data);
}

/* PRIVATE DRAWING FUNCTIONS
 *   get_cell_style
 *   draw_cell_pixmap
 *   draw_row
 *   draw_rows
 *   draw_xor_line
 *   unilist_refresh
 */
static void
get_cell_style (GtkUniList     *unilist,
		GtkUniListRow  *unilist_row,
		gint          state,
		gint          column,
		GtkStyle    **style,
		GdkGC       **fg_gc,
		GdkGC       **bg_gc)
{
  gint fg_state;

  if ((state == GTK_STATE_NORMAL) &&
      (GTK_WIDGET (unilist)->state == GTK_STATE_INSENSITIVE))
    fg_state = GTK_STATE_INSENSITIVE;
  else
    fg_state = state;

  if (unilist_row->cell[column].style)
    {
      if (style)
	*style = unilist_row->cell[column].style;
      if (fg_gc)
	*fg_gc = unilist_row->cell[column].style->fg_gc[fg_state];
      if (bg_gc) {
	if (state == GTK_STATE_SELECTED)
	  *bg_gc = unilist_row->cell[column].style->bg_gc[state];
	else
	  *bg_gc = unilist_row->cell[column].style->base_gc[state];
      }
    }
  else if (unilist_row->style)
    {
      if (style)
	*style = unilist_row->style;
      if (fg_gc)
	*fg_gc = unilist_row->style->fg_gc[fg_state];
      if (bg_gc) {
	if (state == GTK_STATE_SELECTED)
	  *bg_gc = unilist_row->style->bg_gc[state];
	else
	  *bg_gc = unilist_row->style->base_gc[state];
      }
    }
  else
    {
      if (style)
	*style = GTK_WIDGET (unilist)->style;
      if (fg_gc)
	*fg_gc = GTK_WIDGET (unilist)->style->fg_gc[fg_state];
      if (bg_gc) {
	if (state == GTK_STATE_SELECTED)
	  *bg_gc = GTK_WIDGET (unilist)->style->bg_gc[state];
	else
	  *bg_gc = GTK_WIDGET (unilist)->style->base_gc[state];
      }

      if (state != GTK_STATE_SELECTED)
	{
	  if (fg_gc && unilist_row->fg_set)
	    *fg_gc = unilist->fg_gc;
	  if (bg_gc && unilist_row->bg_set)
	    *bg_gc = unilist->bg_gc;
	}
    }
}

static gint
draw_cell_pixmap (GdkWindow    *window,
		  GdkRectangle *clip_rectangle,
		  GdkGC        *fg_gc,
		  GdkPixmap    *pixmap,
		  GdkBitmap    *mask,
		  gint          x,
		  gint          y,
		  gint          width,
		  gint          height)
{
  gint xsrc = 0;
  gint ysrc = 0;

  if (mask)
    {
      gdk_gc_set_clip_mask (fg_gc, mask);
      gdk_gc_set_clip_origin (fg_gc, x, y);
    }

  if (x < clip_rectangle->x)
    {
      xsrc = clip_rectangle->x - x;
      width -= xsrc;
      x = clip_rectangle->x;
    }
  if (x + width > clip_rectangle->x + clip_rectangle->width)
    width = clip_rectangle->x + clip_rectangle->width - x;

  if (y < clip_rectangle->y)
    {
      ysrc = clip_rectangle->y - y;
      height -= ysrc;
      y = clip_rectangle->y;
    }
  if (y + height > clip_rectangle->y + clip_rectangle->height)
    height = clip_rectangle->y + clip_rectangle->height - y;

  gdk_draw_pixmap (window, fg_gc, pixmap, xsrc, ysrc, x, y, width, height);
  gdk_gc_set_clip_origin (fg_gc, 0, 0);
  if (mask)
    gdk_gc_set_clip_mask (fg_gc, NULL);

  return x + MAX (width, 0);
}

static void
draw_row (GtkUniList     *unilist,
	  GdkRectangle *area,
	  gint          row,
	  GtkUniListRow  *unilist_row)
{
  GtkWidget *widget;
  GdkRectangle *rect;
  GdkRectangle row_rectangle;
  GdkRectangle cell_rectangle;
  GdkRectangle clip_rectangle;
  GdkRectangle intersect_rectangle;
  gint last_column;
  gint state;
  gint i;

  g_return_if_fail (unilist != NULL);

  /* bail now if we arn't drawable yet */
  if (!GTK_WIDGET_DRAWABLE (unilist) || row < 0 || row >= unilist->rows)
    return;

  widget = GTK_WIDGET (unilist);

  /* if the function is passed the pointer to the row instead of null,
   * it avoids this expensive lookup */
  if (!unilist_row)
    unilist_row = ROW_ELEMENT (unilist, row)->data;

  /* rectangle of the entire row */
  row_rectangle.x = 0;
  row_rectangle.y = ROW_TOP_YPIXEL (unilist, row);
  row_rectangle.width = unilist->unilist_window_width;
  row_rectangle.height = unilist->row_height;

  /* rectangle of the cell spacing above the row */
  cell_rectangle.x = 0;
  cell_rectangle.y = row_rectangle.y - CELL_SPACING;
  cell_rectangle.width = row_rectangle.width;
  cell_rectangle.height = CELL_SPACING;

  /* rectangle used to clip drawing operations, its y and height
   * positions only need to be set once, so we set them once here. 
   * the x and width are set withing the drawing loop below once per
   * column */
  clip_rectangle.y = row_rectangle.y;
  clip_rectangle.height = row_rectangle.height;

  if (unilist_row->state == GTK_STATE_NORMAL)
    {
      if (unilist_row->fg_set)
	gdk_gc_set_foreground (unilist->fg_gc, &unilist_row->foreground);
      if (unilist_row->bg_set)
	gdk_gc_set_foreground (unilist->bg_gc, &unilist_row->background);
    }

  state = unilist_row->state;

  /* draw the cell borders and background */
  if (area)
    {
      rect = &intersect_rectangle;
      if (gdk_rectangle_intersect (area, &cell_rectangle,
				   &intersect_rectangle))
	gdk_draw_rectangle (unilist->unilist_window,
			    widget->style->base_gc[GTK_STATE_NORMAL],
			    TRUE,
			    intersect_rectangle.x,
			    intersect_rectangle.y,
			    intersect_rectangle.width,
			    intersect_rectangle.height);

      /* the last row has to clear its bottom cell spacing too */
      if (unilist_row == unilist->row_list_end->data)
	{
	  cell_rectangle.y += unilist->row_height + CELL_SPACING;

	  if (gdk_rectangle_intersect (area, &cell_rectangle,
				       &intersect_rectangle))
	    gdk_draw_rectangle (unilist->unilist_window,
				widget->style->base_gc[GTK_STATE_NORMAL],
				TRUE,
				intersect_rectangle.x,
				intersect_rectangle.y,
				intersect_rectangle.width,
				intersect_rectangle.height);
	}

      if (!gdk_rectangle_intersect (area, &row_rectangle,&intersect_rectangle))
	return;

    }
  else
    {
      rect = &clip_rectangle;
      gdk_draw_rectangle (unilist->unilist_window,
			  widget->style->base_gc[GTK_STATE_NORMAL],
			  TRUE,
			  cell_rectangle.x,
			  cell_rectangle.y,
			  cell_rectangle.width,
			  cell_rectangle.height);

      /* the last row has to clear its bottom cell spacing too */
      if (unilist_row == unilist->row_list_end->data)
	{
	  cell_rectangle.y += unilist->row_height + CELL_SPACING;

	  gdk_draw_rectangle (unilist->unilist_window,
			      widget->style->base_gc[GTK_STATE_NORMAL],
			      TRUE,
			      cell_rectangle.x,
			      cell_rectangle.y,
			      cell_rectangle.width,
			      cell_rectangle.height);     
	}	  
    }
  
  for (last_column = unilist->columns - 1;
       last_column >= 0 && !unilist->column[last_column].visible; last_column--)
    ;

  /* iterate and draw all the columns (row cells) and draw their contents */
  for (i = 0; i < unilist->columns; i++)
    {
      GtkStyle *style;
      GdkGC *fg_gc;
      GdkGC *bg_gc;
      PangoLayout *layout;
      PangoRectangle logical_rect;

      gint width;
      gint height;
      gint pixmap_width;
      gint offset = 0;

      if (!unilist->column[i].visible)
	continue;

      get_cell_style (unilist, unilist_row, state, i, &style, &fg_gc, &bg_gc);

      clip_rectangle.x = unilist->column[i].area.x + unilist->hoffset;
      clip_rectangle.width = unilist->column[i].area.width;

      /* calculate clipping region clipping region */
      clip_rectangle.x -= COLUMN_INSET + CELL_SPACING;
      clip_rectangle.width += (2 * COLUMN_INSET + CELL_SPACING +
			       (i == last_column) * CELL_SPACING);
      
      if (area && !gdk_rectangle_intersect (area, &clip_rectangle,
					    &intersect_rectangle))
	continue;

      gdk_draw_rectangle (unilist->unilist_window, bg_gc, TRUE,
			  rect->x, rect->y, rect->width, rect->height);

      clip_rectangle.x += COLUMN_INSET + CELL_SPACING;
      clip_rectangle.width -= (2 * COLUMN_INSET + CELL_SPACING +
			       (i == last_column) * CELL_SPACING);


      /* calculate real width for column justification */
      
      layout = _gtk_unilist_create_cell_layout (unilist, unilist_row, i);
      if (layout)
	{
	  pango_layout_get_pixel_extents (layout, NULL, &logical_rect);
	  width = logical_rect.width;
	}
      else
	width = 0;

      pixmap_width = 0;
      offset = 0;
      switch (unilist_row->cell[i].type)
	{
	case GTK_CELL_PIXMAP:
	  gdk_window_get_size (GTK_CELL_PIXMAP (unilist_row->cell[i])->pixmap,
			       &pixmap_width, &height);
	  width += pixmap_width;
	  break;
	case GTK_CELL_PIXTEXT:
	  gdk_window_get_size (GTK_CELL_PIXTEXT (unilist_row->cell[i])->pixmap,
			       &pixmap_width, &height);
	  width += pixmap_width + GTK_CELL_PIXTEXT (unilist_row->cell[i])->spacing;
	  break;
	default:
	  break;
	}

      switch (unilist->column[i].justification)
	{
	case GTK_JUSTIFY_LEFT:
	  offset = clip_rectangle.x + unilist_row->cell[i].horizontal;
	  break;
	case GTK_JUSTIFY_RIGHT:
	  offset = (clip_rectangle.x + unilist_row->cell[i].horizontal +
		    clip_rectangle.width - width);
	  break;
	case GTK_JUSTIFY_CENTER:
	case GTK_JUSTIFY_FILL:
	  offset = (clip_rectangle.x + unilist_row->cell[i].horizontal +
		    (clip_rectangle.width / 2) - (width / 2));
	  break;
	};

      /* Draw Text and/or Pixmap */
      switch (unilist_row->cell[i].type)
	{
	case GTK_CELL_PIXMAP:
	  draw_cell_pixmap (unilist->unilist_window, &clip_rectangle, fg_gc,
			    GTK_CELL_PIXMAP (unilist_row->cell[i])->pixmap,
			    GTK_CELL_PIXMAP (unilist_row->cell[i])->mask,
			    offset,
			    clip_rectangle.y + unilist_row->cell[i].vertical +
			    (clip_rectangle.height - height) / 2,
			    pixmap_width, height);
	  break;
	case GTK_CELL_PIXTEXT:
	  offset =
	    draw_cell_pixmap (unilist->unilist_window, &clip_rectangle, fg_gc,
			      GTK_CELL_PIXTEXT (unilist_row->cell[i])->pixmap,
			      GTK_CELL_PIXTEXT (unilist_row->cell[i])->mask,
			      offset,
			      clip_rectangle.y + unilist_row->cell[i].vertical+
			      (clip_rectangle.height - height) / 2,
			      pixmap_width, height);
	  offset += GTK_CELL_PIXTEXT (unilist_row->cell[i])->spacing;

	  /* Fall through */
	case GTK_CELL_TEXT:
	  if (layout)
	    {
	      gint row_center_offset = (unilist->row_height - logical_rect.height - 1) / 2;

	      gdk_gc_set_clip_rectangle (fg_gc, &clip_rectangle);
	      gdk_draw_layout (unilist->unilist_window, fg_gc,
			       offset,
			       row_rectangle.y + row_center_offset + unilist_row->cell[i].vertical,
			       layout);
              g_object_unref (G_OBJECT (layout));
	      gdk_gc_set_clip_rectangle (fg_gc, NULL);
	    }
	  break;
	default:
	  break;
	}
    }

  /* draw focus rectangle */
  if (unilist->focus_row == row &&
      GTK_WIDGET_CAN_FOCUS (widget) && GTK_WIDGET_HAS_FOCUS(widget))
    {
      if (!area)
	gdk_draw_rectangle (unilist->unilist_window, unilist->xor_gc, FALSE,
			    row_rectangle.x, row_rectangle.y,
			    row_rectangle.width - 1, row_rectangle.height - 1);
      else if (gdk_rectangle_intersect (area, &row_rectangle,
					&intersect_rectangle))
	{
	  gdk_gc_set_clip_rectangle (unilist->xor_gc, &intersect_rectangle);
	  gdk_draw_rectangle (unilist->unilist_window, unilist->xor_gc, FALSE,
			      row_rectangle.x, row_rectangle.y,
			      row_rectangle.width - 1,
			      row_rectangle.height - 1);
	  gdk_gc_set_clip_rectangle (unilist->xor_gc, NULL);
	}
    }
}

static void
draw_rows (GtkUniList     *unilist,
	   GdkRectangle *area)
{
  GList *list;
  GtkUniListRow *unilist_row;
  gint i;
  gint first_row;
  gint last_row;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (unilist->row_height == 0 ||
      !GTK_WIDGET_DRAWABLE (unilist))
    return;

  if (area)
    {
      first_row = ROW_FROM_YPIXEL (unilist, area->y);
      last_row = ROW_FROM_YPIXEL (unilist, area->y + area->height);
    }
  else
    {
      first_row = ROW_FROM_YPIXEL (unilist, 0);
      last_row = ROW_FROM_YPIXEL (unilist, unilist->unilist_window_height);
    }

  /* this is a small special case which exposes the bottom cell line
   * on the last row -- it might go away if I change the wall the cell
   * spacings are drawn
   */
  if (unilist->rows == first_row)
    first_row--;

  list = ROW_ELEMENT (unilist, first_row);
  i = first_row;
  while (list)
    {
      unilist_row = list->data;
      list = list->next;

      if (i > last_row)
	return;

      GTK_UNILIST_GET_CLASS (unilist)->draw_row (unilist, area, i, unilist_row);
      i++;
    }

  if (!area)
    gdk_window_clear_area (unilist->unilist_window, 0,
			   ROW_TOP_YPIXEL (unilist, i), 0, 0);
}

static void                          
draw_xor_line (GtkUniList *unilist)
{
  GtkWidget *widget;

  g_return_if_fail (unilist != NULL);

  widget = GTK_WIDGET (unilist);

  gdk_draw_line (widget->window, unilist->xor_gc,
                 unilist->x_drag,
		 widget->style->ythickness,
                 unilist->x_drag,
                 unilist->column_title_area.height +
		 unilist->unilist_window_height + 1);
}

static void
unilist_refresh (GtkUniList *unilist)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));
  
  if (UNILIST_UNFROZEN (unilist))
    { 
      adjust_adjustments (unilist, FALSE);
      draw_rows (unilist, NULL);
    }
}

/* get cell from coordinates
 *   get_selection_info
 *   gtk_unilist_get_selection_info
 */
static gint
get_selection_info (GtkUniList *unilist,
		    gint      x,
		    gint      y,
		    gint     *row,
		    gint     *column)
{
  gint trow, tcol;

  g_return_val_if_fail (GTK_IS_UNILIST (unilist), 0);

  /* bounds checking, return false if the user clicked 
   * on a blank area */
  trow = ROW_FROM_YPIXEL (unilist, y);
  if (trow >= unilist->rows)
    return 0;

  if (row)
    *row = trow;

  tcol = COLUMN_FROM_XPIXEL (unilist, x);
  if (tcol >= unilist->columns)
    return 0;

  if (column)
    *column = tcol;

  return 1;
}

gint
gtk_unilist_get_selection_info (GtkUniList *unilist, 
			      gint      x, 
			      gint      y, 
			      gint     *row, 
			      gint     *column)
{
  g_return_val_if_fail (GTK_IS_UNILIST (unilist), 0);
  return get_selection_info (unilist, x, y, row, column);
}

/* PRIVATE ADJUSTMENT FUNCTIONS
 *   adjust_adjustments
 *   vadjustment_changed
 *   hadjustment_changed
 *   vadjustment_value_changed
 *   hadjustment_value_changed 
 *   check_exposures
 */
static void
adjust_adjustments (GtkUniList *unilist,
		    gboolean  block_resize)
{
  if (unilist->vadjustment)
    {
      unilist->vadjustment->page_size = unilist->unilist_window_height;
      unilist->vadjustment->step_increment = unilist->row_height;
      unilist->vadjustment->page_increment =
	MAX (unilist->vadjustment->page_size - unilist->vadjustment->step_increment,
	     unilist->vadjustment->page_size / 2);
      unilist->vadjustment->lower = 0;
      unilist->vadjustment->upper = LIST_HEIGHT (unilist);

      if (unilist->unilist_window_height - unilist->voffset > LIST_HEIGHT (unilist) ||
	  (unilist->voffset + (gint)unilist->vadjustment->value) != 0)
	{
	  unilist->vadjustment->value = MAX (0, (LIST_HEIGHT (unilist) -
					       unilist->unilist_window_height));
	  gtk_signal_emit_by_name (GTK_OBJECT (unilist->vadjustment),
				   "value_changed");
	}
      gtk_signal_emit_by_name (GTK_OBJECT (unilist->vadjustment), "changed");
    }

  if (unilist->hadjustment)
    {
      unilist->hadjustment->page_size = unilist->unilist_window_width;
      unilist->hadjustment->step_increment = 10;
      unilist->vadjustment->page_increment =
	MAX (unilist->vadjustment->page_size - unilist->vadjustment->step_increment,
	     unilist->vadjustment->page_size / 2);
      unilist->hadjustment->lower = 0;
      unilist->hadjustment->upper = LIST_WIDTH (unilist);

      if (unilist->unilist_window_width - unilist->hoffset > LIST_WIDTH (unilist) ||
	  (unilist->hoffset + (gint)unilist->hadjustment->value) != 0)
	{
	  unilist->hadjustment->value = MAX (0, (LIST_WIDTH (unilist) -
					       unilist->unilist_window_width));
	  gtk_signal_emit_by_name (GTK_OBJECT (unilist->hadjustment),
				   "value_changed");
	}
      gtk_signal_emit_by_name (GTK_OBJECT (unilist->hadjustment), "changed");
    }

  if (!block_resize && (!unilist->vadjustment || !unilist->hadjustment))
    {
      GtkWidget *widget;
      GtkRequisition requisition;

      widget = GTK_WIDGET (unilist);
      gtk_widget_size_request (widget, &requisition);

      if ((!unilist->hadjustment &&
	   requisition.width != widget->allocation.width) ||
	  (!unilist->vadjustment &&
	   requisition.height != widget->allocation.height))
	gtk_widget_queue_resize (widget);
    }
}

static void
vadjustment_changed (GtkAdjustment *adjustment,
		     gpointer       data)
{
  GtkUniList *unilist;

  g_return_if_fail (adjustment != NULL);
  g_return_if_fail (data != NULL);

  unilist = GTK_UNILIST (data);
}

static void
hadjustment_changed (GtkAdjustment *adjustment,
		     gpointer       data)
{
  GtkUniList *unilist;

  g_return_if_fail (adjustment != NULL);
  g_return_if_fail (data != NULL);

  unilist = GTK_UNILIST (data);
}

static void
vadjustment_value_changed (GtkAdjustment *adjustment,
			   gpointer       data)
{
  GtkUniList *unilist;
  gint dy, value;

  g_return_if_fail (adjustment != NULL);
  g_return_if_fail (GTK_IS_UNILIST (data));

  unilist = GTK_UNILIST (data);

  if (adjustment != unilist->vadjustment)
    return;

  value = -adjustment->value;
  dy = value - unilist->voffset;
  unilist->voffset = value;

  if (GTK_WIDGET_DRAWABLE (unilist))
    {
      gdk_window_scroll (unilist->unilist_window, 0, dy);
      gdk_window_process_updates (unilist->unilist_window, FALSE);
    }
  
  return;
}

typedef struct
{
  GdkWindow *window;
  gint dx;
} ScrollData;

/* The window to which widget->window is relative */
#define ALLOCATION_WINDOW(widget)		\
   (GTK_WIDGET_NO_WINDOW (widget) ?		\
    (widget)->window :                          \
     gdk_window_get_parent ((widget)->window))

static void
adjust_allocation_recurse (GtkWidget *widget,
			   gpointer   data)
{
  ScrollData *scroll_data = data;
  
  if (!GTK_WIDGET_REALIZED (widget))
    {
      if (GTK_WIDGET_VISIBLE (widget))
	{
	  GdkRectangle tmp_rectangle = widget->allocation;
	  tmp_rectangle.x += scroll_data->dx;
      
	  gtk_widget_size_allocate (widget, &tmp_rectangle);
	}
    }
  else
    {
      if (ALLOCATION_WINDOW (widget) == scroll_data->window)
	{
	  widget->allocation.x += scroll_data->dx;

	  if (GTK_IS_CONTAINER (widget))
	    gtk_container_forall (GTK_CONTAINER (widget),
				  adjust_allocation_recurse,
				  data);
	}
    }
}

static void
adjust_allocation (GtkWidget *widget,
		   gint       dx)
{
  ScrollData scroll_data;

  if (GTK_WIDGET_REALIZED (widget))
    scroll_data.window = ALLOCATION_WINDOW (widget);
  else
    scroll_data.window = NULL;
    
  scroll_data.dx = dx;
  
  adjust_allocation_recurse (widget, &scroll_data);
}

static void
hadjustment_value_changed (GtkAdjustment *adjustment,
			   gpointer       data)
{
  GtkUniList *unilist;
  GtkContainer *container;
  GdkRectangle area;
  gint i;
  gint y = 0;
  gint value;
  gint dx;

  g_return_if_fail (adjustment != NULL);
  g_return_if_fail (GTK_IS_UNILIST (data));

  unilist = GTK_UNILIST (data);
  container = GTK_CONTAINER (data);

  if (adjustment != unilist->hadjustment)
    return;

  value = adjustment->value;

  dx = -value - unilist->hoffset;

  if (GTK_WIDGET_REALIZED (unilist))
    gdk_window_scroll (unilist->title_window, dx, 0);

  /* adjust the column button's allocations */
  for (i = 0; i < unilist->columns; i++)
    if (unilist->column[i].button)
      adjust_allocation (unilist->column[i].button, dx);

  unilist->hoffset = -value;

  if (GTK_WIDGET_DRAWABLE (unilist))
    {
      if (GTK_WIDGET_CAN_FOCUS(unilist) && GTK_WIDGET_HAS_FOCUS(unilist) &&
          !container->focus_child && GTK_UNILIST_ADD_MODE(unilist))
        {
          y = ROW_TOP_YPIXEL (unilist, unilist->focus_row);
      
          gdk_draw_rectangle (unilist->unilist_window, unilist->xor_gc, FALSE, 0, y,
                              unilist->unilist_window_width - 1,
                              unilist->row_height - 1);
        }
 
      gdk_window_scroll (unilist->unilist_window, dx, 0);
      gdk_window_process_updates (unilist->unilist_window, FALSE);

      if (GTK_WIDGET_CAN_FOCUS(unilist) && GTK_WIDGET_HAS_FOCUS(unilist) &&
          !container->focus_child)
        {
          if (GTK_UNILIST_ADD_MODE(unilist))
            {
              gint focus_row;
	  
              focus_row = unilist->focus_row;
              unilist->focus_row = -1;
              draw_rows (unilist, &area);
              unilist->focus_row = focus_row;
	  
              gdk_draw_rectangle (unilist->unilist_window, unilist->xor_gc,
                                  FALSE, 0, y, unilist->unilist_window_width - 1,
                                  unilist->row_height - 1);
              return;
            }
          else if (ABS(dx) < unilist->unilist_window_width - 1)
            {
              gint x0;
              gint x1;
	  
              if (dx > 0)
                {
                  x0 = unilist->unilist_window_width - 1;
                  x1 = dx;
                }
              else
                {
                  x0 = 0;
                  x1 = unilist->unilist_window_width - 1 + dx;
                }

              y = ROW_TOP_YPIXEL (unilist, unilist->focus_row);
              gdk_draw_line (unilist->unilist_window, unilist->xor_gc,
                             x0, y + 1, x0, y + unilist->row_height - 2);
              gdk_draw_line (unilist->unilist_window, unilist->xor_gc,
                             x1, y + 1, x1, y + unilist->row_height - 2);
            }
        }
    }
}

/* PRIVATE 
 * Memory Allocation/Distruction Routines for GtkUniList stuctures
 *
 * functions:
 *   columns_new
 *   column_title_new
 *   columns_delete
 *   row_new
 *   row_delete
 */
static GtkUniListColumn *
columns_new (GtkUniList *unilist)
{
  GtkUniListColumn *column;
  gint i;

  column = g_new (GtkUniListColumn, unilist->columns);

  for (i = 0; i < unilist->columns; i++)
    {
      column[i].area.x = 0;
      column[i].area.y = 0;
      column[i].area.width = 0;
      column[i].area.height = 0;
      column[i].title = NULL;
      column[i].button = NULL;
      column[i].window = NULL;
      column[i].width = 0;
      column[i].min_width = -1;
      column[i].max_width = -1;
      column[i].visible = TRUE;
      column[i].width_set = FALSE;
      column[i].resizeable = TRUE;
      column[i].auto_resize = FALSE;
      column[i].button_passive = FALSE;
      column[i].justification = GTK_JUSTIFY_LEFT;
    }

  return column;
}

static void
column_title_new (GtkUniList    *unilist,
		  gint         column,
		  const gchar *title)
{
  if (unilist->column[column].title)
    g_free (unilist->column[column].title);

  unilist->column[column].title = g_strdup (title);
}

static void
columns_delete (GtkUniList *unilist)
{
  gint i;

  for (i = 0; i < unilist->columns; i++)
    if (unilist->column[i].title)
      g_free (unilist->column[i].title);
      
  g_free (unilist->column);
}

static GtkUniListRow *
row_new (GtkUniList *unilist)
{
  int i;
  GtkUniListRow *unilist_row;

  unilist_row = g_chunk_new (GtkUniListRow, unilist->row_mem_chunk);
  unilist_row->cell = g_chunk_new (GtkCell, unilist->cell_mem_chunk);

  for (i = 0; i < unilist->columns; i++)
    {
      unilist_row->cell[i].type = GTK_CELL_EMPTY;
      unilist_row->cell[i].vertical = 0;
      unilist_row->cell[i].horizontal = 0;
      unilist_row->cell[i].style = NULL;
    }

  unilist_row->fg_set = FALSE;
  unilist_row->bg_set = FALSE;
  unilist_row->style = NULL;
  unilist_row->selectable = TRUE;
  unilist_row->state = GTK_STATE_NORMAL;
  unilist_row->data = NULL;
  unilist_row->destroy = NULL;

  return unilist_row;
}

static void
row_delete (GtkUniList    *unilist,
	    GtkUniListRow *unilist_row)
{
  gint i;

  for (i = 0; i < unilist->columns; i++)
    {
      GTK_UNILIST_GET_CLASS (unilist)->set_cell_contents
	(unilist, unilist_row, i, GTK_CELL_EMPTY, NULL, 0, NULL, NULL);
      if (unilist_row->cell[i].style)
	{
	  if (GTK_WIDGET_REALIZED (unilist))
	    gtk_style_detach (unilist_row->cell[i].style);
	  g_object_unref (unilist_row->cell[i].style);
	}
    }

  if (unilist_row->style)
    {
      if (GTK_WIDGET_REALIZED (unilist))
        gtk_style_detach (unilist_row->style);
      g_object_unref (unilist_row->style);
    }

  if (unilist_row->destroy)
    unilist_row->destroy (unilist_row->data);

  g_mem_chunk_free (unilist->cell_mem_chunk, unilist_row->cell);
  g_mem_chunk_free (unilist->row_mem_chunk, unilist_row);
}

/* FOCUS FUNCTIONS
 *   gtk_unilist_focus_content_area
 *   gtk_unilist_focus
 *   gtk_unilist_draw_focus
 *   gtk_unilist_focus_in
 *   gtk_unilist_focus_out
 *   title_focus
 */
static void
gtk_unilist_focus_content_area (GtkUniList *unilist)
{
  if (unilist->focus_row < 0)
    {
      unilist->focus_row = 0;
      
      if ((unilist->selection_mode == GTK_SELECTION_BROWSE ||
	   unilist->selection_mode == GTK_SELECTION_MULTIPLE) &&
	  !unilist->selection)
	gtk_signal_emit (GTK_OBJECT (unilist),
			 unilist_signals[SELECT_ROW],
			 unilist->focus_row, -1, NULL);
    }
  gtk_widget_grab_focus (GTK_WIDGET (unilist));
}

static gboolean
gtk_unilist_focus (GtkWidget        *widget,
		 GtkDirectionType  direction)
{
  GtkUniList *unilist = GTK_UNILIST (widget);
  GtkWidget *focus_child;
  gboolean is_current_focus;

  if (!GTK_WIDGET_IS_SENSITIVE (widget))
    return FALSE;

  focus_child = GTK_CONTAINER (widget)->focus_child;
  
  is_current_focus = gtk_widget_is_focus (GTK_WIDGET (unilist));
			  
  if (focus_child &&
      gtk_widget_child_focus (focus_child, direction))
    return TRUE;
      
  switch (direction)
    {
    case GTK_DIR_LEFT:
    case GTK_DIR_RIGHT:
      if (focus_child)
	{
	  if (title_focus_move (unilist, direction))
	    return TRUE;
	}
      else if (!is_current_focus)
	{
	  gtk_unilist_focus_content_area (unilist);
	  return TRUE;
	}
      break;
    case GTK_DIR_DOWN:
    case GTK_DIR_TAB_FORWARD:
      if (!focus_child && !is_current_focus)
	{
	  if (title_focus_in (unilist, direction))
	    return TRUE;
	}
      
      if (!is_current_focus && unilist->rows)
	{
	  gtk_unilist_focus_content_area (unilist);
	  return TRUE;
	}
      break;
    case GTK_DIR_UP:
    case GTK_DIR_TAB_BACKWARD:
      if (!focus_child && is_current_focus)
	{
	  if (title_focus_in (unilist, direction))
	    return TRUE;
	}
      
      if (!is_current_focus && !focus_child && unilist->rows)
	{
	  gtk_unilist_focus_content_area (unilist);
	  return TRUE;
	}
      break;
    default:
      break;
    }

  return FALSE;
}

static void
gtk_unilist_set_focus_child (GtkContainer *container,
			   GtkWidget    *child)
{
  GtkUniList *unilist = GTK_UNILIST (container);
  gint i;

  for (i = 0; i < unilist->columns; i++)
    if (unilist->column[i].button == child)
      unilist->focus_header_column = i;
  
  parent_class->set_focus_child (container, child);
}

static void
gtk_unilist_draw_focus (GtkWidget *widget)
{
  GtkUniList *unilist;

  g_return_if_fail (GTK_IS_UNILIST (widget));

  if (!GTK_WIDGET_DRAWABLE (widget) || !GTK_WIDGET_CAN_FOCUS (widget))
    return;

  unilist = GTK_UNILIST (widget);
  if (unilist->focus_row >= 0)
    gdk_draw_rectangle (unilist->unilist_window, unilist->xor_gc, FALSE,
			0, ROW_TOP_YPIXEL(unilist, unilist->focus_row),
			unilist->unilist_window_width - 1,
			unilist->row_height - 1);
}

static gint
gtk_unilist_focus_in (GtkWidget     *widget,
		    GdkEventFocus *event)
{
  GtkUniList *unilist = GTK_UNILIST (widget);

  if (unilist->selection_mode == GTK_SELECTION_BROWSE &&
      unilist->selection == NULL && unilist->focus_row > -1)
    {
      GList *list;

      list = g_list_nth (unilist->row_list, unilist->focus_row);
      if (list && GTK_UNILIST_ROW (list)->selectable)
	gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[SELECT_ROW],
			 unilist->focus_row, -1, event);
      else
	gtk_unilist_draw_focus (widget);
    }
  else
    gtk_unilist_draw_focus (widget);

  return FALSE;
}

static gint
gtk_unilist_focus_out (GtkWidget     *widget,
		     GdkEventFocus *event)
{
  GtkUniList *unilist = GTK_UNILIST (widget);

  gtk_unilist_draw_focus (widget);
  
  GTK_UNILIST_GET_CLASS (widget)->resync_selection (unilist, (GdkEvent *) event);

  return FALSE;
}

static gboolean
focus_column (GtkUniList *unilist, gint column, gint dir)
{
  GtkWidget *child = unilist->column[column].button;
  
  if (gtk_widget_child_focus (child, dir))
    {
      return TRUE;
    }
  else if (GTK_WIDGET_CAN_FOCUS (child))
    {
      gtk_widget_grab_focus (child);
      return TRUE;
    }

  return FALSE;
}

/* Focus moved onto the headers. Focus first focusable and visible child.
 * (FIXME: focus the last focused child if visible)
 */
static gboolean
title_focus_in (GtkUniList *unilist, gint dir)
{
  gint i;
  gint left, right;

  if (!GTK_UNILIST_SHOW_TITLES (unilist))
    return FALSE;

  /* Check last focused column */
  if (unilist->focus_header_column != -1)
    {
      i = unilist->focus_header_column;
      
      left = COLUMN_LEFT_XPIXEL (unilist, i);
      right = left + unilist->column[i].area.width;
      
      if (left >= 0 && right <= unilist->unilist_window_width)
	{
	  if (focus_column (unilist, i, dir))
	    return TRUE;
	}
    }

  /* Check fully visible columns */
  for (i = 0 ; i < unilist->columns ; i++)
    {
      left = COLUMN_LEFT_XPIXEL (unilist, i);
      right = left + unilist->column[i].area.width;
      
      if (left >= 0 && right <= unilist->unilist_window_width)
	{
	  if (focus_column (unilist, i, dir))
	    return TRUE;
	}
    }

  /* Check partially visible columns */
  for (i = 0 ; i < unilist->columns ; i++)
    {
      left = COLUMN_LEFT_XPIXEL (unilist, i);
      right = left + unilist->column[i].area.width;

      if ((left < 0 && right > 0) ||
	  (left < unilist->unilist_window_width && right > unilist->unilist_window_width))
	{
	  if (focus_column (unilist, i, dir))
	    return TRUE;
	}
    }
      
  return FALSE;
}

/* Move the focus right or left within the title buttons, scrolling
 * as necessary to keep the focused child visible.
 */
static gboolean
title_focus_move (GtkUniList *unilist,
		  gint      dir)
{
  GtkWidget *focus_child;
  gboolean return_val = FALSE;
  gint d = 0;
  gint i = -1;
  gint j;

  if (!GTK_UNILIST_SHOW_TITLES(unilist))
    return FALSE;

  focus_child = GTK_CONTAINER (unilist)->focus_child;
  g_assert (focus_child);

  /* Movement direction within headers
   */
  switch (dir)
    {
    case GTK_DIR_RIGHT:
      d = 1;
      break;
    case GTK_DIR_LEFT:
      d = -1;
      break;
    }
  
  for (i = 0; i < unilist->columns; i++)
    if (unilist->column[i].button == focus_child)
      break;
  
  g_assert (i != -1);		/* Have a starting column */
  
  j = i + d;
  while (!return_val && j >= 0 && j < unilist->columns)
    {
      if (unilist->column[j].button &&
	  GTK_WIDGET_VISIBLE (unilist->column[j].button))
	{
	  if (focus_column (unilist, j, dir))
	    {
	      return_val = TRUE;
	      break;
	    }
	}
      j += d;
    }

  /* If we didn't find it, wrap around and keep looking
   */
  if (!return_val)
    {
      j = d > 0 ? 0 : unilist->columns - 1;

      while (!return_val && j != i)
	{
	  if (unilist->column[j].button &&
	      GTK_WIDGET_VISIBLE (unilist->column[j].button))
	    {
	      if (focus_column (unilist, j, dir))
		{
		  return_val = TRUE;
		  break;
		}
	    }
	  j += d;
	}
    }

  /* Scroll horizontally so focused column is visible
   */
  if (return_val)
    {
      if (COLUMN_LEFT_XPIXEL (unilist, j) < CELL_SPACING + COLUMN_INSET)
	gtk_unilist_moveto (unilist, -1, j, 0, 0);
      else if (COLUMN_LEFT_XPIXEL(unilist, j) + unilist->column[j].area.width >
	       unilist->unilist_window_width)
	{
	  gint last_column;
	  
	  for (last_column = unilist->columns - 1;
	       last_column >= 0 && !unilist->column[last_column].visible; last_column--);

	  if (j == last_column)
	    gtk_unilist_moveto (unilist, -1, j, 0, 0);
	  else
	    gtk_unilist_moveto (unilist, -1, j, 0, 1);
	}
    }
  return TRUE;			/* Even if we didn't find a new one, we can keep the
				 * focus in the same place.
				 */
}

/* PRIVATE SCROLLING FUNCTIONS
 *   move_focus_row
 *   scroll_horizontal
 *   scroll_vertical
 *   move_horizontal
 *   move_vertical
 *   horizontal_timeout
 *   vertical_timeout
 *   remove_grab
 */
static void
move_focus_row (GtkUniList      *unilist,
		GtkScrollType  scroll_type,
		gfloat         position)
{
  GtkWidget *widget;

  g_return_if_fail (unilist != 0);
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  widget = GTK_WIDGET (unilist);

  switch (scroll_type)
    {
    case GTK_SCROLL_STEP_UP:
    case GTK_SCROLL_STEP_BACKWARD:
      if (unilist->focus_row <= 0)
	return;
      gtk_unilist_draw_focus (widget);
      unilist->focus_row--;
      gtk_unilist_draw_focus (widget);
      break;

    case GTK_SCROLL_STEP_DOWN:
    case GTK_SCROLL_STEP_FORWARD:
      if (unilist->focus_row >= unilist->rows - 1)
	return;
      gtk_unilist_draw_focus (widget);
      unilist->focus_row++;
      gtk_unilist_draw_focus (widget);
      break;
    case GTK_SCROLL_PAGE_UP:
    case GTK_SCROLL_PAGE_BACKWARD:
      if (unilist->focus_row <= 0)
	return;
      gtk_unilist_draw_focus (widget);
      unilist->focus_row = MAX (0, unilist->focus_row -
			      (2 * unilist->unilist_window_height -
			       unilist->row_height - CELL_SPACING) / 
			      (2 * (unilist->row_height + CELL_SPACING)));
      gtk_unilist_draw_focus (widget);
      break;
    case GTK_SCROLL_PAGE_DOWN:
    case GTK_SCROLL_PAGE_FORWARD:
      if (unilist->focus_row >= unilist->rows - 1)
	return;
      gtk_unilist_draw_focus (widget);
      unilist->focus_row = MIN (unilist->rows - 1, unilist->focus_row + 
			      (2 * unilist->unilist_window_height -
			       unilist->row_height - CELL_SPACING) / 
			      (2 * (unilist->row_height + CELL_SPACING)));
      gtk_unilist_draw_focus (widget);
      break;
    case GTK_SCROLL_JUMP:
      if (position >= 0 && position <= 1)
	{
	  gtk_unilist_draw_focus (widget);
	  unilist->focus_row = position * (unilist->rows - 1);
	  gtk_unilist_draw_focus (widget);
	}
      break;
    default:
      break;
    }
}

static void
scroll_horizontal (GtkUniList      *unilist,
		   GtkScrollType  scroll_type,
		   gfloat         position)
{
  gint column = 0;
  gint last_column;

  g_return_if_fail (unilist != 0);
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (gdk_pointer_is_grabbed () && GTK_WIDGET_HAS_GRAB (unilist))
    return;

  for (last_column = unilist->columns - 1;
       last_column >= 0 && !unilist->column[last_column].visible; last_column--)
    ;

  switch (scroll_type)
    {
    case GTK_SCROLL_STEP_BACKWARD:
      column = COLUMN_FROM_XPIXEL (unilist, 0);
      if (COLUMN_LEFT_XPIXEL (unilist, column) - CELL_SPACING - COLUMN_INSET >= 0
	  && column > 0)
	column--;
      break;
    case GTK_SCROLL_STEP_FORWARD:
      column = COLUMN_FROM_XPIXEL (unilist, unilist->unilist_window_width);
      if (column < 0)
	return;
      if (COLUMN_LEFT_XPIXEL (unilist, column) +
	  unilist->column[column].area.width +
	  CELL_SPACING + COLUMN_INSET - 1 <= unilist->unilist_window_width &&
	  column < last_column)
	column++;
      break;
    case GTK_SCROLL_PAGE_BACKWARD:
    case GTK_SCROLL_PAGE_FORWARD:
      return;
    case GTK_SCROLL_JUMP:
      if (position >= 0 && position <= 1)
	{
	  gint vis_columns = 0;
	  gint i;

	  for (i = 0; i <= last_column; i++)
 	    if (unilist->column[i].visible)
	      vis_columns++;

	  column = position * vis_columns;

	  for (i = 0; i <= last_column && column > 0; i++)
	    if (unilist->column[i].visible)
	      column--;

	  column = i;
	}
      else
	return;
      break;
    default:
      break;
    }

  if (COLUMN_LEFT_XPIXEL (unilist, column) < CELL_SPACING + COLUMN_INSET)
    gtk_unilist_moveto (unilist, -1, column, 0, 0);
  else if (COLUMN_LEFT_XPIXEL (unilist, column) + CELL_SPACING + COLUMN_INSET - 1
	   + unilist->column[column].area.width > unilist->unilist_window_width)
    {
      if (column == last_column)
	gtk_unilist_moveto (unilist, -1, column, 0, 0);
      else
	gtk_unilist_moveto (unilist, -1, column, 0, 1);
    }
}

static void
scroll_vertical (GtkUniList      *unilist,
		 GtkScrollType  scroll_type,
		 gfloat         position)
{
  gint old_focus_row;
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (gdk_pointer_is_grabbed () && GTK_WIDGET_HAS_GRAB (unilist))
    return;

  switch (unilist->selection_mode)
    {
    case GTK_SELECTION_MULTIPLE:
      if (unilist->anchor >= 0)
	return;
    case GTK_SELECTION_BROWSE:

      old_focus_row = unilist->focus_row;
      move_focus_row (unilist, scroll_type, position);

      if (old_focus_row != unilist->focus_row)
	{
	  if (unilist->selection_mode == GTK_SELECTION_BROWSE)
	    gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[UNSELECT_ROW],
			     old_focus_row, -1, NULL);
	  else if (!GTK_UNILIST_ADD_MODE(unilist))
	    {
	      gtk_unilist_unselect_all (unilist);
	      unilist->undo_anchor = old_focus_row;
	    }
	}

      switch (gtk_unilist_row_is_visible (unilist, unilist->focus_row))
	{
	case GTK_VISIBILITY_NONE:
//	  if (old_focus_row != unilist->focus_row &&
//	      !(unilist->selection_mode == GTK_SELECTION_MULTIPLE &&
//		GTK_UNILIST_ADD_MODE(unilist)))
//	    gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[SELECT_ROW],
//			     unilist->focus_row, -1, NULL);
	  switch (scroll_type)
	    {
            case GTK_SCROLL_PAGE_UP:
            case GTK_SCROLL_STEP_UP:
	    case GTK_SCROLL_STEP_BACKWARD:
	    case GTK_SCROLL_PAGE_BACKWARD:
	      gtk_unilist_moveto (unilist, unilist->focus_row, -1, 0, 0);
	      break;
            case GTK_SCROLL_PAGE_DOWN:
            case GTK_SCROLL_STEP_DOWN:
	    case GTK_SCROLL_STEP_FORWARD:
	    case GTK_SCROLL_PAGE_FORWARD:
	      gtk_unilist_moveto (unilist, unilist->focus_row, -1, 1, 0);
	      break;
	    case GTK_SCROLL_JUMP:
	      gtk_unilist_moveto (unilist, unilist->focus_row, -1, 0.5, 0);
	      break;
	    default:
	      break;
	    }
	  break;
	case GTK_VISIBILITY_PARTIAL:
	  switch (scroll_type)
	    {
	    case GTK_SCROLL_STEP_BACKWARD:
	    case GTK_SCROLL_PAGE_BACKWARD:
	      gtk_unilist_moveto (unilist, unilist->focus_row, -1, 0, 0);
	      break;
	    case GTK_SCROLL_STEP_FORWARD:
	    case GTK_SCROLL_PAGE_FORWARD:
	      gtk_unilist_moveto (unilist, unilist->focus_row, -1, 1, 0);
	      break;
	    case GTK_SCROLL_JUMP:
	      gtk_unilist_moveto (unilist, unilist->focus_row, -1, 0.5, 0);
	      break;
	    default:
	      break;
	    }
	default:
//	  if (old_focus_row != unilist->focus_row &&
//	      !(unilist->selection_mode == GTK_SELECTION_MULTIPLE &&
//		GTK_UNILIST_ADD_MODE(unilist)))
//	    gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[SELECT_ROW],
//			     unilist->focus_row, 0, NULL);
//	    gtk_signal_emit (GTK_OBJECT (unilist), unilist_signals[UNSELECT_ROW],
//			     unilist->focus_row, -1, NULL);
	  break;
	}
      break;
    default:
      move_focus_row (unilist, scroll_type, position);

      if (ROW_TOP_YPIXEL (unilist, unilist->focus_row) + unilist->row_height >
	  unilist->unilist_window_height)
	gtk_unilist_moveto (unilist, unilist->focus_row, -1, 1, 0);
      else if (ROW_TOP_YPIXEL (unilist, unilist->focus_row) < 0)
	gtk_unilist_moveto (unilist, unilist->focus_row, -1, 0, 0);
      break;
    }
}

static void
move_horizontal (GtkUniList *unilist,
		 gint      diff)
{
  gdouble value;

  if (!unilist->hadjustment)
    return;

  value = CLAMP (unilist->hadjustment->value + diff, 0.0,
		 unilist->hadjustment->upper - unilist->hadjustment->page_size);
  gtk_adjustment_set_value (unilist->hadjustment, value);
}

static void
move_vertical (GtkUniList *unilist,
	       gint      row,
	       gfloat    align)
{
  gdouble value;

  if (!unilist->vadjustment)
    return;

  value = (ROW_TOP_YPIXEL (unilist, row) - unilist->voffset -
	   align * (unilist->unilist_window_height - unilist->row_height) +
	   (2 * align - 1) * CELL_SPACING);

  if (value + unilist->vadjustment->page_size > unilist->vadjustment->upper)
    value = unilist->vadjustment->upper - unilist->vadjustment->page_size;

  gtk_adjustment_set_value (unilist->vadjustment, value);
}

static gint
horizontal_timeout (GtkUniList *unilist)
{
  GdkEventMotion event;

  memset (&event, 0, sizeof (event));

  GDK_THREADS_ENTER ();

  unilist->htimer = 0;

  event.type = GDK_MOTION_NOTIFY;
  event.send_event = TRUE;

  gtk_unilist_motion (GTK_WIDGET (unilist), &event);

  GDK_THREADS_LEAVE ();
  
  return FALSE;
}

static gint
vertical_timeout (GtkUniList *unilist)
{
  GdkEventMotion event;

  memset (&event, 0, sizeof (event));

  GDK_THREADS_ENTER ();

  unilist->vtimer = 0;

  event.type = GDK_MOTION_NOTIFY;
  event.send_event = TRUE;

  gtk_unilist_motion (GTK_WIDGET (unilist), &event);

  GDK_THREADS_LEAVE ();

  return FALSE;
}

static void
remove_grab (GtkUniList *unilist)
{
  if (GTK_WIDGET_HAS_GRAB (unilist))
    {
      gtk_grab_remove (GTK_WIDGET (unilist));
      if (gdk_pointer_is_grabbed ())
	gdk_pointer_ungrab (GDK_CURRENT_TIME);
    }

  if (unilist->htimer)
    {
      gtk_timeout_remove (unilist->htimer);
      unilist->htimer = 0;
    }

  if (unilist->vtimer)
    {
      gtk_timeout_remove (unilist->vtimer);
      unilist->vtimer = 0;
    }
}

/* PUBLIC SORTING FUNCTIONS
 * gtk_unilist_sort
 * gtk_unilist_set_compare_func
 * gtk_unilist_set_auto_sort
 * gtk_unilist_set_sort_type
 * gtk_unilist_set_sort_column
 */
void
gtk_unilist_sort (GtkUniList *unilist)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  GTK_UNILIST_GET_CLASS (unilist)->sort_list (unilist);
}

void
gtk_unilist_set_compare_func (GtkUniList            *unilist,
			    GtkUniListCompareFunc  cmp_func)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  unilist->compare = (cmp_func) ? cmp_func : default_compare;
}

void       
gtk_unilist_set_auto_sort (GtkUniList *unilist,
			 gboolean  auto_sort)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));
  
  if (GTK_UNILIST_AUTO_SORT(unilist) && !auto_sort)
    GTK_UNILIST_UNSET_FLAG (unilist, UNILIST_AUTO_SORT);
  else if (!GTK_UNILIST_AUTO_SORT(unilist) && auto_sort)
    {
      GTK_UNILIST_SET_FLAG (unilist, UNILIST_AUTO_SORT);
      gtk_unilist_sort (unilist);
    }
}

void       
gtk_unilist_set_sort_type (GtkUniList    *unilist,
			 GtkSortType  sort_type)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));
  
  unilist->sort_type = sort_type;
}

void
gtk_unilist_set_sort_column (GtkUniList *unilist,
			   gint      column)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (column < 0 || column >= unilist->columns)
    return;

  unilist->sort_column = column;
}

/* PRIVATE SORTING FUNCTIONS
 *   default_compare
 *   real_sort_list
 *   gtk_unilist_merge
 *   gtk_unilist_mergesort
 */
static gint
default_compare (GtkUniList      *unilist,
		 gconstpointer  ptr1,
		 gconstpointer  ptr2)
{
  char *text1 = NULL;
  char *text2 = NULL;

  GtkUniListRow *row1 = (GtkUniListRow *) ptr1;
  GtkUniListRow *row2 = (GtkUniListRow *) ptr2;

  switch (row1->cell[unilist->sort_column].type)
    {
    case GTK_CELL_TEXT:
      text1 = GTK_CELL_TEXT (row1->cell[unilist->sort_column])->text;
      break;
    case GTK_CELL_PIXTEXT:
      text1 = GTK_CELL_PIXTEXT (row1->cell[unilist->sort_column])->text;
      break;
    default:
      break;
    }
 
  switch (row2->cell[unilist->sort_column].type)
    {
    case GTK_CELL_TEXT:
      text2 = GTK_CELL_TEXT (row2->cell[unilist->sort_column])->text;
      break;
    case GTK_CELL_PIXTEXT:
      text2 = GTK_CELL_PIXTEXT (row2->cell[unilist->sort_column])->text;
      break;
    default:
      break;
    }

  if (!text2)
    return (text1 != NULL);

  if (!text1)
    return -1;

  return strcmp (text1, text2);
}

static void
real_sort_list (GtkUniList *unilist)
{
  GList *list;
  GList *work;
  gint i;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (unilist->rows <= 1)
    return;

  if (gdk_pointer_is_grabbed () && GTK_WIDGET_HAS_GRAB (unilist))
    return;

  gtk_unilist_freeze (unilist);

  if (unilist->anchor != -1 && unilist->selection_mode == GTK_SELECTION_MULTIPLE)
    {
      GTK_UNILIST_GET_CLASS (unilist)->resync_selection (unilist, NULL);
      g_list_free (unilist->undo_selection);
      g_list_free (unilist->undo_unselection);
      unilist->undo_selection = NULL;
      unilist->undo_unselection = NULL;
    }
   
  unilist->row_list = gtk_unilist_mergesort (unilist, unilist->row_list, unilist->rows);

  work = unilist->selection;

  for (i = 0, list = unilist->row_list; i < unilist->rows; i++, list = list->next)
    {
      if (GTK_UNILIST_ROW (list)->state == GTK_STATE_SELECTED)
	{
	  work->data = GINT_TO_POINTER (i);
	  work = work->next;
	}
      
      if (i == unilist->rows - 1)
	unilist->row_list_end = list;
    }

  gtk_unilist_thaw (unilist);
}

static GList *
gtk_unilist_merge (GtkUniList *unilist,
		 GList    *a,         /* first list to merge */
		 GList    *b)         /* second list to merge */
{
  GList z = { 0 };                    /* auxiliary node */
  GList *c;
  gint cmp;

  c = &z;

  while (a || b)
    {
      if (a && !b)
	{
	  c->next = a;
	  a->prev = c;
	  c = a;
	  a = a->next;
	  break;
	}
      else if (!a && b)
	{
	  c->next = b;
	  b->prev = c;
	  c = b;
	  b = b->next;
	  break;
	}
      else /* a && b */
	{
	  cmp = unilist->compare (unilist, GTK_UNILIST_ROW (a), GTK_UNILIST_ROW (b));
	  if ((cmp >= 0 && unilist->sort_type == GTK_SORT_DESCENDING) ||
	      (cmp <= 0 && unilist->sort_type == GTK_SORT_ASCENDING) ||
	      (a && !b))
	    {
	      c->next = a;
	      a->prev = c;
	      c = a;
	      a = a->next;
	    }
	  else
	    {
	      c->next = b;
	      b->prev = c;
	      c = b;
	      b = b->next;
	    }
	}
    }

  z.next->prev = NULL;
  return z.next;
}

static GList *
gtk_unilist_mergesort (GtkUniList *unilist,
		     GList    *list,         /* the list to sort */
		     gint      num)          /* the list's length */
{
  GList *half;
  gint i;

  if (num <= 1)
    {
      return list;
    }
  else
    {
      /* move "half" to the middle */
      half = list;
      for (i = 0; i < num / 2; i++)
	half = half->next;

      /* cut the list in two */
      half->prev->next = NULL;
      half->prev = NULL;

      /* recursively sort both lists */
      return gtk_unilist_merge (unilist,
		       gtk_unilist_mergesort (unilist, list, num / 2),
		       gtk_unilist_mergesort (unilist, half, num - num / 2));
    }
}

/************************/

static void
drag_source_info_destroy (gpointer data)
{
  GtkUniListCellInfo *info = data;

  g_free (info);
}

static void
drag_dest_info_destroy (gpointer data)
{
  GtkUniListDestInfo *info = data;

  g_free (info);
}

static void
drag_dest_cell (GtkUniList         *unilist,
		gint              x,
		gint              y,
		GtkUniListDestInfo *dest_info)
{
  GtkWidget *widget;

  widget = GTK_WIDGET (unilist);

  dest_info->insert_pos = GTK_UNILIST_DRAG_NONE;

  y -= (GTK_CONTAINER (unilist)->border_width +
	widget->style->ythickness +
	unilist->column_title_area.height);

  dest_info->cell.row = ROW_FROM_YPIXEL (unilist, y);
  if (dest_info->cell.row >= unilist->rows)
    {
      dest_info->cell.row = unilist->rows - 1;
      y = ROW_TOP_YPIXEL (unilist, dest_info->cell.row) + unilist->row_height;
    }
  if (dest_info->cell.row < -1)
    dest_info->cell.row = -1;
  
  x -= GTK_CONTAINER (widget)->border_width + widget->style->xthickness;

  dest_info->cell.column = COLUMN_FROM_XPIXEL (unilist, x);

  if (dest_info->cell.row >= 0)
    {
      gint y_delta;
      gint h = 0;

      y_delta = y - ROW_TOP_YPIXEL (unilist, dest_info->cell.row);
      
      if (GTK_UNILIST_DRAW_DRAG_RECT(unilist))
	{
	  dest_info->insert_pos = GTK_UNILIST_DRAG_INTO;
	  h = unilist->row_height / 4;
	}
      else if (GTK_UNILIST_DRAW_DRAG_LINE(unilist))
	{
	  dest_info->insert_pos = GTK_UNILIST_DRAG_BEFORE;
	  h = unilist->row_height / 2;
	}

      if (GTK_UNILIST_DRAW_DRAG_LINE(unilist))
	{
	  if (y_delta < h)
	    dest_info->insert_pos = GTK_UNILIST_DRAG_BEFORE;
	  else if (unilist->row_height - y_delta < h)
	    dest_info->insert_pos = GTK_UNILIST_DRAG_AFTER;
	}
    }
}

static void
gtk_unilist_drag_begin (GtkWidget	     *widget,
		      GdkDragContext *context)
{
  GtkUniList *unilist;
  GtkUniListCellInfo *info;

  g_return_if_fail (GTK_IS_UNILIST (widget));
  g_return_if_fail (context != NULL);

  unilist = GTK_UNILIST (widget);

  unilist->drag_button = 0;
  remove_grab (unilist);

  switch (unilist->selection_mode)
    {
    case GTK_SELECTION_MULTIPLE:
      update_extended_selection (unilist, unilist->focus_row);
//      GTK_UNILIST_GET_CLASS (unilist)->resync_selection (unilist, NULL);
      break;
    case GTK_SELECTION_SINGLE:
      unilist->anchor = -1;
    case GTK_SELECTION_BROWSE:
      break;
    default:
      g_assert_not_reached ();
    }

  info = g_dataset_get_data (context, "gtk-unilist-drag-source");

  if (!info)
    {
      info = g_new (GtkUniListCellInfo, 1);

      if (unilist->click_cell.row < 0)
	unilist->click_cell.row = 0;
      else if (unilist->click_cell.row >= unilist->rows)
	unilist->click_cell.row = unilist->rows - 1;
      info->row = unilist->click_cell.row;
      info->column = unilist->click_cell.column;

      g_dataset_set_data_full (context, "gtk-unilist-drag-source", info,
			       drag_source_info_destroy);
    }

  if (GTK_UNILIST_USE_DRAG_ICONS (unilist))
    gtk_drag_set_icon_default (context);
}

static void
gtk_unilist_drag_end (GtkWidget	   *widget,
		    GdkDragContext *context)
{
  GtkUniList *unilist;

  g_return_if_fail (GTK_IS_UNILIST (widget));
  g_return_if_fail (context != NULL);

  unilist = GTK_UNILIST (widget);

  unilist->click_cell.row = -1;
  unilist->click_cell.column = -1;
}

static void
gtk_unilist_drag_leave (GtkWidget      *widget,
		      GdkDragContext *context,
		      guint           time)
{
  GtkUniList *unilist;
  GtkUniListDestInfo *dest_info;

  g_return_if_fail (GTK_IS_UNILIST (widget));
  g_return_if_fail (context != NULL);

  unilist = GTK_UNILIST (widget);

  dest_info = g_dataset_get_data (context, "gtk-unilist-drag-dest");
  
  if (dest_info)
    {
      if (dest_info->cell.row >= 0 &&
	  GTK_UNILIST_REORDERABLE(unilist) &&
	  gtk_drag_get_source_widget (context) == widget)
	{
	  GList *list;
	  GdkAtom atom = gdk_atom_intern ("gtk-unilist-drag-reorder", FALSE);

	  list = context->targets;
	  while (list)
	    {
	      if (atom == GDK_POINTER_TO_ATOM (list->data))
		{
		  GTK_UNILIST_GET_CLASS (unilist)->draw_drag_highlight
		    (unilist,
		     g_list_nth (unilist->row_list, dest_info->cell.row)->data,
		     dest_info->cell.row, dest_info->insert_pos);
		  unilist->drag_highlight_row = -1;
		  break;
		}
	      list = list->next;
	    }
	}
      g_dataset_remove_data (context, "gtk-unilist-drag-dest");
    }
}

static gint
gtk_unilist_drag_motion (GtkWidget      *widget,
		       GdkDragContext *context,
		       gint            x,
		       gint            y,
		       guint           time)
{
  GtkUniList *unilist;
  GtkUniListDestInfo new_info;
  GtkUniListDestInfo *dest_info;

  g_return_val_if_fail (GTK_IS_UNILIST (widget), FALSE);

  unilist = GTK_UNILIST (widget);

  dest_info = g_dataset_get_data (context, "gtk-unilist-drag-dest");

  if (!dest_info)
    {
      dest_info = g_new (GtkUniListDestInfo, 1);

      dest_info->insert_pos  = GTK_UNILIST_DRAG_NONE;
      dest_info->cell.row    = -1;
      dest_info->cell.column = -1;

      g_dataset_set_data_full (context, "gtk-unilist-drag-dest", dest_info,
			       drag_dest_info_destroy);
    }

  drag_dest_cell (unilist, x, y, &new_info);

  if (GTK_UNILIST_REORDERABLE (unilist))
    {
      GList *list;
      GdkAtom atom = gdk_atom_intern ("gtk-unilist-drag-reorder", FALSE);

      list = context->targets;
      while (list)
	{
	  if (atom == GDK_POINTER_TO_ATOM (list->data))
	    break;
	  list = list->next;
	}

      if (list)
	{
	  if (gtk_drag_get_source_widget (context) != widget ||
	      new_info.insert_pos == GTK_UNILIST_DRAG_NONE ||
	      new_info.cell.row == unilist->click_cell.row ||
	      (new_info.cell.row == unilist->click_cell.row - 1 &&
	       new_info.insert_pos == GTK_UNILIST_DRAG_AFTER) ||
	      (new_info.cell.row == unilist->click_cell.row + 1 &&
	       new_info.insert_pos == GTK_UNILIST_DRAG_BEFORE))
	    {
	      if (dest_info->cell.row < 0)
		{
		  gdk_drag_status (context, GDK_ACTION_DEFAULT, time);
		  return FALSE;
		}
	      return TRUE;
	    }
		
	  if (new_info.cell.row != dest_info->cell.row ||
	      (new_info.cell.row == dest_info->cell.row &&
	       dest_info->insert_pos != new_info.insert_pos))
	    {
	      if (dest_info->cell.row >= 0)
		GTK_UNILIST_GET_CLASS (unilist)->draw_drag_highlight
		  (unilist, g_list_nth (unilist->row_list,
				      dest_info->cell.row)->data,
		   dest_info->cell.row, dest_info->insert_pos);

	      dest_info->insert_pos  = new_info.insert_pos;
	      dest_info->cell.row    = new_info.cell.row;
	      dest_info->cell.column = new_info.cell.column;
	      
	      GTK_UNILIST_GET_CLASS (unilist)->draw_drag_highlight
		(unilist, g_list_nth (unilist->row_list,
				    dest_info->cell.row)->data,
		 dest_info->cell.row, dest_info->insert_pos);
	      
	      unilist->drag_highlight_row = dest_info->cell.row;
	      unilist->drag_highlight_pos = dest_info->insert_pos;

	      gdk_drag_status (context, context->suggested_action, time);
	    }
	  return TRUE;
	}
    }

  dest_info->insert_pos  = new_info.insert_pos;
  dest_info->cell.row    = new_info.cell.row;
  dest_info->cell.column = new_info.cell.column;
  return TRUE;
}

static gboolean
gtk_unilist_drag_drop (GtkWidget      *widget,
		     GdkDragContext *context,
		     gint            x,
		     gint            y,
		     guint           time)
{
  g_return_val_if_fail (GTK_IS_UNILIST (widget), FALSE);
  g_return_val_if_fail (context != NULL, FALSE);

  if (GTK_UNILIST_REORDERABLE (widget) &&
      gtk_drag_get_source_widget (context) == widget)
    {
      GList *list;
      GdkAtom atom = gdk_atom_intern ("gtk-unilist-drag-reorder", FALSE);

      list = context->targets;
      while (list)
	{
	  if (atom == GDK_POINTER_TO_ATOM (list->data))
	    return TRUE;
	  list = list->next;
	}
    }
  return FALSE;
}

static void
gtk_unilist_drag_data_received (GtkWidget        *widget,
			      GdkDragContext   *context,
			      gint              x,
			      gint              y,
			      GtkSelectionData *selection_data,
			      guint             info,
			      guint             time)
{
  GtkUniList *unilist;

  g_return_if_fail (GTK_IS_UNILIST (widget));
  g_return_if_fail (context != NULL);
  g_return_if_fail (selection_data != NULL);

  unilist = GTK_UNILIST (widget);

  if (GTK_UNILIST_REORDERABLE (unilist) &&
      gtk_drag_get_source_widget (context) == widget &&
      selection_data->target ==
      gdk_atom_intern ("gtk-unilist-drag-reorder", FALSE) &&
      selection_data->format == 8 &&
      selection_data->length == sizeof (GtkUniListCellInfo))
    {
      GtkUniListCellInfo *source_info;

      source_info = (GtkUniListCellInfo *)(selection_data->data);
      if (source_info)
	{
	  GtkUniListDestInfo dest_info;

	  drag_dest_cell (unilist, x, y, &dest_info);

	  if (dest_info.insert_pos == GTK_UNILIST_DRAG_AFTER)
	    dest_info.cell.row++;
	  if (source_info->row < dest_info.cell.row)
	    dest_info.cell.row--;
	  if (dest_info.cell.row != source_info->row)
	    gtk_unilist_row_move (unilist, source_info->row, dest_info.cell.row);

	  g_dataset_remove_data (context, "gtk-unilist-drag-dest");
	}
    }
}

static void  
gtk_unilist_drag_data_get (GtkWidget        *widget,
			 GdkDragContext   *context,
			 GtkSelectionData *selection_data,
			 guint             info,
			 guint             time)
{
  g_return_if_fail (GTK_IS_UNILIST (widget));
  g_return_if_fail (context != NULL);
  g_return_if_fail (selection_data != NULL);

  if (selection_data->target ==
      gdk_atom_intern ("gtk-unilist-drag-reorder", FALSE))
    {
      GtkUniListCellInfo *info;

      info = g_dataset_get_data (context, "gtk-unilist-drag-source");

      if (info)
	{
	  GtkUniListCellInfo ret_info;

	  ret_info.row = info->row;
	  ret_info.column = info->column;

	  gtk_selection_data_set (selection_data, selection_data->target,
				  8, (guchar *) &ret_info,
				  sizeof (GtkUniListCellInfo));
	}
    }
}

static void
draw_drag_highlight (GtkUniList        *unilist,
		     GtkUniListRow     *dest_row,
		     gint             dest_row_number,
		     GtkUniListDragPos  drag_pos)
{
  gint y;

  y = ROW_TOP_YPIXEL (unilist, dest_row_number) - 1;

  switch (drag_pos)
    {
    case GTK_UNILIST_DRAG_NONE:
      break;
    case GTK_UNILIST_DRAG_AFTER:
      y += unilist->row_height + 1;
    case GTK_UNILIST_DRAG_BEFORE:
      gdk_draw_line (unilist->unilist_window, unilist->xor_gc,
		     0, y, unilist->unilist_window_width, y);
      break;
    case GTK_UNILIST_DRAG_INTO:
      gdk_draw_rectangle (unilist->unilist_window, unilist->xor_gc, FALSE, 0, y,
			  unilist->unilist_window_width - 1, unilist->row_height);
      break;
    }
}

void
gtk_unilist_set_reorderable (GtkUniList *unilist, 
			   gboolean  reorderable)
{
  GtkWidget *widget;

  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if ((GTK_UNILIST_REORDERABLE(unilist) != 0) == reorderable)
    return;

  widget = GTK_WIDGET (unilist);

  if (reorderable)
    {
      GTK_UNILIST_SET_FLAG (unilist, UNILIST_REORDERABLE);
      gtk_drag_dest_set (widget,
			 GTK_DEST_DEFAULT_MOTION | GTK_DEST_DEFAULT_DROP,
			 &unilist_target_table, 1, GDK_ACTION_MOVE);
    }
  else
    {
      GTK_UNILIST_UNSET_FLAG (unilist, UNILIST_REORDERABLE);
      gtk_drag_dest_unset (GTK_WIDGET (unilist));
    }
}

void
gtk_unilist_set_use_drag_icons (GtkUniList *unilist,
			      gboolean  use_icons)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));

  if (use_icons != 0)
    GTK_UNILIST_SET_FLAG (unilist, UNILIST_USE_DRAG_ICONS);
  else
    GTK_UNILIST_UNSET_FLAG (unilist, UNILIST_USE_DRAG_ICONS);
}

void
gtk_unilist_set_button_actions (GtkUniList *unilist,
			      guint     button,
			      guint8    button_actions)
{
  g_return_if_fail (GTK_IS_UNILIST (unilist));
  
  if (button < MAX_BUTTON)
    {
      if (gdk_pointer_is_grabbed () || GTK_WIDGET_HAS_GRAB (unilist))
	{
	  remove_grab (unilist);
	  unilist->drag_button = 0;
	}

      GTK_UNILIST_GET_CLASS (unilist)->resync_selection (unilist, NULL);

      unilist->button_actions[button] = button_actions;
    }
}      
gint selection_changed (GtkUniList *unilist){}

